Package: ggstratify
Title: Fast Stratified Descriptive Figures with a Point-and-Click GUI
Version: 0.0.1
Authors@R: 
    c(
      person(
        "Akihiro", "Shiroshita",
        email = "akihirokun8@gmail.com",
        role = c("aut", "cre", "cph")
      ),
      person(
        "Yuki", "Kataoka",
        email = "youkiti@gmail.com",
        role = "aut"
      )
    )
Description: A point-and-click 'shiny' interface for the descriptive analysis
    that comes before any model is chosen. Pass a data frame, pick the
    variable to describe, and add the layers you want to see it within: a
    second variable becomes the panels of a 'ggplot2' facet_wrap(), and
    further variables become separate figures, one file each, taken either one
    variable at a time or crossed. Every stratum is reported with the number
    of observations behind it, on the figure and on each of its panels; strata
    that contain none are listed rather than dropped, and rows with a missing
    value in a layer variable are excluded and counted. A continuous variable
    can be categorized into quantile groups, equal-width bins or user-supplied
    cut points and then used as a layer; the figure types follow those offered
    by the 'ggplotgui' package and add the line plot for change over time, an
    optional LOWESS smoother, and the Kaplan-Meier curve estimated by
    'survival', with an optional number-at-risk table. Columns are described
    as they are typed, so convert each to the type you mean first. Figures are
    written as PNG or SVG, and the application prints the 'ggplot2' code 
    behind the figure on screen, so that a description can be repeated, shared
    or accounted for later. Everything runs locally, with no network access and no AI involved.
License: GPL-3
URL: https://github.com/AkiShiroshita/ggstratify,
        https://akishiroshita.github.io/ggstratify/
BugReports: https://github.com/AkiShiroshita/ggstratify/issues
Encoding: UTF-8
Depends: R (>= 4.1)
Imports: bslib (>= 0.5.0), checkmate (>= 2.1.0), data.table (>=
        1.14.0), ggplot2 (>= 3.4.0), grDevices, patchwork (>= 1.1.0),
        ragg, shiny (>= 1.7.0), stats, survival (>= 3.2.0), utils
Suggests: knitr, rmarkdown, svglite (>= 2.1.0), testthat (>= 3.0.0),
        tibble
VignetteBuilder: knitr
LazyData: true
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-08-22 09:00:30 UTC; shiroa1
Author: Akihiro Shiroshita [aut, cre, cph],
  Yuki Kataoka [aut]
Maintainer: Akihiro Shiroshita <akihirokun8@gmail.com>
Repository: CRAN
Date/Publication: 2026-09-02 11:40:02 UTC

# Generated by roxygen2: do not edit by hand

export(ggstratify)
import(data.table)
import(ggplot2)
import(shiny)
importFrom(grDevices,svg)
importFrom(patchwork,wrap_plots)
importFrom(stats,setNames)
importFrom(survival,Surv)
importFrom(survival,survfit)
importFrom(utils,head)
importFrom(utils,modifyList)

# ggstratify

<a href='https://github.com/AkiShiroshita/ggstratify'>
  <img src='man/figures/logo.png'
       alt="Logo of ggstratify"
       align="right"
       height="120" />
</a>
<br clear="right">

**A one-function, point-and-click Shiny interface for the descriptive analysis.**

```r
ggstratify(your_data)
```

That is the only function you need to remember.

![ggstratify demo](man/figures/demo.gif)

## Overview

Descriptive analysis is essential in every research study. Humans (not AI) need to understand the data before making decisions, such as choosing an appropriate statistical model.

In particular, understanding how variables are distributed across strata defined by other variables is often critical.

* Visually inspect your data without repeatedly writing code.

* `ggstratify` runs entirely locally and requires neither a network connection nor a language model.

* Easily export figures to share with collaborators.


### Before you start: decide the variable types

**This package describes what it is given. It does not guess what you meant.**
Convert each column to the type you intend before handing it over.

```r
dat <- transform(
  dat,
  sex       = factor(sex, levels = c("Male", "Female")),   # groups as factors
  severity  = factor(severity, levels = c("Mild", "Moderate", "Severe")),
  age       = as.numeric(age)                              # measurements as numeric
)
ggstratify(dat)
```

A grouping variable left as `1, 2, 3` will be described as a number. The
**order of a factor's levels** becomes the order of the panels, the figures and
the axis.

## Installation

```r
# install.packages("remotes")
remotes::install_github("AkiShiroshita/ggstratify")
```

## Usage

```r
library(ggstratify)

ggstratify(epi_cohort)           # the example data that ships with the package
ggstratify(iris)                 # data.frame / tibble / data.table / matrix
```

## The screen

| Tab | Contents |
|---|---|
| **Plot** | Either every figure at once as panels (fastest) or one at a time, enlarged. **The variables you chose decide which** (see below) |
| **Data** | The first 200 rows |
| **Strata** | Every stratum: variable, level, N and the file name it will be written to. Strata with no file name get no figure. Plus the rows excluded for missing values |
| **R-code** | The `ggplot2` code for the figure on screen. One button copies it |

## Figure types

`Boxplot` / `Density` / `Dot + Error` / `Dotplot` / `Histogram` /
`Kaplan-Meier curve` / `Line` / `Scatter` / `Violin`

## Acknowledgements

- Claude Code (Anthropic's Claude Opus 5) assisted with adding notes, testing and
  English-language proofreading. The design, decisions and final
  responsibility remain the author's.
- The package design was inspired by
  [ggplotgui](https://github.com/gertstulp/ggplotgui/) (Gert Stulp, GPL-3).

## License

GPL-3

# Data handling helpers. Everything here is data.table so that the app stays
# responsive on large data sets; the results that are expensive to compute
# (variable classification, stratum counts) are computed once and cached by the
# server rather than recomputed on every input change.

#' Coerce an input into a standalone data.table
#'
#' Accepts anything that inherits from data.frame (tibble and data.table
#' included) or a matrix. Subclasses are dropped: the result is always a fresh
#' plain `data.table` that shares no columns with the caller's object, so the
#' app can add helper columns by reference without mutating the user's data.
#'
#' A file path is deliberately not accepted. Reading a file here would set
#' every column's type by guess, and the app would then describe the data on
#' those guesses -- see the *Before you start* section of [ggstratify()].
#'
#' @param x A data.frame-like object or a matrix.
#' @return A `data.table`.
#' @keywords internal
#' @noRd
gs_prepare_data <- function(x) {
  if (is.matrix(x)) {
    x <- as.data.frame(x, stringsAsFactors = FALSE)
  }
  if (!is.data.frame(x)) {
    stop("`dataset` must be a data.frame or a matrix, already read into the ",
         "session with the column types you mean it to have.", call. = FALSE)
  }
  # A list-column is a column of objects rather than of values. There is no
  # level to stratify on and no value to plot, and left alone it surfaces much
  # later as a data.table error from inside gs_classify_vars(). Naming it here
  # is the same bargain as the type check above: one clear message at the
  # call, rather than a puzzle from inside the app.
  listy <- names(x)[!vapply(x, is.atomic, logical(1L))]
  if (length(listy)) {
    stop("These columns hold lists rather than values, and cannot be ",
         "described: ", paste(listy, collapse = ", "),
         ". Drop them, or unnest them into columns of values, before ",
         "calling ggstratify().", call. = FALSE)
  }
  # copy() first: as.data.table()/setDT() on a data.table would otherwise let
  # later `:=` calls write straight into the caller's object.
  dt <- data.table::as.data.table(data.table::copy(x))
  gs_clean_names(dt)
}

#' Drop empty/duplicated column names so that generated code is always valid
#' @keywords internal
#' @noRd
gs_clean_names <- function(dt) {
  nms <- names(dt)
  blank <- is.na(nms) | !nzchar(nms)
  if (any(blank)) nms[blank] <- paste0("V", seq_along(nms))[blank]
  # Reserved names used by the preview machinery and by the Kaplan-Meier
  # fit, which would otherwise be overwritten by a column of the same name.
  reserved <- c(".strat_var", ".strat_level", ".strat_label", ".gs_stratum",
                GS_FACET_COL, GS_KM_COLS)
  nms[nms %in% reserved] <- paste0(nms[nms %in% reserved], "_")
  # Last, and after the two renames above rather than before: filling a blank
  # name and escaping a reserved one can each land on a name already in use --
  # data holding both `.strat_var` and `.strat_var_` is the case that bites --
  # and a duplicate name here is one the app would silently read the wrong
  # column through.
  if (anyDuplicated(nms)) nms <- make.unique(nms, sep = "_")
  data.table::setnames(dt, nms)
  dt[]
}

#' Classify every column as continuous / categorical
#'
#' Computed once per data set. `n_levels` is what the UI shows next to each
#' stratifying variable, and what decides whether a variable is a sensible
#' thing to stratify on at all.
#'
#' @param dt A `data.table`.
#' @param max_levels Largest number of distinct values still treated as
#'   categorical for stratification purposes.
#' @param min_continuous Smallest number of distinct numeric values before a
#'   numeric column counts as continuous rather than a coded category.
#' @return A `data.table` with one row per column.
#' @keywords internal
#' @noRd
gs_classify_vars <- function(dt, max_levels = 50L, min_continuous = 10L) {
  nms <- names(dt)
  if (!length(nms)) {
    return(data.table::data.table(
      var = character(), class = character(), n_levels = integer(),
      n_missing = integer(), is_continuous = logical(),
      is_categorical = logical(), is_numeric = logical(), is_event = logical(),
      can_stratify = logical()
    ))
  }

  info <- data.table::rbindlist(lapply(nms, function(v) {
    col <- dt[[v]]
    cls <- class(col)[1L]
    nlev <- data.table::uniqueN(col, na.rm = TRUE)
    numericish <- is.numeric(col) && !inherits(col, c("Date", "POSIXct"))
    list(
      var = v,
      class = cls,
      n_levels = as.integer(nlev),
      n_missing = as.integer(sum(is.na(col))),
      is_continuous = numericish && nlev >= min_continuous,
      is_categorical = !numericish || nlev < min_continuous,
      # Both are for the Kaplan-Meier selectors: survival::Surv() needs a
      # numeric time and a two-valued event indicator.
      is_numeric = numericish,
      is_event = gs_is_event_col(col)
    )
  }))

  # A variable is worth stratifying on when it has at least two levels but
  # few enough that one figure per level is still meaningful.
  info[, can_stratify := is_categorical & n_levels >= 2L & n_levels <= max_levels]
  info[]
}

#' Can a column be handed to `survival::Surv()` as the event indicator?
#'
#' The app draws one kind of survival curve -- plain right-censored -- and so
#' accepts the codings that mean that: logical, 0/1 and 1/2.
#'
#' A factor is rejected, but not because `Surv()` refuses one. It does not:
#' `Surv(time, factor)` builds a multi-state outcome (`type = "mright"`),
#' which fits and plots perfectly well as something this app has no way to
#' draw. Rejecting it here is what keeps a two-level factor from quietly
#' becoming a different analysis than the one on the screen.
#' @keywords internal
#' @noRd
gs_is_event_col <- function(col) {
  if (is.logical(col)) return(TRUE)
  if (!is.numeric(col) || inherits(col, c("Date", "POSIXct"))) return(FALSE)
  vals <- unique(col[!is.na(col)])
  if (!length(vals) || length(vals) > 2L) return(FALSE)
  all(vals %in% c(0, 1)) || all(vals %in% c(1, 2))
}

#' Names of columns matching a role
#' @keywords internal
#' @noRd
gs_vars_of <- function(info, role = c("continuous", "categorical", "stratify",
                                      "numeric", "event", "all")) {
  role <- match.arg(role)
  if (is.null(info) || !nrow(info)) return(character())
  switch(
    role,
    continuous  = info[is_continuous == TRUE, var],
    categorical = info[is_categorical == TRUE, var],
    stratify    = info[can_stratify == TRUE, var],
    numeric     = info[is_numeric == TRUE, var],
    event       = info[is_event == TRUE, var],
    all         = info$var
  )
}

#' What each variable selector may be set to, for one data set
#'
#' Every list is offered `"(none)"` first, and every selector opens on it. A
#' pre-filled box is a claim the app is in no position to make -- it does not
#' know which column the question is about -- and one that is easy to miss: a
#' figure of whichever column sorted first, drawn before anything was asked
#' for, looks exactly like a figure that was asked for.
#'
#' @param info Output of `gs_classify_vars()`.
#' @return A named list of character vectors, one per `selectInput` id.
#' @keywords internal
#' @noRd
gs_selector_choices <- function(info) {
  all_vars <- gs_vars_of(info, "all")
  continuous <- gs_vars_of(info, "continuous")
  # Fall back to every variable when nothing looks continuous, so the app is
  # still usable on oddly typed data.
  y_choices <- if (length(continuous)) continuous else all_vars

  out <- list(
    yvar = y_choices,
    xvar = all_vars,
    group = all_vars,
    # A subject identifier has as many levels as there are subjects, so it is
    # drawn from every variable rather than from the stratifying pool.
    idvar = all_vars,
    # The facet layer and the separate-figure layers come from the same pool:
    # a variable with few enough levels to be worth splitting on.
    facet = gs_vars_of(info, "stratify"),
    # survival::Surv() needs a numeric time and a two-valued event indicator.
    timevar = gs_vars_of(info, "numeric"),
    eventvar = gs_vars_of(info, "event"),
    # Categorizing a variable that is already categorical would do nothing.
    cut_var = continuous
  )
  lapply(out, function(x) c(GS_NONE, x))
}

#' Choices for the layer selectors, labelled with level counts
#' @keywords internal
#' @noRd
gs_stratify_choices <- function(info) {
  vars <- gs_vars_of(info, "stratify")
  if (!length(vars)) return(character())
  sub <- info[var %in% vars]
  stats::setNames(sub$var, sprintf("%s  (%d)", sub$var, sub$n_levels))
}

#' The levels of a column, in the order they should be reported
#'
#' A factor is enumerated from `levels()`, so a declared level that matches no
#' rows is still visible; anything else from the values actually present.
#' `NA` is never a level: a row that does not say which stratum it belongs to
#' cannot be placed in one.
#' @keywords internal
#' @noRd
gs_levels_of <- function(col) {
  lv <- if (is.factor(col)) levels(col) else sort(unique(col[!is.na(col)]))
  as.character(lv)
}

#' Rows that carry a value for every layer variable
#' @keywords internal
#' @noRd
gs_complete_layers <- function(dt, vars) {
  vars <- intersect(vars, names(dt))
  if (!length(vars)) return(rep(TRUE, nrow(dt)))
  Reduce(`&`, lapply(vars, function(v) !is.na(dt[[v]])))
}

#' How many rows each layer variable is missing
#'
#' Reported rather than silently dropped: "n = 480" means something different
#' when 120 patients had no recorded site.
#'
#' @param dt A `data.table`.
#' @param vars Layer variables, see `gs_layer_vars()`.
#' @return A `data.table` with columns `var` and `n_missing`, only for the
#'   variables that actually have missing values.
#' @keywords internal
#' @noRd
gs_missing_report <- function(dt, vars) {
  vars <- intersect(vars, names(dt))
  out <- data.table::data.table(
    var = vars,
    n_missing = vapply(vars, function(v) sum(is.na(dt[[v]])), integer(1L),
                       USE.NAMES = FALSE)
  )
  out[n_missing > 0L]
}

#' The name of the stratum each row belongs to, in crossed mode
#'
#' `sex_Male__site_Site A`. The same construction is emitted into the
#' generated script, so the app and the script name their files identically.
#' @keywords internal
#' @noRd
gs_stratum_key <- function(dt, strat_vars) {
  parts <- lapply(strat_vars, function(v) paste0(v, "_", as.character(dt[[v]])))
  do.call(paste, c(parts, list(sep = "__")))
}

#' The human-readable label of the stratum each row belongs to
#'
#' `sex: Male | site: Site A`. Used for the Strata tab, the figure titles and
#' the facet strips alike, so that one stratum reads the same everywhere.
#' @keywords internal
#' @noRd
gs_stratum_label <- function(dt, strat_vars) {
  parts <- lapply(strat_vars,
                  function(v) sprintf("%s: %s", v, as.character(dt[[v]])))
  do.call(paste, c(parts, list(sep = " | ")))
}

#' One row per stratum
#'
#' In `"independent"` mode each stratifying variable is handled on its own, so
#' selecting `sex` and `treatment` yields `sex_M`, `sex_F`, `treatment_A`, ...
#' rather than their cross-product. In `"crossed"` mode one stratum is one
#' observed combination of all of them, which is what a third or fourth layer
#' of description means.
#'
#' Rows missing any stratifying variable are excluded before anything is
#' counted; `gs_missing_report()` is what tells the user about them.
#'
#' @param dt A `data.table`.
#' @param strat_vars Character vector of stratifying column names.
#' @param min_n Strata with fewer than this many rows are marked `keep = FALSE`.
#' @param mode `"independent"` or `"crossed"`.
#' @return A `data.table` with columns `var`, `level`, `label`, `file`, `n`,
#'   `keep`, `status`; zero rows when `strat_vars` is empty.
#' @keywords internal
#' @noRd
gs_strata_table <- function(dt, strat_vars, min_n = 0L,
                            mode = c("independent", "crossed")) {
  mode <- match.arg(mode)
  empty <- data.table::data.table(
    var = character(), level = character(), label = character(),
    file = character(), n = integer(), keep = logical(), status = character()
  )
  strat_vars <- intersect(strat_vars, names(dt))
  if (!length(strat_vars)) return(empty)
  min_n <- as.integer(min_n)
  if (length(min_n) != 1L || is.na(min_n)) min_n <- 0L
  min_n <- max(min_n, 0L)

  dt <- dt[gs_complete_layers(dt, strat_vars)]

  out <- if (mode == "crossed") {
    gs_strata_crossed(dt, strat_vars)
  } else {
    gs_strata_independent(dt, strat_vars)
  }

  if (!nrow(out)) return(empty)
  out[, keep := n > 0L & n >= min_n]

  # Ordered so that the n == 0 message wins over the min_n message.
  out[, status := "exported"]
  out[n < min_n, status := sprintf("skipped: n < %d", min_n)]
  out[n == 0L, status := "skipped: n = 0, no figure"]

  data.table::setcolorder(
    out, c("var", "level", "label", "file", "n", "keep", "status"))
  out[]
}

#' One stratum per level of each variable, taken one variable at a time
#' @keywords internal
#' @noRd
gs_strata_independent <- function(dt, strat_vars) {
  out <- data.table::rbindlist(lapply(strat_vars, function(v) {
    all_levels <- gs_levels_of(dt[[v]])

    # .N by a single column is the cheapest way to get the sizes.
    counts <- dt[, .N, by = c(v)]
    data.table::setnames(counts, c("level", "n"))
    counts[, level := as.character(level)]

    n <- counts$n[match(all_levels, counts$level)]
    n[is.na(n)] <- 0L
    data.table::data.table(var = v, level = all_levels, n = as.integer(n))
  }), use.names = TRUE)

  if (!nrow(out)) return(out)
  out[, label := sprintf("%s: %s", var, level)]
  out[, file := gs_safe_name(sprintf("%s_%s", var, level))]
  out[]
}

#' One stratum per observed combination of every variable
#'
#' Only combinations that occur are listed. The full grid is what the
#' package's own documentation warns against -- five levels crossed with five
#' gives twenty-five mostly empty figures -- and how much of it is empty is
#' reported as a count instead, by `attr(x, "n_possible")`.
#' @keywords internal
#' @noRd
gs_strata_crossed <- function(dt, strat_vars) {
  counts <- dt[, .N, by = c(strat_vars)]
  if (!nrow(counts)) return(data.table::data.table())

  # Combinations are listed in the level order of the variables, outermost
  # first, so the list reads like the nesting it describes.
  keys <- lapply(strat_vars, function(v) gs_levels_of(dt[[v]]))
  ord <- do.call(order, lapply(seq_along(strat_vars), function(i) {
    match(as.character(counts[[strat_vars[i]]]), keys[[i]])
  }))
  counts <- counts[ord]

  values <- lapply(strat_vars, function(v) as.character(counts[[v]]))
  out <- data.table::data.table(
    var = paste(strat_vars, collapse = " x "),
    level = do.call(paste, c(values, list(sep = " | "))),
    label = gs_stratum_label(counts, strat_vars),
    file = gs_safe_name(gs_stratum_key(counts, strat_vars)),
    n = as.integer(counts$N)
  )
  # How many combinations the variables could have formed, so that the app can
  # say how much of the cross-product is empty.
  data.table::setattr(out, "n_possible", prod(vapply(keys, length, integer(1L))))
  out[]
}

#' Append the stratum size to a label, e.g. "sex: M (N = 30)"
#'
#' One spelling of the size, used by the figure titles, the facet strips of the
#' all-figures preview and the panel strips of a facetted figure alike, so that
#' a reader meets the same notation wherever a count appears.
#' @keywords internal
#' @noRd
gs_label_n <- function(label, n) sprintf("%s (N = %d)", label, as.integer(n))

#' Make a string safe to use as a file name
#' @keywords internal
#' @noRd
gs_safe_name <- function(x) {
  x <- gsub("[^A-Za-z0-9._-]+", "_", as.character(x))
  x <- gsub("^_+|_+$", "", x)
  x[!nzchar(x)] <- "NA"
  x
}

#' Build the long data used by the all-strata facet preview
#'
#' In independent mode each stratifying variable contributes one copy of the
#' data, labelled `variable: level`, so that a single `facet_wrap()` shows
#' every stratum in one render pass. In crossed mode there is nothing to
#' duplicate: each row belongs to exactly one combination. Only the columns
#' actually used by the plot are carried over, which keeps the duplication
#' cheap even on wide data.
#'
#' @param dt A `data.table`.
#' @param strat_vars Stratifying column names.
#' @param plot_cols Columns referenced by the plot's aesthetics.
#' @param min_n Strata smaller than this are dropped.
#' @param mode `"independent"` or `"crossed"`.
#' @param label_n Whether the strip carries the stratum size, following the
#'   same switch as the figure titles.
#' @return A `data.table` with an extra `.strat_label` factor column.
#' @keywords internal
#' @noRd
gs_long_strata <- function(dt, strat_vars, plot_cols, min_n = 0L,
                           mode = c("independent", "crossed"),
                           label_n = TRUE) {
  mode <- match.arg(mode)
  strat_vars <- intersect(strat_vars, names(dt))
  plot_cols <- intersect(unique(plot_cols), names(dt))
  dt <- dt[gs_complete_layers(dt, strat_vars)]

  # Panels appear in the order the Strata tab lists them. Sorting the labels
  # as text would be wrong for a categorized variable: "[0.5,3.9]" sorts after
  # "(3.9,8.13]".
  panel_order <- gs_strata_table(dt, strat_vars, 0L, mode)$label

  narrow <- function(d) {
    if (length(plot_cols)) d[, c(plot_cols, ".strat_label"), with = FALSE] else d
  }

  long <- if (mode == "crossed") {
    d <- dt[, unique(c(plot_cols, strat_vars)), with = FALSE]
    d[, .strat_label := gs_stratum_label(d, strat_vars)]
    narrow(d)
  } else {
    parts <- lapply(strat_vars, function(v) {
      d <- dt[, unique(c(plot_cols, v)), with = FALSE]
      d[, .strat_label := sprintf("%s: %s", v, as.character(d[[v]]))]
      narrow(d)
    })
    data.table::rbindlist(parts, use.names = TRUE, fill = TRUE)
  }
  if (!nrow(long)) return(long)
  if (min_n > 0L) {
    long <- long[, if (.N >= min_n) .SD, by = .strat_label]
  }
  if (!nrow(long)) return(long)

  # Facet strips carry the stratum size, so the preview answers "how many are
  # in this panel?" without a trip to the Strata tab. Strata with no rows
  # cannot appear here at all, which is why they are reported separately.
  sizes <- long[, .N, by = .strat_label]
  labels <- if (isTRUE(label_n)) {
    gs_label_n(sizes$.strat_label, sizes$N)
  } else {
    sizes$.strat_label
  }
  long[, .strat_label := factor(
    labels[match(.strat_label, sizes$.strat_label)],
    levels = labels[order(match(sizes$.strat_label, panel_order))]
  )]
  long[]
}

#' Subset the data to a single stratum
#'
#' `row` is one row of `gs_strata_table()`, which is what the preview's
#' dropdown selects, so the rows shown are the rows the table counted.
#' @keywords internal
#' @noRd
gs_one_stratum <- function(dt, spec, row) {
  strat <- intersect(spec$strat_vars, names(dt))
  if (!length(strat) || is.null(row) || !nrow(row)) return(dt)

  if (identical(spec$strat_mode, "crossed")) {
    return(dt[gs_stratum_label(dt, strat) == row$label[1L]])
  }
  v <- row$var[1L]
  if (!v %in% names(dt)) return(dt)
  col <- as.character(dt[[v]])
  dt[!is.na(col) & col == as.character(row$level[1L])]
}

#' Split the data into the named list the export iterates over
#'
#' Mirrors the `split()` + `setNames()` construction emitted by
#' `gs_code_script()`, so that the "Export all PNGs" button and the generated
#' script produce byte-for-byte the same set of files. Columns are narrowed
#' before the split, which is what keeps this cheap on wide data.
#'
#' `drop = FALSE` keeps unused factor levels as zero-row elements, so that an
#' empty stratum can be counted and reported rather than disappearing. Those
#' elements are removed here, after the caller has had a chance to see them.
#'
#' With no stratifying variables the result is a single unnamed element, which
#' `gs_file_stem()` turns into a plain prefix.
#'
#' @param dt A `data.table`.
#' @param spec A spec list.
#' @param drop_empty Whether to remove strata that fail the size rules.
#' @return A named list of `data.table`s.
#' @keywords internal
#' @noRd
gs_split_strata <- function(dt, spec, drop_empty = TRUE) {
  cols <- gs_spec_cols(spec)
  strat <- intersect(spec$strat_vars, names(dt))
  narrow <- function(d, extra = character()) {
    keep_cols <- unique(c(cols, extra))
    if (length(keep_cols)) d[, keep_cols, with = FALSE] else d
  }

  # A row that does not say which stratum it belongs to is not put in one.
  dt <- dt[gs_complete_layers(dt, gs_layer_vars(spec))]

  if (!length(strat)) {
    return(stats::setNames(list(narrow(dt)), ""))
  }

  if (identical(spec$strat_mode, "crossed")) {
    d <- narrow(dt, strat)
    d[, .gs_stratum := gs_stratum_key(d, strat)]
    out <- split(d, by = ".gs_stratum", sorted = TRUE, drop = TRUE,
                 keep.by = FALSE)
  } else {
    out <- unlist(lapply(strat, function(v) {
      parts <- split(narrow(dt, v), by = v, sorted = TRUE, drop = FALSE)
      stats::setNames(parts, sprintf("%s_%s", v, names(parts)))
    }), recursive = FALSE)
  }

  if (!drop_empty) return(out)

  min_n <- as.integer(spec$min_n)
  if (is.na(min_n)) min_n <- 0L
  sizes <- vapply(out, nrow, integer(1L))
  # n == 0 is dropped regardless of min_n: there is nothing to draw.
  out[sizes > 0L & sizes >= min_n]
}

#' File-name stems for exported figures
#'
#' Vectorised over `nm`, because the Strata tab lists every name at once and
#' has to list the names the export will actually write.
#'
#' An empty `nm` is the unstratified case -- one figure, drawn from everything
#' -- which has no stratum to be named after and falls back to the prefix.
#' @keywords internal
#' @noRd
gs_file_stem <- function(prefix, nm) {
  safe <- gs_safe_name(nm)
  out <- if (nzchar(prefix)) paste0(prefix, "_", safe) else safe
  out[!nzchar(nm)] <- if (nzchar(prefix)) prefix else "plot"
  out
}

scatter_spec <- gs_spec(plot_type = "Scatter", x = "x", y = "y")
box_spec <- gs_spec(plot_type = "Boxplot", y = "y")

test_that("gs_downsample leaves small data alone", {
  d <- data.table::data.table(x = rnorm(100), y = rnorm(100))
  expect_identical(gs_downsample(d, scatter_spec, TRUE), d)
  expect_identical(gs_downsample(d, box_spec, TRUE), d)
})

# The real caps are in the millions of rows, so the sampling *rules* are tested
# against a small threshold passed in, and the caps themselves are tested as
# the numbers they are. Building a table of the real cap's size to watch it be
# halved would cost more memory than the check machine is promised.
CAP <- 500L

test_that("gs_downsample is a no-op when the user switches it off", {
  d <- data.table::data.table(x = rnorm(CAP + 100L))
  expect_identical(gs_downsample(d, scatter_spec, FALSE, threshold = CAP), d)
})

test_that("the caps are large enough that ordinary data is never sampled", {
  # Sampling is the one place where the preview stops matching the data, so it
  # has to be reached by data that will not draw, not by data that is merely
  # large. A cohort of a million rows draws its own boxplot.
  expect_gte(gs_sample_threshold(scatter_spec), 1e6)
  expect_gte(gs_sample_threshold(box_spec), 1e7)
})

test_that("scatter is capped harder than the other plot types", {
  expect_lt(gs_sample_threshold(scatter_spec), gs_sample_threshold(box_spec))
  # A line plot costs the same as a scatter plot of the same data.
  expect_equal(gs_sample_threshold(gs_spec(plot_type = GS_LINE)),
               gs_sample_threshold(scatter_spec))

  d <- data.table::data.table(x = rnorm(CAP * 2L))
  expect_equal(nrow(gs_downsample(d, scatter_spec, TRUE, threshold = CAP)), CAP)
  # The same data is under the real cap for either type, so it is left whole.
  expect_equal(nrow(gs_downsample(d, scatter_spec, TRUE)), nrow(d))
  expect_equal(nrow(gs_downsample(d, box_spec, TRUE)), nrow(d))
})

test_that("a sampled line plot keeps whole trajectories", {
  # Half a trajectory is a line nobody has, so subjects are sampled, not rows.
  per_id <- 4L
  n_id <- CAP                          # 4x the cap, in rows
  d <- data.table::data.table(
    id = rep(sprintf("P%05d", seq_len(n_id)), each = per_id),
    visit = rep(seq_len(per_id), times = n_id),
    crp = rnorm(n_id * per_id)
  )
  spec <- gs_spec(plot_type = GS_LINE, x = "visit", y = "crp", id = "id")
  out <- gs_downsample(d, spec, TRUE, threshold = CAP)

  expect_lte(nrow(out), CAP)
  expect_gt(nrow(out), 0L)
  # Every ID that survived kept all of its visits.
  expect_true(all(out[, .N, by = id]$N == per_id))
})

test_that("group-wise sampling keeps small strata intact", {
  n <- CAP + 1000L
  d <- data.table::data.table(
    y = rnorm(n),
    .strat_label = c(rep("g: huge", n - 20L), rep("g: tiny", 20L))
  )
  out <- gs_downsample(d, box_spec, TRUE, by = ".strat_label", threshold = CAP)

  expect_lte(nrow(out), CAP)
  # Every stratum survives, and the tiny one is not thinned at all.
  expect_setequal(unique(out$.strat_label), c("g: huge", "g: tiny"))
  expect_equal(nrow(out[.strat_label == "g: tiny"]), 20L)
})

# --- server wiring ------------------------------------------------------------

gs_test_inputs <- function(outdir) {
  list(
    plot_type = "Boxplot", yvar = "Sepal.Length", xvar = GS_NONE,
    group = GS_NONE, strat_vars = "Species", min_n = 10,
    strat_mode = "independent",
    jitter = TRUE, binwidth = NA, alpha = 0.6, bw_adjust = 1,
    idvar = GS_NONE, line_points = FALSE, smooth = FALSE, smooth_se = FALSE,
    smooth_span = 0.75,
    facet = GS_NONE, theme = "theme_bw()",
    palette = "", title = "", lab_x = "", lab_y = "", lab_legend = "",
    xlim_min = NA, xlim_max = NA, ylim_min = NA, ylim_max = NA,
    data_name = "iris", outdir = outdir, prefix = "", width = 7, height = 5,
    dpi = 300, format = "png", preview_mode = "facet", preview_stratum = "1",
    sample_big = TRUE, show_n = TRUE,
    timevar = "", eventvar = "", km_ci = FALSE, km_censor = TRUE,
    km_ylim = TRUE, km_risk = FALSE, cut_var = "Sepal.Length",
    cut_method = "quantile", cut_n = 4, cut_points = "", cut_name = ""
  )
}

test_that("the server renders, lists strata and generates code", {
  outdir <- file.path(tempdir(), "gs-test-server")
  unlink(outdir, recursive = TRUE)

  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(outdir))
    session$elapse(500)

    expect_equal(nrow(dat()), 150L)
    expect_equal(gs_vars_of(info(), "stratify"), "Species")
    expect_length(problems(), 0L)

    expect_equal(nrow(strata()), 3L)
    expect_true(all(strata()$keep))
    expect_true(facet_preview())
    expect_no_error(output$plot)

    # The code is one figure's, and says which one.
    expect_match(output$code, "p <- ggplot(", fixed = TRUE)
    expect_false(grepl("ggsave", output$code, fixed = TRUE))
    expect_match(output$code, "# Figure: Species: setosa", fixed = TRUE)

    # Single-stratum preview.
    session$setInputs(preview_mode = "single", preview_stratum = "2")
    session$elapse(500)
    expect_false(facet_preview())
    expect_no_error(output$plot)
  })
})

test_that("the export button writes one PNG per kept stratum", {
  outdir <- file.path(tempdir(), "gs-test-export")
  unlink(outdir, recursive = TRUE)

  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(outdir))
    session$elapse(500)
    session$setInputs(export = 1)

    expect_equal(
      sort(list.files(outdir, pattern = "\\.png$")),
      c("boxplot_Species_setosa.png", "boxplot_Species_versicolor.png",
        "boxplot_Species_virginica.png")
    )
  })
  unlink(outdir, recursive = TRUE)
})

test_that("the code on the R-code tab runs, and draws the figure shown", {
  # The preview, the export and this code are all the same generated text
  # evaluated against the same rows, so what has to hold is that the text
  # runs on its own -- in a session that has never heard of ggstratify.
  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(preview_mode = "single", preview_stratum = "2")
    session$elapse(500)

    env <- new.env(parent = globalenv())
    assign("iris", iris, envir = env)
    p <- eval(parse(text = output$code), envir = env)

    expect_s3_class(p, "ggplot")
    # The second stratum, as the selector says, and only it.
    expect_equal(nrow(p$data), 50L)
    expect_equal(as.character(unique(p$data$Species)), "versicolor")
  })
})

test_that("the export button writes what the Strata tab says it will", {
  outdir <- file.path(tempdir(), "gs-svg")
  unlink(outdir, recursive = TRUE)

  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(outdir))
    session$setInputs(format = "svg")
    session$elapse(500)
    session$setInputs(export = 1)

    expect_equal(sort(list.files(outdir)),
                 sort(paste0("boxplot_", strata()[keep == TRUE, file], ".svg")))
  })
  expect_length(list.files(outdir, pattern = "\\.svg$"), 3L)
  unlink(outdir, recursive = TRUE)
})

test_that("a facet layer disables the all-figures preview rather than double-faceting", {
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", xvar = "treatment",
                      strat_vars = "sex", facet = "severity")
    session$elapse(500)

    expect_false(facet_preview())
    expect_no_error(output$plot)
    expect_match(output$code, "facet_wrap(~ .facet_label)", fixed = TRUE)
  })
})

test_that("one variable cannot be two layers at once", {
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", strat_vars = c("sex", "severity"),
                      facet = "severity")
    session$elapse(500)

    # Faceting by it wins; splitting the figures by it as well would leave one
    # panel per figure.
    expect_equal(spec_r()$strat_vars, "sex")
    expect_equal(spec_r()$facet, "severity")
    expect_no_error(output$plot)
  })
})

test_that("switching plot type keeps the app in a drawable state", {
  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$elapse(500)

    # The survival curve is exercised separately: iris has nothing to follow up.
    for (type in setdiff(GS_PLOT_TYPES, GS_KM)) {
      # Scatter and Line put a variable on each axis.
      session$setInputs(plot_type = type,
                        xvar = if (type %in% GS_XY_TYPES) "Sepal.Width" else "Species")
      session$elapse(500)
      expect_length(problems(), 0L)
      expect_no_error(output$plot)
    }
  })
})

# --- the preview mode ---------------------------------------------------------

test_that("the chosen variables, not the switch, decide what the preview shows", {
  # The switch itself is only offered when both answers are possible, and the
  # server keeps its value in step with updateRadioButtons(). What is asserted
  # here is the part that survives a stale switch: the mock session never
  # feeds an update back into input$, so preview_mode stays at "facet"
  # throughout and every mode below is the one the variables imply.
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", xvar = "treatment",
                      strat_vars = character(), facet = GS_NONE)
    session$elapse(500)

    # Nothing splits the figures, so there is only one figure to show.
    expect_false(preview_choice())
    expect_false(facet_preview())
    expect_no_error(output$plot)

    # Ticking a variable makes several: show them all at once.
    session$setInputs(strat_vars = "sex")
    session$elapse(500)
    expect_true(preview_choice())
    expect_true(facet_preview())

    # Panelling by a variable spends the panels the all-figures preview needs,
    # so it goes back to one at a time -- without the user touching anything.
    session$setInputs(facet = "severity")
    session$elapse(500)
    expect_false(preview_choice())
    expect_false(facet_preview())
    expect_no_error(output$plot)

    session$setInputs(facet = GS_NONE)
    session$elapse(500)
    expect_true(preview_choice())
    expect_true(facet_preview())
  })
})

test_that("changing how a figure looks does not change which figure is shown", {
  # update*Input() moves a control back to its first choice, and the reactives
  # behind the two preview controls are invalidated by every input in the
  # sidebar -- not only by the ones that decide what there is to preview. So
  # what has to hold is that nothing is pushed back to the browser at all when
  # the appearance changes; the mock session cannot show it, because it never
  # feeds an update into input$, so the two update calls are recorded instead.
  pushed <- character()
  record <- function(session, inputId, ...) {
    pushed <<- c(pushed, inputId)
    invisible(NULL)
  }
  local_mocked_bindings(updateSelectInput = record,
                        updateRadioButtons = record, .package = "shiny")

  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(preview_mode = "single", preview_stratum = "3")
    session$elapse(500)
    expect_equal(current_stratum()$label, "Species: virginica")
    # The figures themselves were pushed once, when the data arrived.
    expect_true("preview_stratum" %in% pushed)

    for (change in list(list(alpha = 0.3), list(theme = "theme_minimal()"),
                        list(palette = "Set1"), list(jitter = FALSE),
                        list(ylim_min = 4.5))) {
      pushed <<- character()
      do.call(session$setInputs, change)
      session$elapse(500)
      expect_equal(pushed, character(), info = names(change))
    }
    expect_equal(current_stratum()$label, "Species: virginica")

    # A variable that does decide what there is to preview still pushes the
    # new list, or the box would go on offering figures that are not drawn.
    pushed <<- character()
    session$setInputs(strat_vars = character())
    session$elapse(500)
    expect_true("preview_stratum" %in% pushed)
  })
})

test_that("a typed axis range reaches the figure and the code beside it", {
  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(ylim_min = 4.5, ylim_max = 7)
    session$elapse(500)

    expect_equal(spec_r()$ylim_min, 4.5)
    expect_match(output$code, "coord_cartesian(ylim = c(4.5, 7))", fixed = TRUE)
    expect_no_error(output$plot)

    # Clearing a box is NA, which is the automatic range it started with.
    session$setInputs(ylim_min = NA, ylim_max = NA)
    session$elapse(500)
    expect_false(grepl("coord_cartesian", output$code, fixed = TRUE))
  })
})

# --- line plots ---------------------------------------------------------------

test_that("the server draws a line plot and writes the ID into the code", {
  # Repeated measurements per patient, which is what a line plot describes.
  visits <- data.table::rbindlist(lapply(1:6, function(v) {
    d <- data.table::as.data.table(epi_cohort)[, .(id, sex, treatment, crp)]
    d[, visit := v]
    d[, crp := crp * (1 - 0.15 * v)]
    d
  }))

  shiny::testServer(gs_server(visits), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(plot_type = GS_LINE, xvar = "visit", yvar = "crp",
                      idvar = "id", group = "treatment", strat_vars = "sex",
                      line_points = TRUE, smooth = TRUE, smooth_span = 0.8,
                      data_name = "visits")
    session$elapse(500)

    expect_length(problems(), 0L)
    expect_equal(spec_r()$id, "id")
    expect_no_error(output$plot)

    expect_match(output$code, "geom_line(aes(group = id), alpha = 0.6)",
                 fixed = TRUE)
    expect_match(output$code, "geom_point(alpha = 0.6)", fixed = TRUE)
    expect_match(output$code, 'geom_smooth(method = "loess", formula = y ~ x, span = 0.8',
                 fixed = TRUE)

    # The ID is a column the figure reads, so it survives the narrowing.
    expect_true("id" %in% gs_spec_cols(spec_r()))

    # Switching away takes the line-only options with it.
    session$setInputs(plot_type = "Boxplot", xvar = "treatment")
    session$elapse(500)
    expect_equal(spec_r()$id, "")
    expect_false(grepl("geom_smooth", output$code, fixed = TRUE))
  })
})

# --- Kaplan-Meier -------------------------------------------------------------

test_that("the server draws and exports stratified survival curves", {
  outdir <- file.path(tempdir(), "gs-test-km")
  unlink(outdir, recursive = TRUE)

  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(outdir))
    session$setInputs(plot_type = GS_KM, timevar = "fu_days",
                      eventvar = "death", group = "treatment",
                      strat_vars = "sex", data_name = "epi_cohort")
    session$elapse(500)

    expect_length(problems(), 0L)
    # The time and event variables replace the x/y selection entirely.
    expect_equal(spec_r()$x, "")
    expect_equal(spec_r()$time, "fu_days")
    expect_no_error(output$plot)

    expect_match(output$code, "km_data <- function", fixed = TRUE)
    expect_match(output$code, 'km_data(d, "fu_days", "death", by = c("treatment"))',
                 fixed = TRUE)

    # The risk table is drawn on request, and reaches the code with the curve.
    session$setInputs(km_risk = TRUE)
    session$elapse(500)
    expect_match(output$code, "km_risk <- function", fixed = TRUE)
    expect_no_error(output$plot)
    session$setInputs(km_risk = FALSE)
    session$elapse(500)

    session$setInputs(preview_mode = "single", preview_stratum = "1")
    session$elapse(500)
    expect_no_error(output$plot)

    session$setInputs(export = 1)
  })

  expect_equal(sort(list.files(outdir, pattern = "\\.png$")),
               c("kaplan_meier_curve_sex_Female.png",
                 "kaplan_meier_curve_sex_Male.png"))
  unlink(outdir, recursive = TRUE)
})

test_that("the app says what a survival curve is missing rather than erroring", {
  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(plot_type = GS_KM, timevar = "Sepal.Length",
                      eventvar = "Species")
    session$elapse(500)

    expect_gt(length(problems()), 0L)
    expect_match(output$code, "^# Cannot generate code yet")
  })
})

# --- categorizing -------------------------------------------------------------

test_that("a categorized variable joins the selectors and can be stratified on", {
  outdir <- file.path(tempdir(), "gs-test-cut")
  unlink(outdir, recursive = TRUE)

  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(outdir))
    session$setInputs(yvar = "los_days", xvar = "sex", strat_vars = character(),
                      data_name = "epi_cohort", cut_var = "age",
                      cut_method = "breaks", cut_points = "65",
                      cut_name = "age65")
    session$elapse(500)
    session$setInputs(add_cut = 1)
    session$elapse(500)

    expect_length(cuts_used(), 1L)
    expect_true("age65" %in% names(dat()))
    expect_true("age65" %in% gs_vars_of(info(), "stratify"))
    # The original data is untouched; only the app's copy grew a column.
    expect_false("age65" %in% names(epi_cohort))

    # The rule reaches the spec, and therefore the generated script.
    expect_equal(spec_r()$cuts[[1]]$new, "age65")
    expect_match(output$code, "dt[, age65 := cut(age, breaks = c(-Inf, 65, Inf))]",
                 fixed = TRUE)

    session$setInputs(strat_vars = "age65")
    session$elapse(500)
    expect_equal(nrow(strata()), 2L)
    expect_no_error(output$plot)
    session$setInputs(export = 1)

    # Removing the rule takes the column and the stratification with it.
    session$setInputs(cut_remove = "1", remove_cut = 1)
    session$elapse(500)
    expect_length(cuts_used(), 0L)
    expect_false("age65" %in% names(dat()))
    expect_length(spec_r()$strat_vars, 0L)
    expect_false(grepl("age65", output$code, fixed = TRUE))
  })

  expect_length(list.files(outdir, pattern = "\\.png$"), 2L)
  unlink(outdir, recursive = TRUE)
})

test_that("an impossible categorization is reported and not added", {
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(cut_var = "age", cut_method = "breaks",
                      cut_points = "sixty five", cut_name = "age65")
    session$elapse(500)
    session$setInputs(add_cut = 1)
    session$elapse(500)

    expect_length(cuts_used(), 0L)
    expect_false("age65" %in% names(dat()))
  })
})

test_that("adding a categorized variable leaves the other selections alone", {
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", xvar = "treatment", group = "sex",
                      strat_vars = "site", cut_var = "age",
                      cut_method = "quantile", cut_n = 3, cut_name = "age_cat")
    session$elapse(500)
    session$setInputs(add_cut = 1)
    session$elapse(500)

    spec <- spec_r()
    expect_equal(spec$y, "los_days")
    expect_equal(spec$x, "treatment")
    expect_equal(spec$group, "sex")
    expect_equal(spec$strat_vars, "site")
  })
})

test_that("the app reports a problem instead of erroring on an unusable spec", {
  shiny::testServer(gs_server(iris), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(plot_type = "Boxplot", yvar = "Species")
    session$elapse(500)

    expect_gt(length(problems()), 0L)
    expect_match(output$code, "^# Cannot generate code yet")
  })
})

# --- empty strata -------------------------------------------------------------

test_that("an n = 0 stratum is listed but never exported", {
  outdir <- file.path(tempdir(), "gs-empty")
  unlink(outdir, recursive = TRUE)

  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(outdir))
    session$setInputs(yvar = "los_days", xvar = "treatment",
                      strat_vars = "site", min_n = 10)
    session$elapse(500)

    st <- strata()
    expect_equal(nrow(st), 4L)                       # Site D is included
    expect_equal(st[level == "Site D", n], 0L)
    expect_false(st[level == "Site D", keep])
    expect_match(st[level == "Site D", status], "n = 0")
    expect_no_error(output$strata_table)
    expect_no_error(output$plot)

    session$setInputs(export = 1)
  })

  files <- list.files(outdir, pattern = "\\.png$")
  expect_length(files, 3L)                           # A, B, C -- not D
  expect_false(any(grepl("Site_D", files)))
  unlink(outdir, recursive = TRUE)
})

test_that("the strata summary counts empty and undersized strata", {
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", strat_vars = c("site", "severity"),
                      min_n = 30)
    session$elapse(500)

    st <- strata()
    expect_equal(sum(st$n == 0L), 1L)                # Site D
    expect_gt(sum(!st$keep & st$n > 0L), 0L)         # Severe
    expect_true(all(st[keep == TRUE, n] >= 30L))
    expect_no_error(output$strata_summary)
  })
})

test_that("the stratum size reaches the figure title and the facet strips", {
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", xvar = "treatment",
                      strat_vars = "sex", min_n = 10, preview_mode = "single",
                      preview_stratum = "1", show_n = TRUE)
    session$elapse(500)
    expect_match(output$code, 'labs(title = "sex_Male (N = ', fixed = TRUE)

    session$setInputs(show_n = FALSE)
    session$elapse(500)
    expect_match(output$code, 'labs(title = "sex_Male")', fixed = TRUE)
  })

  # Facet strips carry the size too.
  dt <- gs_prepare_data(epi_cohort)
  long <- gs_long_strata(dt, "sex", c("treatment", "los_days"), 10L)
  expect_true(all(grepl("\\(N = \\d+\\)$", levels(long$.strat_label))))

  # ...and follow the same switch the title does.
  plain <- gs_long_strata(dt, "sex", c("treatment", "los_days"), 10L,
                          label_n = FALSE)
  expect_false(any(grepl("N = ", levels(plain$.strat_label), fixed = TRUE)))
})

# --- the layers ---------------------------------------------------------------

test_that("one layer is a plain figure and no code loops over anything", {
  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", xvar = GS_NONE,
                      strat_vars = character(), facet = GS_NONE,
                      data_name = "epi_cohort")
    session$elapse(500)

    expect_equal(gs_n_layers(spec_r()), 1L)
    expect_equal(nrow(strata()), 0L)
    expect_match(output$code, "d <- epi_cohort", fixed = TRUE)
    expect_false(grepl("facet_", output$code, fixed = TRUE))
    expect_no_error(output$plot)
  })
})

test_that("two layers are one figure of panels, three are separate figures", {
  outdir <- file.path(tempdir(), "gs-test-layers")
  unlink(outdir, recursive = TRUE)

  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(outdir))
    session$setInputs(yvar = "los_days", xvar = "treatment",
                      strat_vars = character(), facet = "severity",
                      data_name = "epi_cohort")
    session$elapse(500)

    expect_equal(gs_n_layers(spec_r()), 2L)
    # The panels carry their sizes, so they are drawn from a labelled copy.
    expect_match(output$code, "facet_wrap(~ .facet_label)", fixed = TRUE)
    expect_match(output$code, 'add_facet_n(d, "severity")', fixed = TRUE)

    # A third layer no longer fits in one figure, so it makes several.
    session$setInputs(strat_vars = "sex")
    session$elapse(500)
    expect_equal(gs_n_layers(spec_r()), 3L)
    expect_match(output$code, 'd <- dt[as.character(sex) == "Male"]',
                 fixed = TRUE)
    session$setInputs(export = 1)
  })

  expect_equal(sort(list.files(outdir, pattern = "\\.png$")),
               c("boxplot_sex_Female.png", "boxplot_sex_Male.png"))
  unlink(outdir, recursive = TRUE)
})

test_that("crossed layers give one figure per observed combination", {
  by_button <- file.path(tempdir(), "gs-crossed-button")
  unlink(by_button, recursive = TRUE)

  shiny::testServer(gs_server(epi_cohort), {
    do.call(session$setInputs, gs_test_inputs(by_button))
    session$setInputs(yvar = "los_days", xvar = "treatment",
                      strat_vars = c("sex", "site"), strat_mode = "crossed",
                      min_n = 0, data_name = "epi_cohort")
    session$elapse(500)

    st <- strata()
    # Site D recruited nobody, so six of the eight combinations occur.
    expect_equal(nrow(st), 6L)
    expect_equal(attr(st, "n_possible"), 8L)
    expect_true(all(grepl("^sex: (Male|Female) \\| site: Site [ABC]$", st$label)))
    expect_true(all(grepl("^sex_(Male|Female)__site_Site_[ABC]$", st$file)))
    expect_equal(sum(st$n), nrow(epi_cohort))
    expect_no_error(output$plot)

    session$setInputs(export = 1)

    # The code for one of those figures subsets on both variables at once.
    expect_match(output$code, "as.character(sex) == ", fixed = TRUE)
    expect_match(output$code, "as.character(site) == ", fixed = TRUE)
    env <- new.env(parent = globalenv())
    assign("epi_cohort", epi_cohort, envir = env)
    expect_s3_class(eval(parse(text = output$code), envir = env), "ggplot")
  })

  expect_length(list.files(by_button, pattern = "\\.png$"), 6L)
  unlink(by_button, recursive = TRUE)
})

test_that("rows missing a layer variable are excluded and counted", {
  cohort <- data.table::as.data.table(epi_cohort)
  cohort[1:40, sex := NA]
  cohort[1:15, severity := NA]
  n_kept <- nrow(cohort) - 40L

  shiny::testServer(gs_server(cohort), {
    do.call(session$setInputs, gs_test_inputs(tempdir()))
    session$setInputs(yvar = "los_days", strat_vars = "sex",
                      facet = "severity", data_name = "cohort")
    session$elapse(500)

    # The 15 rows missing severity are among the 40 missing sex, so 40 go.
    expect_equal(nrow(layer_data()), n_kept)
    expect_equal(missing_report()$n_missing, c(40L, 15L))
    expect_equal(sum(strata()$n), n_kept)
    expect_false(any(is.na(strata()$level)))
    expect_no_error(output$strata_summary)

    # And the generated script does the same thing, out loud.
    expect_match(output$code, 'layer_vars <- c("sex", "severity")', fixed = TRUE)
    expect_match(output$code, "Excluded ", fixed = TRUE)
  })
})

test_that("the UI builds and carries every control the server reads", {
  # as.character() renders the tags without pulling in htmltools by name.
  ui <- as.character(gs_ui())
  for (id in c("plot_type", "yvar", "xvar", "group", "timevar", "eventvar",
               "km_ci", "km_risk", "idvar", "line_points", "smooth",
               "smooth_span", "smooth_se", "cut_var", "cut_method", "add_cut",
               "facet", "strat_vars", "strat_mode", "min_n", "show_n",
               "export", "format", "outdir", "preview_mode", "strata_table",
               "layer_summary", "xlim_min", "xlim_max", "ylim_min",
               "ylim_max")) {
    expect_true(grepl(id, ui, fixed = TRUE), info = id)
  }
})

test_that("every selector opens on nothing chosen", {
  # A pre-filled box is a claim the app is not in a position to make, and a
  # figure drawn before anything was asked for looks like one that was.
  choices <- gs_selector_choices(gs_classify_vars(gs_prepare_data(epi_cohort)))

  expect_setequal(names(choices),
                  c("yvar", "xvar", "group", "idvar", "facet", "timevar",
                    "eventvar", "cut_var"))
  for (id in names(choices)) {
    expect_equal(choices[[id]][1], GS_NONE, info = id)
  }
  # Each list still offers the variables it is meant to.
  expect_true("los_days" %in% choices$yvar)
  expect_true("fu_days" %in% choices$timevar)
  expect_true("death" %in% choices$eventvar)
  expect_false("death" %in% choices$timevar[-1] && "sex" %in% choices$eventvar)
  expect_true("site" %in% choices$facet)
  expect_false("id" %in% choices$facet)     # too many levels to split on
})

test_that("nothing is drawn until something has been chosen", {
  shiny::testServer(gs_server(epi_cohort), {
    session$setInputs(plot_type = "Boxplot")
    session$elapse(500)

    expect_equal(spec_r()$y, "")
    expect_equal(spec_r()$facet, "")
    expect_gt(length(problems()), 0L)
    expect_match(output$code, "^# Cannot generate code yet")
  })
})

test_that("gs_device refuses a format it has no device for", {
  expect_identical(gs_device("png"), ragg::agg_png)
  expect_true(is.function(gs_device("svg")))
  expect_error(gs_device("pdf"), "Unknown figure format")
})

test_that("the JavaScript the conditionalPanels are built from is escaped", {
  expect_equal(gs_js_str("Boxplot"), "'Boxplot'")
  expect_equal(gs_js_str("Tufte's boxplot"), "'Tufte\\'s boxplot'")
  expect_equal(gs_js_str("a\\b"), "'a\\\\b'")
  expect_equal(gs_js_array(c("a", "b")), "['a','b']")
  # The real constants must survive the round trip unchanged.
  expect_equal(gs_js_type(GS_KM),
               sprintf("input.plot_type == '%s'", GS_KM))
  expect_equal(gs_js_has_facet(), sprintf("input.facet != '%s'", GS_NONE))
})

test_that("gs_prepare_data never modifies the caller's object", {
  df <- data.frame(a = 1:3, b = letters[1:3], stringsAsFactors = FALSE)
  dt <- gs_prepare_data(df)

  expect_true(data.table::is.data.table(dt))
  expect_false(data.table::is.data.table(df))

  # Writing by reference into the result must not reach back into `df`.
  dt[, c := a * 2]
  expect_named(df, c("a", "b"))
})

test_that("gs_prepare_data copies data.tables too", {
  src <- data.table::data.table(a = 1:3)
  out <- gs_prepare_data(src)
  out[, b := 1L]
  expect_named(src, "a")
})

test_that("gs_prepare_data accepts a tibble and drops its subclass", {
  skip_if_not_installed("tibble")
  src <- tibble::as_tibble(data.frame(a = 1:3, b = letters[1:3]))
  out <- gs_prepare_data(src)

  # A plain data.table, not a tbl_df wearing a data.table hat: the generated
  # code is written against data.table alone.
  expect_identical(class(out), c("data.table", "data.frame"))
  out[, a := 99L]
  expect_identical(src$a, 1:3)
})

test_that("gs_prepare_data keeps every row of a subclassed data.frame", {
  # Stands in for a grouped tibble: a grouping carried on the class is not a
  # stratification, so it is dropped and no row is summarised away.
  src <- data.frame(g = c("a", "a", "b"), y = 1:3)
  class(src) <- c("grouped_df", "tbl_df", "tbl", "data.frame")
  out <- gs_prepare_data(src)

  expect_identical(class(out), c("data.table", "data.frame"))
  expect_equal(nrow(out), 3L)
})

test_that("gs_prepare_data rejects non-tabular input", {
  expect_error(gs_prepare_data(1:10), "data.frame")
})

test_that("gs_prepare_data does not read files", {
  # A path is not data. Reading one here would type every column by guess,
  # and the app would then describe the data on those guesses.
  path <- tempfile(fileext = ".csv")
  on.exit(unlink(path), add = TRUE)
  utils::write.csv(iris, path, row.names = FALSE)
  expect_error(gs_prepare_data(path), "data.frame")
})

test_that("gs_clean_names fixes blank, duplicated and reserved names", {
  dt <- data.table::data.table(1:2, 1:2, 1:2)
  data.table::setnames(dt, c("x", "x", ".strat_label"))
  out <- gs_clean_names(dt)
  expect_equal(names(out), c("x", "x_1", ".strat_label_"))
})

test_that("gs_classify_vars separates continuous from categorical", {
  dt <- gs_prepare_data(iris)
  info <- gs_classify_vars(dt)

  expect_true(info[var == "Sepal.Length", is_continuous])
  expect_false(info[var == "Species", is_continuous])
  expect_true(info[var == "Species", can_stratify])
  expect_equal(info[var == "Species", n_levels], 3L)

  expect_equal(gs_vars_of(info, "stratify"), "Species")
  expect_true("Sepal.Width" %in% gs_vars_of(info, "continuous"))
})

test_that("a coded numeric with few levels counts as categorical", {
  dt <- data.table::data.table(grp = rep(1:3, each = 20), y = rnorm(60))
  info <- gs_classify_vars(dt)
  expect_true(info[var == "grp", is_categorical])
  expect_true(info[var == "grp", can_stratify])
  expect_true(info[var == "y", is_continuous])
})

test_that("high-cardinality variables are not offered for stratification", {
  dt <- data.table::data.table(id = as.character(1:200))
  info <- gs_classify_vars(dt)
  expect_true(info[var == "id", is_categorical])
  expect_false(info[var == "id", can_stratify])
})

test_that("gs_stratify_choices labels variables with their level count", {
  info <- gs_classify_vars(gs_prepare_data(iris))
  ch <- gs_stratify_choices(info)
  expect_equal(unname(ch), "Species")
  expect_match(names(ch), "Species\\s+\\(3\\)")
})

test_that("gs_strata_table treats each variable independently", {
  dt <- data.table::data.table(
    sex = rep(c("M", "F"), each = 30),
    arm = rep(c("A", "B", "C"), 20),
    y = rnorm(60)
  )
  st <- gs_strata_table(dt, c("sex", "arm"), min_n = 0L)

  expect_equal(nrow(st), 5L)                     # 2 + 3, not 2 * 3
  expect_setequal(st$label,
                  c("sex: F", "sex: M", "arm: A", "arm: B", "arm: C"))
  expect_equal(st[var == "sex", sum(n)], 60L)
  expect_equal(st[var == "arm", sum(n)], 60L)
  expect_true(all(st$file %in% c("sex_F", "sex_M", "arm_A", "arm_B", "arm_C")))
})

test_that("gs_strata_table flags strata below min_n", {
  dt <- data.table::data.table(g = c(rep("big", 50), "small"), y = rnorm(51))
  st <- gs_strata_table(dt, "g", min_n = 10L)
  expect_equal(st[level == "small", keep], FALSE)
  expect_match(st[level == "small", status], "n < 10")
  expect_equal(st[level == "big", keep], TRUE)
  expect_equal(st[level == "big", status], "exported")
})

test_that("an unused factor level is reported with n = 0", {
  dt <- data.table::data.table(
    g = factor(c("a", "a", "b"), levels = c("a", "b", "empty")),
    y = rnorm(3)
  )
  st <- gs_strata_table(dt, "g", min_n = 0L)

  # The empty level is listed, not dropped.
  expect_equal(nrow(st), 3L)
  expect_equal(st[level == "empty", n], 0L)
  expect_false(st[level == "empty", keep])
  expect_match(st[level == "empty", status], "n = 0")
  # Levels that do have rows are still exported.
  expect_true(all(st[level != "empty", keep]))
})

test_that("n = 0 is skipped even when min_n allows everything", {
  dt <- data.table::data.table(
    g = factor(c("a", "b"), levels = c("a", "b", "empty")),
    y = rnorm(2)
  )
  st <- gs_strata_table(dt, "g", min_n = 0L)
  expect_false(st[level == "empty", keep])
  expect_true(all(st[n > 0L, keep]))
})

test_that("character columns cannot produce an n = 0 stratum", {
  dt <- data.table::data.table(g = c("a", "a", "b"), y = rnorm(3))
  st <- gs_strata_table(dt, "g", min_n = 0L)
  expect_equal(nrow(st), 2L)
  expect_true(all(st$n > 0L))
})

test_that("gs_label_n appends the stratum size", {
  expect_equal(gs_label_n("sex: M", 30), "sex: M (N = 30)")
  expect_equal(gs_label_n("site_D", 0), "site_D (N = 0)")
})

test_that("a cleared minimum n does not poison the size rules", {
  dt <- data.table::data.table(g = c("a", "b"), y = 1:2)
  st <- gs_strata_table(dt, "g", min_n = NA_integer_)
  expect_true(all(st$keep))
})

test_that("gs_strata_table returns an empty table for no stratifying variables", {
  dt <- gs_prepare_data(iris)
  expect_equal(nrow(gs_strata_table(dt, character())), 0L)
  expect_equal(nrow(gs_strata_table(dt, "not_a_column")), 0L)
})

test_that("gs_safe_name sanitises anything usable as a file name", {
  expect_equal(gs_safe_name("sex: M"), "sex_M")
  expect_equal(gs_safe_name("a/b\\c"), "a_b_c")
  expect_equal(gs_safe_name("keep.me-1"), "keep.me-1")
  expect_equal(gs_safe_name("***"), "NA")
})

test_that("gs_long_strata stacks one copy per stratifying variable", {
  dt <- data.table::data.table(
    sex = rep(c("M", "F"), each = 30),
    arm = rep(c("A", "B", "C"), 20),
    y = rnorm(60)
  )
  long <- gs_long_strata(dt, c("sex", "arm"), plot_cols = "y", min_n = 0L)

  expect_equal(nrow(long), 120L)                 # 60 rows per variable
  expect_setequal(names(long), c("y", ".strat_label"))
  expect_equal(nlevels(long$.strat_label), 5L)
})

test_that("the facet panels follow the level order, not the label spelling", {
  # The labels a categorized variable produces sort the wrong way as text:
  # "[0.5,3.9]" comes after "(3.9,8.13]".
  dt <- data.table::data.table(y = rnorm(60))
  dt[, band := cut(seq_len(60), breaks = c(0, 20, 40, 60),
                   labels = c("[0,20]", "(20,40]", "(40,60]"))]
  long <- gs_long_strata(dt, "band", plot_cols = "y", min_n = 0L)

  expect_equal(levels(long$.strat_label),
               c("band: [0,20] (N = 20)", "band: (20,40] (N = 20)",
                 "band: (40,60] (N = 20)"))
})

test_that("gs_long_strata drops small strata and leaves the source untouched", {
  dt <- data.table::data.table(g = c(rep("big", 50), "small"), y = rnorm(51))
  long <- gs_long_strata(dt, "g", plot_cols = "y", min_n = 10L)

  expect_equal(nrow(long), 50L)
  expect_named(dt, c("g", "y"))                  # no .strat_label leaked back
})

test_that("gs_split_strata mirrors the generated script's split", {
  dt <- data.table::data.table(
    sex = rep(c("M", "F"), each = 30),
    arm = rep(c("A", "B", "C"), 20),
    bmi = rnorm(60),
    unused = rnorm(60)
  )
  spec <- gs_spec(plot_type = "Boxplot", x = "arm", y = "bmi",
                  strat_vars = c("sex", "arm"), min_n = 0L)
  out <- gs_split_strata(dt, spec)

  expect_named(out, c("sex_F", "sex_M", "arm_A", "arm_B", "arm_C"))
  expect_equal(vapply(out, nrow, integer(1L)),
               c(sex_F = 30L, sex_M = 30L, arm_A = 20L, arm_B = 20L, arm_C = 20L))
  # Columns are narrowed to what the plot reads, plus the stratifying variable.
  expect_setequal(names(out$sex_M), c("arm", "bmi", "sex"))
})

test_that("gs_split_strata applies min_n", {
  dt <- data.table::data.table(g = c(rep("big", 50), "small"), y = rnorm(51))
  spec <- gs_spec(plot_type = "Boxplot", y = "y", strat_vars = "g", min_n = 10L)
  expect_named(gs_split_strata(dt, spec), "g_big")
})

test_that("gs_split_strata exposes then removes empty factor levels", {
  dt <- data.table::data.table(
    g = factor(c("a", "a", "b"), levels = c("a", "b", "empty")),
    y = rnorm(3)
  )
  spec <- gs_spec(plot_type = "Boxplot", y = "y", strat_vars = "g", min_n = 0L)

  # drop_empty = FALSE is what lets the caller count and report the zero.
  raw <- gs_split_strata(dt, spec, drop_empty = FALSE)
  expect_named(raw, c("g_a", "g_b", "g_empty"))
  expect_equal(nrow(raw$g_empty), 0L)

  # By default the empty stratum never reaches the plotting code.
  expect_named(gs_split_strata(dt, spec), c("g_a", "g_b"))
})

test_that("gs_split_strata returns one unnamed element without stratification", {
  dt <- gs_prepare_data(iris)
  spec <- gs_spec(plot_type = "Boxplot", x = "Species", y = "Sepal.Length")
  out <- gs_split_strata(dt, spec)

  expect_length(out, 1L)
  expect_equal(names(out), "")
  expect_equal(nrow(out[[1]]), 150L)
})

test_that("gs_file_stem combines prefix and stratum name", {
  expect_equal(gs_file_stem("boxplot", "sex_M"), "boxplot_sex_M")
  expect_equal(gs_file_stem("", "sex_M"), "sex_M")
  expect_equal(gs_file_stem("boxplot", ""), "boxplot")
  expect_equal(gs_file_stem("", ""), "plot")
  expect_equal(gs_file_stem("fig", "arm: A/B"), "fig_arm_A_B")
})

test_that("gs_one_stratum subsets to the row of the strata table it is given", {
  dt <- gs_prepare_data(iris)
  spec <- gs_spec(plot_type = "Boxplot", y = "Sepal.Length",
                  strat_vars = "Species")
  st <- gs_strata_table(dt, "Species", 0L)

  out <- gs_one_stratum(dt, spec, st[level == "setosa"])
  expect_equal(nrow(out), 50L)
  expect_equal(unique(as.character(out$Species)), "setosa")

  # With no stratifying variables there is nothing to subset to.
  expect_equal(nrow(gs_one_stratum(dt, gs_spec(y = "Sepal.Length"), st[1])), 150L)
})

# --- crossed layers -----------------------------------------------------------

crossed_dt <- function() {
  data.table::data.table(
    sex = factor(rep(c("M", "F"), each = 30)),
    arm = factor(rep(c("A", "B", "C"), 20), levels = c("A", "B", "C", "D")),
    y = rnorm(60)
  )
}

test_that("crossed mode lists one stratum per observed combination", {
  st <- gs_strata_table(crossed_dt(), c("sex", "arm"), 0L, "crossed")

  # 2 x 3 observed; arm "D" recruited nobody, so 8 were possible.
  expect_equal(nrow(st), 6L)
  expect_equal(attr(st, "n_possible"), 8L)
  expect_equal(sum(st$n), 60L)
  expect_equal(st$var[1], "sex x arm")
  expect_equal(st$label[1], "sex: F | arm: A")
  expect_equal(st$file[1], "sex_F__arm_A")
  # Combinations are listed in level order -- F before M, because that is the
  # factor's own order -- with the outermost variable moving slowest.
  expect_equal(st$level, c("F | A", "F | B", "F | C", "M | A", "M | B", "M | C"))
})

test_that("crossed mode splits and previews the same combinations", {
  dt <- crossed_dt()
  spec <- gs_spec(plot_type = "Boxplot", y = "y",
                  strat_vars = c("sex", "arm"), strat_mode = "crossed",
                  min_n = 0L)

  out <- gs_split_strata(dt, spec)
  expect_length(out, 6L)
  expect_setequal(names(out), gs_strata_table(dt, c("sex", "arm"), 0L,
                                              "crossed")$file)
  expect_equal(sum(vapply(out, nrow, integer(1L))), 60L)

  # The all-figures preview labels each row with its one combination; nothing
  # is duplicated the way independent mode duplicates it.
  long <- gs_long_strata(dt, c("sex", "arm"), "y", 0L, "crossed")
  expect_equal(nrow(long), 60L)
  expect_equal(nlevels(long$.strat_label), 6L)
  expect_match(levels(long$.strat_label)[1], "sex: F | arm: A (N = 10)",
               fixed = TRUE)
})

# --- missing values -----------------------------------------------------------

test_that("a row that does not say which stratum it is in is excluded", {
  dt <- data.table::data.table(
    g = c(rep("a", 20), rep("b", 20), rep(NA_character_, 5)),
    # The one missing f is a row whose g is present, so it is an extra
    # exclusion rather than one already covered by g.
    f = c(NA_character_, rep(c("x", "y"), 22)),
    y = rnorm(45)
  )

  st <- gs_strata_table(dt, "g", 0L)
  expect_equal(nrow(st), 2L)
  expect_equal(sum(st$n), 40L)
  expect_false(any(is.na(st$level)))

  # gs_split_strata excludes the same rows, so no "g_NA" file is ever written.
  spec <- gs_spec(plot_type = "Boxplot", y = "y", strat_vars = "g", min_n = 0L)
  out <- gs_split_strata(dt, spec)
  expect_named(out, c("g_a", "g_b"))

  # The facet layer counts as a layer: its missing rows go too.
  spec2 <- gs_spec(plot_type = "Boxplot", y = "y", strat_vars = "g",
                   facet = "f", min_n = 0L)
  expect_equal(sum(vapply(gs_split_strata(dt, spec2), nrow, integer(1L))), 39L)
})

test_that("gs_missing_report counts only the variables that are missing", {
  dt <- data.table::data.table(a = c(1, NA, 3), b = c("x", "y", "z"),
                               c = c(NA, NA, 1))
  rep <- gs_missing_report(dt, c("a", "b", "c"))
  expect_equal(rep$var, c("a", "c"))
  expect_equal(rep$n_missing, c(1L, 2L))
  expect_equal(nrow(gs_missing_report(dt, "b")), 0L)
  expect_equal(nrow(gs_missing_report(dt, character())), 0L)
})

test_that("escaping a reserved name cannot collide with a real column", {
  # .strat_var is renamed to .strat_var_, which the data already has. Before
  # make.unique() ran last, both columns came out called .strat_var_ and the
  # app read the wrong one.
  d <- data.frame(a = 1:3, b = 4:6)
  names(d) <- c(".strat_var", ".strat_var_")
  out <- gs_prepare_data(d)
  expect_equal(anyDuplicated(names(out)), 0L)
  expect_equal(length(names(out)), 2L)

  # The same for a blank name that fills to one already in use.
  d2 <- data.frame(V1 = 1:3, x = 4:6)
  names(d2) <- c("V1", "")
  out2 <- gs_prepare_data(d2)
  expect_equal(anyDuplicated(names(out2)), 0L)
})

test_that("a list-column is refused at the door, not inside the app", {
  d <- data.frame(g = c("a", "b"), stringsAsFactors = FALSE)
  d$lst <- list(1:2, 3:4)
  expect_error(gs_prepare_data(d), "lst")
  expect_error(gs_prepare_data(d), "lists rather than values")
})

test_that("the shipped example data is what its documentation says it is", {
  expect_equal(dim(epi_cohort), c(600L, 11L))
  # The n = 0 stratum only works if Site D survives as a declared level.
  expect_true("Site D" %in% levels(epi_cohort$site))
  expect_equal(sum(epi_cohort$site == "Site D"), 0L)
  # The documented Severe count, and the fact that the default minimum keeps
  # it: the group is there to try the control on, not to be filtered by it.
  expect_equal(sum(epi_cohort$severity == "Severe"), 20L)
  expect_gt(sum(epi_cohort$severity == "Severe"), gs_spec()$min_n)
  # Follow-up is bounded by the one-year study end, and death is 0/1.
  expect_true(all(epi_cohort$fu_days >= 1 & epi_cohort$fu_days <= 365))
  expect_setequal(unique(epi_cohort$death), c(0L, 1L))
})

# The code generator is the single source of truth for what a figure looks
# like. `gs_code_plot()` emits the ggplot2 expression as text; the preview, the
# "Export all figures" button and the R-code tab all draw their figure by
# evaluating exactly that text. A figure can therefore never drift from the
# code the user is told to run.
#
# The claim is about the figure, and stops where the figure does. Getting to
# `d` -- the categorized columns, the rows dropped for a missing layer value,
# the subset that makes this figure its stratum -- is written here too, by
# gs_code_script(), but the app reaches `d` by calling the helpers rather than
# by evaluating those lines. The two stay in step because they share the one
# implementation underneath: gs_apply_cuts() runs the very lines
# gs_code_cut_line() writes, and gs_split_strata() takes the subset
# gs_code_script() prints. gs_eval_plot() is handed a `d` that is already
# there, and evaluates only the figure.

# --- literal helpers ---------------------------------------------------------

# Reserved words plus the handful of names that would be confusing unquoted.
GS_RESERVED <- c("if", "else", "repeat", "while", "function", "for", "next",
                 "break", "TRUE", "FALSE", "NULL", "Inf", "NaN", "NA",
                 "NA_integer_", "NA_real_", "NA_character_", "in")

#' Quote a column name for use in generated code
#' @keywords internal
#' @noRd
gs_bt <- function(x) {
  ok <- grepl("^[a-zA-Z.][a-zA-Z0-9._]*$", x) &
    !grepl("^\\.[0-9]", x) &
    !x %in% GS_RESERVED
  ifelse(ok, x, paste0("`", gsub("`", "\\\\`", x), "`"))
}

#' Quote a string literal for use in generated code
#' @keywords internal
#' @noRd
gs_dq <- function(x) encodeString(as.character(x), quote = "\"")

#' Format a number without scientific notation
#' @keywords internal
#' @noRd
gs_n <- function(x) {
  if (is.null(x) || length(x) != 1L || is.na(x)) return("NA")
  if (x == round(x) && abs(x) < 1e15) return(format(as.integer(round(x))))
  format(x, scientific = FALSE, trim = TRUE)
}

#' Indent every line of a character vector, leaving blank lines blank
#' @keywords internal
#' @noRd
gs_indent <- function(lines, prefix = "  ") {
  ifelse(nzchar(lines), paste0(prefix, lines), lines)
}

# --- plot semantics ----------------------------------------------------------

#' Which aesthetic a "Group (or colour)" selection maps to for a plot type
#' @keywords internal
#' @noRd
gs_group_aes <- function(plot_type) {
  if (plot_type %in% c("Scatter", GS_LINE, "Dot + Error", GS_KM)) "colour" else "fill"
}

# --- panel strips ------------------------------------------------------------

# The column add_facet_n() writes. Dot-prefixed, and kept out of the data by
# gs_clean_names(), so that it cannot collide with a column of the user's.
GS_FACET_COL <- ".facet_label"

# Emitted once, and only when the panels are asked to carry their size. Written
# out as text rather than exported so that the generated code stands on its own.
GS_FACET_N_HELPER <- c(
  "# A panel that does not say how many observations it holds invites the",
  "# reader to compare shapes drawn from 300 rows and from 8, and a strip",
  "# reading \"[18,53]\" does not say what the range is a range of.",
  "# add_facet_n() writes the variable and the size into the strip label,",
  "# keeping the level order intact so that the panels stay in the order the",
  "# factor declares.",
  "add_facet_n <- function(d, var) {",
  "  d <- data.table::copy(data.table::as.data.table(d))",
  "  v <- as.character(d[[var]])",
  "  lv <- if (is.factor(d[[var]])) levels(d[[var]]) else sort(unique(v))",
  "  sizes <- as.integer(table(factor(v, levels = lv)))",
  "  strips <- sprintf(\"%s: %s (N = %d)\", var, lv, sizes)",
  sprintf("  d[, %s := factor(strips[match(v, lv)], levels = strips)]",
          GS_FACET_COL),
  "  d[]",
  "}",
  ""
)

#' Does this figure label its panels with their size?
#'
#' Only a facet variable of the user's is labelled here. The all-figures
#' preview builds its own `.strat_label`, which already carries the size.
#' @keywords internal
#' @noRd
gs_uses_facet_n <- function(spec, facet_strata = FALSE) {
  !isTRUE(facet_strata) && nzchar(spec$facet) && isTRUE(spec$show_n)
}

#' The column the figure is panelled by, which is not always a real column
#' @keywords internal
#' @noRd
gs_facet_col <- function(spec, facet_strata = FALSE) {
  if (isTRUE(facet_strata)) return(".strat_label")
  if (!nzchar(spec$facet)) return("")
  if (isTRUE(spec$show_n)) GS_FACET_COL else spec$facet
}

# --- categorized variables ---------------------------------------------------

#' The `cut()` expression for one categorization rule
#' @keywords internal
#' @noRd
gs_cut_expr <- function(cut) {
  v <- gs_bt(cut$var)
  switch(
    cut$method,
    quantile = sprintf(paste0("cut(%s, breaks = unique(stats::quantile(%s, ",
                              "probs = seq(0, 1, length.out = %s), na.rm = TRUE)), ",
                              "include.lowest = TRUE)"),
                       v, v, gs_n(cut$n + 1L)),
    equal    = sprintf("cut(%s, breaks = %s)", v, gs_n(cut$n)),
    breaks   = sprintf("cut(%s, breaks = c(-Inf, %s, Inf))", v,
                       paste(gs_n_vec(cut$breaks), collapse = ", ")),
    stop("Unknown cut method: ", cut$method, call. = FALSE)
  )
}

#' The single `data.table` line that adds one derived column
#' @keywords internal
#' @noRd
gs_code_cut_line <- function(cut, data_sym = "dt") {
  sprintf("%s[, %s := %s]", data_sym, gs_bt(cut$new), gs_cut_expr(cut))
}

#' The block that adds every derived column, or `character(0)`
#'
#' Emitted once, straight after the data is loaded, so that a categorized
#' variable can be used anywhere a real column can -- including as a
#' stratifying variable.
#' @keywords internal
#' @noRd
gs_code_cuts <- function(cuts, data_sym = "dt") {
  cuts <- gs_as_cuts(cuts)
  if (!length(cuts)) return(character())
  c("# Categorized continuous variables.",
    vapply(cuts, gs_code_cut_line, character(1L), data_sym = data_sym),
    "")
}

# --- the layers --------------------------------------------------------------

#' The block that excludes rows with no value for a layer variable
#'
#' A row that does not say which panel or which figure it belongs to cannot be
#' drawn in one. Dropping it quietly would change every denominator on the
#' screen without saying so, so the count is reported as the script runs.
#' @keywords internal
#' @noRd
gs_code_layer_na <- function(spec, data_sym = "dt") {
  vars <- gs_layer_vars(spec)
  if (!length(vars)) return(character())
  c("# Rows with no value for a layer variable cannot be placed in a panel or",
    "# a figure, so they are excluded here -- once, before anything is counted.",
    sprintf("layer_vars <- c(%s)", paste(gs_dq(vars), collapse = ", ")),
    "for (v in layer_vars) {",
    sprintf("  n_na <- sum(is.na(%s[[v]]))", data_sym),
    "  if (n_na > 0L) {",
    "    message(\"Excluded \", n_na, \" row(s) with a missing \", v, \".\")",
    "  }",
    sprintf("  %s <- %s[!is.na(%s[[v]])]", data_sym, data_sym, data_sym),
    "}",
    "")
}

# --- Kaplan-Meier ------------------------------------------------------------

# Column names km_data() and km_risk() produce. Dot-prefixed so that they
# cannot collide with a grouping variable the user is fitting by;
# gs_clean_names() keeps the data free of them for the same reason.
GS_KM_COLS <- c(".time", ".surv", ".lower", ".upper", ".ncens", ".nrisk")

# The helper the generated script defines once and the loop then calls. It is
# emitted as text rather than exported so that the script stands on its own:
# running it needs survival and data.table, not ggstratify.
GS_KM_HELPER <- c(
  "# survfit() returns one row per event time. km_data() turns a fit into the",
  "# data frame the step plot is drawn from, one curve per level of `by`, and",
  "# starts every curve at (0, 1). The time and event columns are copied to",
  "# .t and .e first, so that a variable named `time` in the data cannot be",
  "# confused with this function's own argument.",
  "km_data <- function(d, time, event, by = character()) {",
  "  one <- function(tt, ev) {",
  "    f <- survival::survfit(survival::Surv(tt, ev) ~ 1)",
  "    na <- rep(NA_real_, length(f$time))",
  "    list(.time  = c(0, f$time),",
  "         .surv  = c(1, f$surv),",
  "         .lower = c(1, if (is.null(f$lower)) na else f$lower),",
  "         .upper = c(1, if (is.null(f$upper)) na else f$upper),",
  "         .ncens = c(0L, f$n.censor))",
  "  }",
  "  cols <- unique(c(time, event, by))",
  "  d <- data.table::as.data.table(d)[, ..cols]",
  "  .ok <- !is.na(d[[time]]) & !is.na(d[[event]])",
  "  d <- d[.ok]",
  "  d[[\".t\"]] <- as.numeric(d[[time]])",
  "  d[[\".e\"]] <- as.numeric(d[[event]])",
  "  if (length(by)) d[, one(.t, .e), by = by] else d[, one(.t, .e)]",
  "}",
  ""
)

# The second helper, defined only when the number-at-risk table is asked for.
# It answers a different question from km_data(): not "what is the estimate at
# this time" but "how many people were still being followed", which is what
# tells a reader whether the tail of the curve is worth reading at all.
GS_KM_RISK_HELPER <- c(
  "# summary(fit, times =) reports the number still at risk at each requested",
  "# time. extend = TRUE keeps a time beyond the last event in the table, as a",
  "# count of zero, rather than dropping the column out of the figure.",
  "km_risk <- function(d, time, event, times, by = character()) {",
  "  one <- function(tt, ev) {",
  "    f <- survival::survfit(survival::Surv(tt, ev) ~ 1)",
  "    s <- summary(f, times = times, extend = TRUE)",
  "    list(.time = times, .nrisk = as.integer(s$n.risk))",
  "  }",
  "  cols <- unique(c(time, event, by))",
  "  d <- data.table::as.data.table(d)[, ..cols]",
  "  .ok <- !is.na(d[[time]]) & !is.na(d[[event]])",
  "  d <- d[.ok]",
  "  d[[\".t\"]] <- as.numeric(d[[time]])",
  "  d[[\".e\"]] <- as.numeric(d[[event]])",
  "  if (length(by)) d[, one(.t, .e), by = by] else d[, one(.t, .e)]",
  "}",
  ""
)

#' The variables a Kaplan-Meier fit must be computed separately within
#'
#' Anything that splits the figure into curves or panels has to split the fit
#' too: a curve fitted across two facets and then drawn in one of them would
#' be the wrong curve.
#' @keywords internal
#' @noRd
gs_km_by <- function(spec, facet_strata = FALSE) {
  by <- c(spec$group, gs_facet_col(spec, facet_strata))
  unique(by[nzchar(by)])
}

#' Helper definitions a spec's figure needs before it can be drawn
#' @keywords internal
#' @noRd
gs_code_preamble <- function(spec, facet_strata = FALSE) {
  c(if (gs_uses_facet_n(spec, facet_strata)) GS_FACET_N_HELPER,
    if (identical(spec$plot_type, GS_KM)) GS_KM_HELPER,
    if (identical(spec$plot_type, GS_KM) && isTRUE(spec$km_risk)) {
      GS_KM_RISK_HELPER
    })
}

#' Per-figure data preparation, or `character(0)` when there is none
#'
#' The panel labels are written before the survival curves are fitted, so that
#' a panel counts the people in it rather than the rows of the fit, and so that
#' the fit is split by the same column the figure is panelled by.
#' @keywords internal
#' @noRd
gs_code_prep <- function(spec, data_sym = "d", facet_strata = FALSE) {
  lines <- character()
  if (gs_uses_facet_n(spec, facet_strata)) {
    lines <- c(lines, sprintf("%s <- add_facet_n(%s, %s)", data_sym, data_sym,
                              gs_dq(spec$facet)), "")
  }
  if (!identical(spec$plot_type, GS_KM)) return(lines)

  by <- gs_km_by(spec, facet_strata)
  by_arg <- if (length(by)) {
    sprintf(", by = c(%s)", paste(gs_dq(by), collapse = ", "))
  } else {
    ""
  }
  c(lines,
    sprintf("km <- km_data(%s, %s, %s%s)", data_sym, gs_dq(spec$time),
            gs_dq(spec$event), by_arg),
    "")
}

#' The symbol `ggplot()` is handed, which is not the raw data for every type
#' @keywords internal
#' @noRd
gs_plot_sym <- function(spec, data_sym = "d") {
  if (identical(spec$plot_type, GS_KM)) "km" else data_sym
}

#' Preparation and plot together: one complete, evaluable figure
#'
#' What the preview, the in-app export and the R-code tab all evaluate, which
#' is what keeps the figure on screen and the code beside it the same figure.
#' The last expression in the block is the plot.
#' @keywords internal
#' @noRd
gs_code_figure <- function(spec, data_sym = "d", facet_strata = FALSE,
                           title_expr = NULL) {
  c(gs_code_preamble(spec, facet_strata),
    gs_code_prep(spec, data_sym, facet_strata),
    gs_code_plot_block(spec, data_sym, facet_strata, title_expr))
}

#' The plot itself, as one or more statements ending in the figure
#'
#' A plain figure is a single `ggplot()` expression. A survival curve with a
#' number-at-risk table is three: the curve, the table, and the stacking of
#' one on the other.
#'
#' @param sym Symbol to assign the finished figure to, or `NULL` to leave it
#'   as the value of the block.
#' @keywords internal
#' @noRd
gs_code_plot_block <- function(spec, data_sym = "d", facet_strata = FALSE,
                               title_expr = NULL, sym = NULL) {
  assign_to <- function(lines, target) {
    if (is.null(target)) lines else c(paste0(target, " <- ", lines[1]), lines[-1])
  }
  risk <- identical(spec$plot_type, GS_KM) && isTRUE(spec$km_risk)
  plot_lines <- gs_code_plot(spec, data_sym, facet_strata, title_expr,
                             shared_x = risk)
  if (!risk) return(assign_to(plot_lines, sym))

  c(gs_code_risk_times(spec),
    assign_to(plot_lines, "p"),
    "",
    gs_code_risk_table(spec, data_sym, facet_strata),
    "",
    "# guides = \"collect\" moves the legend outside both panels, which is also",
    "# what leaves them the same width: a legend inside the curve alone would",
    "# shift its x axis away from the table's.",
    assign_to(
      c(paste0("patchwork::wrap_plots(p, tbl, ncol = 1, ",
               sprintf("heights = c(%s, 1),", gs_n(GS_RISK_HEIGHT_RATIO))),
        "                      guides = \"collect\")"),
      sym))
}

# How much taller the curve is than the table beneath it.
GS_RISK_HEIGHT_RATIO <- 3

#' The times the number at risk is counted at
#'
#' Read off the curve rather than chosen by the user, and then given to both
#' panels, so that a column of the table sits under the point of the curve it
#' describes. An x range typed into the sidebar is what the curve is being
#' looked at through, so the counts are taken across that range instead.
#' @keywords internal
#' @noRd
gs_code_risk_times <- function(spec) {
  from <- gs_num(spec$xlim_min)
  to <- gs_num(spec$xlim_max)
  c("# The curve and the table share these times, so that every count sits",
    "# under the point of the curve it belongs to.",
    sprintf("risk_min <- %s", if (is.na(from)) "0" else gs_n(from)),
    sprintf("risk_max <- %s",
            if (is.na(to)) "max(km$.time, na.rm = TRUE)" else gs_n(to)),
    "risk_times <- pretty(c(risk_min, risk_max), n = 5)",
    "risk_times <- risk_times[risk_times >= risk_min & risk_times <= risk_max]",
    "")
}

#' The number-at-risk table, as a plot of its own
#'
#' One row per curve, labelled by the same variable that colours the curves,
#' so the table is read against the legend without a key of its own.
#' @keywords internal
#' @noRd
gs_code_risk_table <- function(spec, data_sym = "d", facet_strata = FALSE) {
  by <- gs_km_by(spec, facet_strata)
  by_arg <- if (length(by)) {
    sprintf(", by = c(%s)", paste(gs_dq(by), collapse = ", "))
  } else {
    ""
  }
  y <- if (nzchar(spec$group)) gs_bt(spec$group) else "\"All\""
  lab_x <- if (nzchar(spec$lab_x)) spec$lab_x else "Time"

  lines <- c(
    sprintf("risk <- km_risk(%s, %s, %s, risk_times%s)", data_sym,
            gs_dq(spec$time), gs_dq(spec$event), by_arg),
    sprintf("tbl <- ggplot(risk, aes(x = .time, y = %s)) +", y),
    "  geom_text(aes(label = .nrisk), size = 3.4) +",
    "  scale_x_continuous(limits = c(risk_min, risk_max), breaks = risk_times) +",
    # A discrete axis puts the first level at the bottom, which would have the
    # rows running against the legend they are read beside.
    "  scale_y_discrete(limits = rev) +")

  facet <- gs_code_facet(spec, facet_strata)
  if (length(facet)) lines <- c(lines, gs_indent(facet))

  c(lines,
    sprintf("  labs(x = %s, y = NULL, title = \"Number at risk\") +",
            gs_dq(lab_x)),
    paste0("  ", spec$theme, " +"),
    "  theme(panel.grid = element_blank())")
}

#' The ggplot2 expression for one figure, as lines of code
#'
#' The first line is unindented and every following line carries two spaces,
#' so callers can splice the block into a loop body by prefixing a constant
#' indent.
#'
#' @param spec A spec list, see `gs_spec()`.
#' @param data_sym Symbol holding the data inside the generated code.
#' @param facet_strata When `TRUE`, facet by the preview's `.strat_label`
#'   column instead of the user's facet variables.
#' @param title_expr Raw R code for the plot title. When `NULL` the spec's
#'   fixed title is used.
#' @param shared_x When `TRUE`, the x axis is pinned to the breaks the
#'   number-at-risk table underneath is drawn at.
#' @return A character vector of code lines.
#' @keywords internal
#' @noRd
gs_code_plot <- function(spec, data_sym = "d", facet_strata = FALSE,
                         title_expr = NULL, shared_x = FALSE) {
  type <- spec$plot_type
  grp_aes <- gs_group_aes(type)
  has_group <- nzchar(spec$group)

  # --- aes() ---
  mapping <- character()
  if (type == GS_KM) {
    # The axes come from km_data()'s output, not from the raw columns.
    mapping <- c("x = .time", "y = .surv")
  } else if (type %in% GS_XONLY_TYPES) {
    mapping <- c(mapping, sprintf("x = %s", gs_bt(spec$x)))
  } else {
    # Box/violin/dot plots stay valid without an X-variable: everything is
    # drawn in a single unlabelled column.
    mapping <- c(mapping,
                 sprintf("x = %s", if (nzchar(spec$x)) gs_bt(spec$x) else "\"\""))
    mapping <- c(mapping, sprintf("y = %s", gs_bt(spec$y)))
  }
  if (has_group) {
    mapping <- c(mapping, sprintf("%s = %s", grp_aes, gs_bt(spec$group)))
    # The confidence band is filled by group as well as coloured by it.
    if (type == GS_KM && isTRUE(spec$km_ci)) {
      mapping <- c(mapping, sprintf("fill = %s", gs_bt(spec$group)))
    }
  }

  lines <- sprintf("ggplot(%s, aes(%s)) +", gs_plot_sym(spec, data_sym),
                   paste(mapping, collapse = ", "))

  # --- geoms ---
  lines <- c(lines, gs_indent(gs_code_geoms(spec)))

  # --- palette ---
  scales <- gs_code_scales(spec, grp_aes, has_group)
  if (length(scales)) lines <- c(lines, gs_indent(scales))

  # --- coordinates ---
  if (isTRUE(shared_x)) {
    lines <- c(lines, gs_indent(paste(
      "scale_x_continuous(limits = c(risk_min, risk_max),",
      "breaks = risk_times) +")))
  }
  coord <- gs_code_coord(spec, shared_x)
  if (length(coord)) lines <- c(lines, gs_indent(coord))

  # --- facets ---
  facet <- gs_code_facet(spec, facet_strata)
  if (length(facet)) lines <- c(lines, gs_indent(facet))

  # --- labels ---
  labs <- gs_code_labs(spec, grp_aes, has_group, title_expr, type)
  if (length(labs)) lines <- c(lines, gs_indent(labs))

  # --- theme (always last, always present) ---
  if (isTRUE(shared_x)) {
    # The table underneath carries the axis for both, so the curve drops it
    # rather than printing the same numbers twice, one row apart.
    lines <- c(lines, gs_indent(paste0(spec$theme, " +")),
               gs_indent(c(
                 "theme(axis.title.x = element_blank(),",
                 "      axis.text.x = element_blank(),",
                 "      axis.ticks.x = element_blank())")))
  } else {
    lines <- c(lines, gs_indent(spec$theme))
  }
  lines
}

#' Geom layers for a plot type, each on its own line ending in " +"
#' @keywords internal
#' @noRd
gs_code_geoms <- function(spec) {
  type <- spec$plot_type
  alpha <- gs_n(spec$alpha)
  bw <- spec$binwidth
  # The two types whose layers depend on more than the plot type are built by
  # their own functions; switch() cannot dispatch on a name held in a
  # constant, which is what GS_KM and GS_LINE are.
  out <- if (identical(type, GS_KM)) {
    gs_code_km_geoms(spec, alpha)
  } else if (identical(type, GS_LINE)) {
    gs_code_line_geoms(spec, alpha)
  } else {
    switch(
      type,
      "Boxplot" = sprintf("geom_boxplot(alpha = %s, outlier.shape = %s)",
                          alpha,
                          if (isTRUE(spec$jitter)) "NA" else "19"),
      "Violin" = sprintf("geom_violin(alpha = %s)", alpha),
      "Density" = sprintf("geom_density(adjust = %s, alpha = %s)",
                          gs_n(spec$bw_adjust), alpha),
      "Histogram" = paste0(
        "geom_histogram(",
        if (!is.na(bw)) sprintf("binwidth = %s, ", gs_n(bw)) else "",
        sprintf("alpha = %s, position = \"identity\")", alpha)),
      "Dotplot" = paste0(
        "geom_dotplot(binaxis = \"y\", stackdir = \"center\", ",
        if (!is.na(bw)) sprintf("binwidth = %s, ", gs_n(bw)) else "",
        sprintf("alpha = %s)", alpha)),
      "Scatter" = sprintf("geom_point(alpha = %s)", alpha),
      "Dot + Error" = "stat_summary(fun.data = mean_se, geom = \"pointrange\")",
      stop("Unknown plot type: ", type, call. = FALSE)
    )
  }

  # Jitter is offered for the same three types as in ggplotgui.
  if (isTRUE(spec$jitter) && type %in% c("Boxplot", "Violin", "Dot + Error")) {
    # A fixed, local seed means that changing a cosmetic setting (such as
    # opacity) does not make observations appear to move. position_jitter()
    # restores the caller's random-number state after calculating offsets.
    out <- c(out, paste0(
      "geom_point(position = position_jitter(width = 0.2, height = 0, ",
      "seed = 1), alpha = 0.4)"))
  }
  # The smoother goes on top of whatever it is smoothing.
  out <- c(out, gs_code_smooth(spec))
  paste0(out, " +")
}

#' The layers of a line plot
#'
#' The ID is mapped inside `geom_line()` rather than in the figure's own
#' `aes()` deliberately: it is there to say which points belong to one line,
#' and a smoother that inherited it would be fitted per subject, which is not
#' what a trend line means. Without an ID the points are joined in x order,
#' within each colour group -- the shape a set of already-summarised means
#' over time has.
#' @keywords internal
#' @noRd
gs_code_line_geoms <- function(spec, alpha) {
  out <- if (nzchar(spec$id)) {
    sprintf("geom_line(aes(group = %s), alpha = %s)", gs_bt(spec$id), alpha)
  } else {
    sprintf("geom_line(alpha = %s)", alpha)
  }
  if (isTRUE(spec$line_points)) {
    out <- c(out, sprintf("geom_point(alpha = %s)", alpha))
  }
  out
}

#' The LOWESS smoother layer, or `character(0)`
#'
#' `method = "loess"` is named rather than left to ggplot2, which otherwise
#' chooses between loess and a GAM by the number of observations and says so
#' in a message. Naming it keeps the same smoother on the same data at every
#' size. The formula is spelled out for the same reason: it is the default,
#' but leaving it out has ggplot2 announce it on every draw.
#'
#' The smoother inherits the figure's colour grouping, so a grouped figure
#' gets one smoother per group -- the comparison the grouping was asking for.
#' @keywords internal
#' @noRd
gs_code_smooth <- function(spec) {
  if (!isTRUE(spec$smooth)) return(character())
  sprintf("geom_smooth(method = \"loess\", formula = y ~ x, span = %s, se = %s)",
          gs_n(spec$smooth_span),
          if (isTRUE(spec$smooth_se)) "TRUE" else "FALSE")
}

#' The layers of a Kaplan-Meier curve, in drawing order
#'
#' The band is drawn first so that the step line sits on top of it, and the
#' censoring marks last so that they sit on top of both. The band is a plain
#' ribbon between the survfit confidence limits, which interpolates between
#' event times rather than stepping; at the resolution of a figure the
#' difference is invisible, and it keeps the generated code readable.
#' @keywords internal
#' @noRd
gs_code_km_geoms <- function(spec, alpha) {
  out <- character()
  if (isTRUE(spec$km_ci)) {
    out <- c(out, sprintf(
      "geom_ribbon(aes(ymin = .lower, ymax = .upper), alpha = %s, colour = NA)",
      alpha))
  }
  out <- c(out, "geom_step(linewidth = 0.8)")
  if (isTRUE(spec$km_censor)) {
    out <- c(out, paste0(
      "geom_point(data = function(x) x[x$.ncens > 0, ], ",
      "shape = 3, size = 2, show.legend = FALSE)"))
  }
  out
}

#' The coordinate layer, or character(0) when both axes are left automatic
#'
#' `coord_cartesian()` rather than `xlim()` or a scale's `limits`: it zooms
#' the figure into the range asked for, where a scale limit drops the rows
#' outside it first -- which moves a boxplot's median, rescales a density and
#' recounts a histogram. A range is a question about what to look at, not
#' about which rows the description is of.
#'
#' Either end can be left blank on its own: `NA` there is ggplot2's own "as
#' far as the data goes". A Kaplan-Meier curve's 0-to-1 y axis is a default of
#' the same kind, so a range typed in replaces it.
#' @keywords internal
#' @noRd
gs_code_coord <- function(spec, shared_x = FALSE) {
  # With a number-at-risk table the two panels share one x scale, which is
  # already pinned to the range; see gs_code_risk_times().
  xlim <- if (isTRUE(shared_x)) "" else gs_range_arg(spec$xlim_min, spec$xlim_max)
  ylim <- gs_range_arg(spec$ylim_min, spec$ylim_max)
  if (!nzchar(ylim) && identical(spec$plot_type, GS_KM) && isTRUE(spec$km_ylim)) {
    ylim <- "c(0, 1)"
  }
  parts <- c(if (nzchar(xlim)) sprintf("xlim = %s", xlim),
             if (nzchar(ylim)) sprintf("ylim = %s", ylim))
  if (!length(parts)) return(character())
  sprintf("coord_cartesian(%s) +", paste(parts, collapse = ", "))
}

#' One axis range as a `c(from, to)` literal, or `""` when both ends are blank
#' @keywords internal
#' @noRd
gs_range_arg <- function(from, to) {
  from <- gs_num(from)
  to <- gs_num(to)
  if (is.na(from) && is.na(to)) return("")
  sprintf("c(%s, %s)", gs_n(from), gs_n(to))
}

#' The scale layers, or character(0) when the default palette is in use
#'
#' A palette recolours the groups, so it has nothing to act on until a
#' grouping variable is set; that is why nothing is emitted without one.
#'
#' A continuous grouping variable takes `scale_*_distiller()` rather than
#' `scale_*_brewer()`: the brewer scales are discrete and would abort with
#' "Continuous value supplied to discrete scale" instead of drawing anything.
#'
#' A Kaplan-Meier curve with a confidence band uses two aesthetics for one
#' variable, so the palette has to be applied to both.
#' @keywords internal
#' @noRd
gs_code_scales <- function(spec, grp_aes, has_group) {
  if (!has_group || !nzchar(spec$palette)) return(character())
  kind <- if (isTRUE(spec$group_continuous)) "distiller" else "brewer"
  funs <- paste0(if (grp_aes == "fill") "scale_fill_" else "scale_colour_", kind)
  if (identical(spec$plot_type, GS_KM) && isTRUE(spec$km_ci)) {
    funs <- c(funs, paste0("scale_fill_", kind))
  }
  sprintf("%s(palette = %s) +", funs, gs_dq(spec$palette))
}

#' The facet layer, or character(0) when there is none
#'
#' The column panelled by is not always the column the user chose: with the
#' sizes switched on it is the labelled copy `add_facet_n()` writes, so that a
#' strip reads `site: Site A (N = 303)` rather than `Site A`.
#'
#' A plain column is panelled through `label_both()`, which names the variable
#' on the strip the way the labelled copy already does. A level is not always
#' self-explanatory -- a categorized variable's levels are bare ranges, and a
#' panel headed `[18,53]` does not say what the range is a range of.
#' @keywords internal
#' @noRd
gs_code_facet <- function(spec, facet_strata = FALSE) {
  col <- gs_facet_col(spec, facet_strata)
  if (!nzchar(col)) return(character())
  if (isTRUE(facet_strata)) {
    return("facet_wrap(~ .strat_label, scales = \"free_x\") +")
  }
  if (identical(col, GS_FACET_COL)) {
    return(sprintf("facet_wrap(~ %s) +", gs_bt(col)))
  }
  sprintf("facet_wrap(~ %s, labeller = label_both) +", gs_bt(col))
}

#' The labs() layer, or character(0) when nothing is labelled
#'
#' Axis labels normally come from the column names, which is why nothing is
#' emitted unless the user typed something. A Kaplan-Meier curve is the
#' exception: its axes are `.time` and `.surv`, names that mean nothing to a
#' reader, so it names them itself unless told otherwise.
#' @keywords internal
#' @noRd
gs_code_labs <- function(spec, grp_aes, has_group, title_expr = NULL,
                         type = spec$plot_type) {
  parts <- character()
  if (!is.null(title_expr)) {
    parts <- c(parts, sprintf("title = %s", title_expr))
  } else if (nzchar(spec$title)) {
    parts <- c(parts, sprintf("title = %s", gs_dq(spec$title)))
  }
  lab_x <- if (nzchar(spec$lab_x)) spec$lab_x else if (type == GS_KM) "Time" else ""
  lab_y <- if (nzchar(spec$lab_y)) spec$lab_y
           else if (type == GS_KM) "Survival probability" else ""
  if (nzchar(lab_x)) parts <- c(parts, sprintf("x = %s", gs_dq(lab_x)))
  if (nzchar(lab_y)) parts <- c(parts, sprintf("y = %s", gs_dq(lab_y)))
  if (has_group && nzchar(spec$lab_legend)) {
    # Both aesthetics must be renamed together, or ggplot2 splits one legend
    # into two.
    aes_named <- if (type == GS_KM && isTRUE(spec$km_ci)) c(grp_aes, "fill")
                 else grp_aes
    parts <- c(parts, sprintf("%s = %s", aes_named, gs_dq(spec$lab_legend)))
  }
  if (!length(parts)) return(character())
  sprintf("labs(%s) +", paste(parts, collapse = ", "))
}

# --- full script -------------------------------------------------------------

#' Default file-name prefix for a plot type
#' @keywords internal
#' @noRd
gs_default_prefix <- function(plot_type) {
  p <- tolower(gsub("[^A-Za-z0-9]+", "_", plot_type))
  gsub("^_+|_+$", "", p)
}

#' The prefix a spec's exported files actually start with
#'
#' Read in one place by both the export and the Strata tab's file-name column,
#' so that the name the user is shown is the name that appears on disk.
#' @keywords internal
#' @noRd
gs_export_prefix <- function(spec) {
  if (nzchar(spec$prefix)) spec$prefix else gs_default_prefix(spec$plot_type)
}

#' The copy-pasteable R code shown on the R-code tab
#'
#' What is emitted is the `ggplot2` code for **one** figure -- the figure on
#' the Plot tab -- and nothing else: no export loop, no `ggsave()`, no output
#' folder. Writing the files is the "Export all figures" button's job, and a
#' script that also did it would bury the four lines a reader actually wants to
#' take away, edit and put in their own analysis.
#'
#' Everything the figure genuinely needs is still here, in the order it has to
#' run: the categorized columns, the exclusion of rows that cannot be placed in
#' a panel, the subset that makes this figure the stratum it is, and the
#' helpers a survival curve or a labelled panel strip defines for itself.
#'
#' @param spec A spec list, see `gs_spec()`.
#' @param stratum One row of `gs_strata_table()` -- the figure being shown --
#'   or `NULL` when the layers produce a single figure.
#' @param note An extra comment line for the header, used to say when the Plot
#'   tab is showing every figure at once and this is the code for one of them.
#' @return A single string containing the code.
#' @keywords internal
#' @noRd
gs_code_script <- function(spec, stratum = NULL, note = NULL) {
  strat <- spec$strat_vars[nzchar(spec$strat_vars)]
  lv <- gs_stratum_levels(stratum, strat, spec$strat_mode)
  title <- if (!is.null(stratum) && nrow(stratum)) {
    gs_title_literal(spec, stratum$file[1L], stratum$n[1L])
  } else {
    NULL
  }

  # data.table earns its library() call only when something below is written
  # in it: a categorization, the row exclusion, or the subset for this figure.
  # A plain figure needs none of the three, and is better off as ggplot2 code
  # a reader can lift without taking a dependency with it.
  needs_dt <- length(gs_as_cuts(spec$cuts)) > 0L ||
    length(gs_layer_vars(spec)) > 0L

  header <- c(
    "# ---------------------------------------------------------------------",
    "# Generated by ggstratify",
    "# The ggplot2 code for the figure shown on the Plot tab.",
    if (!is.null(stratum) && nrow(stratum)) {
      sprintf("# Figure: %s (N = %d)", stratum$label[1L], stratum$n[1L])
    },
    if (!is.null(note)) paste0("# ", note),
    "# ---------------------------------------------------------------------",
    if (needs_dt) "library(data.table)",
    "library(ggplot2)",
    ""
  )

  if (!needs_dt) {
    # Nothing to derive, exclude or subset: the figure is drawn from the data
    # as it stands.
    return(paste(c(header,
                   sprintf("d <- %s   # <- your data", spec$data_name),
                   "",
                   gs_code_figure_body(spec, title)),
                 collapse = "\n"))
  }

  header <- c(
    header,
    sprintf("dt <- as.data.table(%s)   # <- your data", spec$data_name),
    "",
    gs_code_cuts(spec$cuts, "dt"),
    gs_code_layer_na(spec, "dt")
  )

  subset <- if (length(lv)) {
    c(sprintf("# The rows this figure is drawn from: %s.",
              paste(sprintf("%s = %s", names(lv), lv),
                    collapse = ", ")),
      sprintf("d <- dt[%s]",
              paste(sprintf("as.character(%s) == %s", gs_bt(names(lv)),
                            gs_dq(lv)),
                    collapse = " & ")),
      "")
  } else {
    c("d <- dt", "")
  }

  paste(c(header, subset, gs_code_figure_body(spec, title)), collapse = "\n")
}

#' The helpers, the preparation and the plot, ending in the figure
#'
#' The part of the generated code that is the same however the data reached
#' `d`: with a subset in front of it, or straight from the user's object.
#' @keywords internal
#' @noRd
gs_code_figure_body <- function(spec, title_expr = NULL) {
  c(gs_code_preamble(spec),
    gs_code_prep(spec, data_sym = "d"),
    gs_code_plot_block(spec, data_sym = "d", title_expr = title_expr,
                       sym = "p"),
    "",
    "p")
}

#' The level of each stratifying variable that one figure stands for
#'
#' `gs_strata_table()` reports a crossed stratum as one row whose `level`
#' joins the levels with `" | "`, which is what the Strata tab shows; the
#' generated subset needs them apart again.
#'
#' @return A named character vector, `variable = level`, or `character(0)`
#'   when the row cannot be taken apart with confidence -- a level containing
#'   the separator itself, say -- in which case no subset is generated rather
#'   than a wrong one.
#' @keywords internal
#' @noRd
gs_stratum_levels <- function(row, strat_vars, mode = "independent") {
  if (is.null(row) || !nrow(row) || !length(strat_vars)) return(character())
  if (!identical(mode, "crossed")) {
    v <- as.character(row$var[1L])
    if (!v %in% strat_vars) return(character())
    return(stats::setNames(as.character(row$level[1L]), v))
  }
  parts <- strsplit(as.character(row$level[1L]), " | ", fixed = TRUE)[[1L]]
  if (length(parts) != length(strat_vars)) return(character())
  stats::setNames(parts, strat_vars)
}

#' The title of one already-drawn figure, as a quoted string literal
#'
#' The preview and the in-app export know the stratum's label and size
#' outright, so they build the title here rather than as an expression. The
#' rule is `gs_title_expr()`'s rule, applied to values instead of code.
#'
#' @param spec A spec list.
#' @param label The stratum's name, used when the user typed no title.
#' @param n The stratum size, or `NULL` to leave the size off.
#' @return A quoted string literal, or `NULL` when there is no title.
#' @keywords internal
#' @noRd
gs_title_literal <- function(spec, label = "", n = NULL) {
  base <- if (nzchar(spec$title)) spec$title else label
  if (!nzchar(base)) return(NULL)
  gs_dq(if (isTRUE(spec$show_n) && !is.null(n)) gs_label_n(base, n) else base)
}

#' Evaluate generated plot code against a data.table
#'
#' Used by both the live preview and the in-app export so that neither can
#' diverge from the code on the R-code tab. The block may contain more than
#' one expression -- a Kaplan-Meier figure defines its helper and fits the
#' curves first -- in which case the value of the last one, the plot, is
#' returned.
#'
#' @param code_lines Output of `gs_code_figure()`.
#' @param data The data to bind to the code's data symbol.
#' @param data_sym The symbol name used when the code was generated.
#' @return A `ggplot` object.
#' @keywords internal
#' @noRd
gs_eval_plot <- function(code_lines, data, data_sym = "d") {
  # The package namespace, not ggplot2's: the generated code is written for a
  # session that has attached both ggplot2 and data.table, and data.table
  # quietly falls back to data.frame semantics when it is called from a
  # namespace that does not import it.
  env <- new.env(parent = asNamespace("ggstratify"))
  assign(data_sym, data, envir = env)
  eval(parse(text = paste(code_lines, collapse = "\n")), envir = env)
}

iris_spec <- function(...) {
  gs_spec(x = "Species", y = "Sepal.Length", strat_vars = "Species", ...)
}

new_tempdir <- function() {
  d <- file.path(tempdir(), paste0("ggstratify-", as.integer(runif(1, 1, 1e9))))
  dir.create(d, showWarnings = FALSE, recursive = TRUE)
  d
}

test_that("every plot type generates parseable, drawable code", {
  # iris plus the two columns a Kaplan-Meier curve needs.
  set.seed(1)
  dt <- gs_prepare_data(cbind(
    iris,
    fu = runif(nrow(iris), 1, 100),
    ev = rbinom(nrow(iris), 1, 0.6)
  ))

  for (type in GS_PLOT_TYPES) {
    spec <- gs_spec(
      plot_type = type, y = "Sepal.Length", strat_vars = "Species",
      binwidth = 0.2, time = "fu", event = "ev",
      # Scatter and Line put a variable on each axis, and a factor on the x
      # axis of a line plot is a line per point.
      x = if (type %in% GS_XY_TYPES) "Petal.Length" else "Species")
    code <- gs_code_figure(spec, "d")

    expect_silent(parse(text = paste(code, collapse = "\n")))

    p <- gs_eval_plot(code, dt)
    expect_s3_class(p, "ggplot")
    # ggplot is lazy; building is what actually exercises the layers.
    expect_no_error(ggplot2::ggplot_build(p))
  }
})

test_that("each plot type maps to the expected geom", {
  expected <- c(
    "Boxplot"     = "geom_boxplot",
    "Density"     = "geom_density",
    "Dot + Error" = "stat_summary",
    "Dotplot"     = "geom_dotplot",
    "Histogram"   = "geom_histogram",
    "Scatter"     = "geom_point",
    "Violin"      = "geom_violin"
  )
  expected[[GS_KM]] <- "geom_step"
  expected[[GS_LINE]] <- "geom_line"
  # Every type the app offers is covered here, so a new one cannot be added
  # without saying what it draws.
  expect_setequal(names(expected), GS_PLOT_TYPES)

  for (type in names(expected)) {
    code <- paste(gs_code_plot(iris_spec(plot_type = type)), collapse = "\n")
    expect_match(code, expected[[type]], fixed = TRUE,
                 info = paste("plot type:", type))
  }
})

test_that("a line plot draws one line per ID, and the ID is a layer mapping", {
  # In the figure's own aes() the ID would be inherited by the smoother,
  # which would then be fitted per subject rather than through them.
  spec <- gs_spec(plot_type = GS_LINE, x = "visit", y = "crp", id = "id",
                  group = "treatment", smooth = TRUE)
  code <- paste(gs_code_plot(spec), collapse = "\n")

  expect_match(code, "geom_line(aes(group = id), alpha = 0.6)", fixed = TRUE)
  expect_match(code, "aes(x = visit, y = crp, colour = treatment)", fixed = TRUE)
  expect_false(grepl("aes(x = visit, y = crp, colour = treatment, group = id)",
                     code, fixed = TRUE))

  # No ID: the points are joined in x order, within each colour group.
  plain <- paste(gs_code_plot(gs_spec(plot_type = GS_LINE, x = "visit",
                                      y = "crp")), collapse = "\n")
  expect_match(plain, "geom_line(alpha = 0.6)", fixed = TRUE)
  expect_false(grepl("group =", plain, fixed = TRUE))
})

test_that("the observations are marked only when asked for", {
  with_points <- paste(gs_code_plot(gs_spec(plot_type = GS_LINE, x = "visit",
                                            y = "crp", line_points = TRUE)),
                       collapse = "\n")
  expect_match(with_points, "geom_point(alpha = 0.6)", fixed = TRUE)

  without <- paste(gs_code_plot(gs_spec(plot_type = GS_LINE, x = "visit",
                                        y = "crp")), collapse = "\n")
  expect_false(grepl("geom_point", without, fixed = TRUE))
})

test_that("the LOWESS smoother is opt-in and names its method", {
  spec <- gs_spec(plot_type = "Scatter", x = "Petal.Length", y = "Sepal.Length")
  expect_false(grepl("geom_smooth", paste(gs_code_plot(spec), collapse = "\n"),
                     fixed = TRUE))

  smoothed <- gs_spec(plot_type = "Scatter", x = "Petal.Length",
                      y = "Sepal.Length", smooth = TRUE, smooth_span = 0.5)
  code <- gs_code_plot(smoothed)
  expect_match(paste(code, collapse = "\n"),
               'geom_smooth(method = "loess", formula = y ~ x, span = 0.5, se = FALSE)',
               fixed = TRUE)
  # The band is the other half of the option.
  expect_match(paste(gs_code_plot(gs_spec(plot_type = "Scatter", x = "Petal.Length",
                                          y = "Sepal.Length", smooth = TRUE,
                                          smooth_se = TRUE)), collapse = "\n"),
               "se = TRUE", fixed = TRUE)

  # And it draws: loess through 150 observations, one fit per group.
  dt <- gs_prepare_data(iris)
  grouped <- gs_spec(plot_type = "Scatter", x = "Petal.Length",
                     y = "Sepal.Length", group = "Species", smooth = TRUE)
  expect_no_error(ggplot2::ggplot_build(
    gs_eval_plot(gs_code_figure(grouped, "d"), dt)))
})

test_that("options belonging to one plot type never reach another", {
  # The controls are hidden when the plot type changes, but their last value
  # stays in the browser and would otherwise arrive in the next spec.
  spec <- gs_spec(plot_type = "Boxplot", y = "Sepal.Length", id = "id",
                  line_points = TRUE, smooth = TRUE)
  expect_equal(spec$id, "")
  expect_false(spec$line_points)
  expect_false(spec$smooth)

  code <- paste(gs_code_plot(spec), collapse = "\n")
  expect_false(grepl("geom_smooth", code, fixed = TRUE))
  expect_false(grepl("group = id", code, fixed = TRUE))

  # A line plot keeps all three, and the ID is read from the data.
  line <- gs_spec(plot_type = GS_LINE, x = "visit", y = "crp", id = "id",
                  line_points = TRUE, smooth = TRUE)
  expect_equal(line$id, "id")
  expect_true("id" %in% gs_spec_cols(line))
})

test_that("density and histogram take their variable from y and draw it on x", {
  spec <- gs_spec(plot_type = "Density", y = "Sepal.Length")
  expect_equal(spec$x, "Sepal.Length")
  expect_equal(spec$y, "")

  code <- paste(gs_code_plot(spec), collapse = "\n")
  expect_match(code, "aes(x = Sepal.Length)", fixed = TRUE)
})

test_that("jitter is honoured only where it applies", {
  box <- paste(gs_code_plot(iris_spec(plot_type = "Boxplot", jitter = TRUE)),
               collapse = "\n")
  expect_match(box, "position_jitter(width = 0.2, height = 0, seed = 1)",
               fixed = TRUE)
  expect_match(box, "geom_point", fixed = TRUE)
  # The outliers are hidden when the points themselves are drawn.
  expect_match(box, "outlier.shape = NA", fixed = TRUE)

  # Jitter is meaningless on a histogram and must not be emitted.
  hist <- paste(gs_code_plot(iris_spec(plot_type = "Histogram", jitter = TRUE)),
                collapse = "\n")
  expect_false(grepl("position_jitter", hist, fixed = TRUE))
})

test_that("the group aesthetic follows the plot type", {
  fill <- paste(gs_code_plot(iris_spec(plot_type = "Boxplot", group = "Species")),
                collapse = "\n")
  expect_match(fill, "fill = Species", fixed = TRUE)

  colour <- paste(gs_code_plot(gs_spec(plot_type = "Scatter", x = "Sepal.Width",
                                       y = "Sepal.Length", group = "Species")),
                  collapse = "\n")
  expect_match(colour, "colour = Species", fixed = TRUE)
})

test_that("a palette is emitted only when a grouping variable is set", {
  with_group <- paste(gs_code_plot(iris_spec(group = "Species", palette = "Set1")),
                      collapse = "\n")
  expect_match(with_group, "scale_fill_brewer(palette = \"Set1\")", fixed = TRUE)

  without <- paste(gs_code_plot(iris_spec(palette = "Set1")), collapse = "\n")
  expect_false(grepl("scale_fill_brewer", without, fixed = TRUE))
})

test_that("a continuous grouping variable takes the gradient palette scale", {
  # scale_fill_brewer() is discrete and would abort on a continuous variable.
  cont <- paste(gs_code_plot(iris_spec(group = "Petal.Width", palette = "Blues",
                                       group_continuous = TRUE)),
                collapse = "\n")
  expect_match(cont, "scale_fill_distiller(palette = \"Blues\")", fixed = TRUE)
  expect_false(grepl("scale_fill_brewer", cont, fixed = TRUE))

  disc <- paste(gs_code_plot(iris_spec(group = "Species", palette = "Blues")),
                collapse = "\n")
  expect_match(disc, "scale_fill_brewer(palette = \"Blues\")", fixed = TRUE)
})

test_that("an axis range is emitted only when one was typed", {
  none <- paste(gs_code_plot(iris_spec()), collapse = "\n")
  expect_false(grepl("coord_cartesian", none, fixed = TRUE))

  both <- paste(gs_code_plot(iris_spec(xlim_min = 0, xlim_max = 10,
                                       ylim_min = 4.5, ylim_max = 7.5)),
                collapse = "\n")
  expect_match(both, "coord_cartesian(xlim = c(0, 10), ylim = c(4.5, 7.5))",
               fixed = TRUE)

  # One axis, and one end of it, can be left to the data on its own.
  half <- paste(gs_code_plot(iris_spec(ylim_max = 7.5)), collapse = "\n")
  expect_match(half, "coord_cartesian(ylim = c(NA, 7.5))", fixed = TRUE)
  expect_false(grepl("xlim", half, fixed = TRUE))
})

test_that("a range zooms the figure rather than dropping the rows outside it", {
  # scale limits would filter first, which moves the median of what is drawn.
  # coord_cartesian() leaves the boxplot alone and looks at part of it.
  dt <- gs_prepare_data(iris)
  spec <- gs_spec(plot_type = "Boxplot", y = "Sepal.Length", ylim_min = 5,
                  ylim_max = 6)
  code <- gs_code_figure(spec, "d")

  expect_silent(parse(text = paste(code, collapse = "\n")))
  built <- ggplot2::ggplot_build(gs_eval_plot(code, dt))
  expect_equal(built$data[[1]]$middle, stats::median(iris$Sepal.Length))
})

test_that("facets come from the strata in facet mode and from the spec otherwise", {
  strat <- paste(gs_code_plot(iris_spec(), facet_strata = TRUE), collapse = "\n")
  expect_match(strat, "facet_wrap(~ .strat_label", fixed = TRUE)

  user <- paste(gs_code_plot(iris_spec(facet = "Species", show_n = FALSE)),
                collapse = "\n")
  expect_match(user, "facet_wrap(~ Species, labeller = label_both)", fixed = TRUE)

  none <- paste(gs_code_plot(iris_spec()), collapse = "\n")
  expect_false(grepl("facet_", none, fixed = TRUE))
})

test_that("each panel says how many observations it holds", {
  spec <- iris_spec(facet = "Species", show_n = TRUE)
  code <- gs_code_figure(spec, "d")
  flat <- paste(code, collapse = "\n")

  # The panel is drawn from a labelled copy of the column, not the column, so
  # that the strip can carry the size without touching the user's data.
  expect_match(flat, "add_facet_n <- function", fixed = TRUE)
  expect_match(flat, 'd <- add_facet_n(d, "Species")', fixed = TRUE)
  expect_match(flat, "facet_wrap(~ .facet_label)", fixed = TRUE)

  # Switching the sizes off goes back to the plain column, panelled through
  # label_both() so that the strip still says which variable it is a level of.
  plain <- paste(gs_code_figure(iris_spec(facet = "Species", show_n = FALSE), "d"),
                 collapse = "\n")
  expect_false(grepl("add_facet_n", plain, fixed = TRUE))
  expect_match(plain, "facet_wrap(~ Species, labeller = label_both)", fixed = TRUE)

  # And the labels are the variable and the sizes: setosa, versicolor and
  # virginica are 50 each.
  p <- gs_eval_plot(code, gs_prepare_data(iris))
  strips <- levels(p$data$.facet_label)
  expect_equal(strips, c("Species: setosa (N = 50)", "Species: versicolor (N = 50)",
                         "Species: virginica (N = 50)"))
})

test_that("a panel strip says which variable its level belongs to", {
  # A categorized variable's levels are bare ranges: a strip reading [4.3,5.4]
  # does not say what the range is a range of, and the axis and the legend are
  # describing other variables.
  dt <- gs_prepare_data(iris)
  cut <- gs_cut("Sepal.Length", new = "sl_cat", method = "breaks", breaks = 5.4)
  spec <- gs_spec(plot_type = "Boxplot", y = "Petal.Length", facet = "sl_cat",
                  cuts = list(cut))

  p <- gs_eval_plot(gs_code_figure(spec, "d"), gs_apply_cuts(dt, list(cut)))
  expect_true(all(grepl("^sl_cat: ", levels(p$data$.facet_label))))
})

test_that("a panel strip counts only the rows the figure draws", {
  # A stratum's panels are counted within the stratum, not across the data:
  # the figure and its strips have to agree about who is in it.
  dt <- gs_prepare_data(data.frame(
    g = rep(c("A", "B"), each = 30L),
    f = rep(c("x", "y", "x"), times = 20L),
    y = rnorm(60)
  ))
  spec <- gs_spec(plot_type = "Boxplot", y = "y", facet = "f",
                  strat_vars = "g", show_n = TRUE)
  p <- gs_eval_plot(gs_code_figure(spec, "d"), dt[g == "A"])

  expect_equal(sum(as.integer(sub(".*N = (\\d+).*", "\\1",
                                  levels(p$data$.facet_label)))), 30L)
})

test_that("boxplots survive having no x-variable", {
  spec <- gs_spec(plot_type = "Boxplot", y = "Sepal.Length")
  code <- gs_code_plot(spec)
  expect_match(paste(code, collapse = "\n"), 'aes(x = "", y = Sepal.Length)',
               fixed = TRUE)
  expect_no_error(ggplot2::ggplot_build(gs_eval_plot(code, gs_prepare_data(iris))))
})

test_that("awkward column names are backtick-quoted", {
  expect_equal(gs_bt("age"), "age")
  expect_equal(gs_bt("Sepal.Length"), "Sepal.Length")
  expect_equal(gs_bt("my var"), "`my var`")
  expect_equal(gs_bt("2010"), "`2010`")
  expect_equal(gs_bt("if"), "`if`")
  # Non-ASCII column names. Written as escapes so the source stays ASCII.
  jp <- "\u5e74\u9f62"
  expect_equal(gs_bt(jp), paste0("`", jp, "`"))

  spec <- gs_spec(plot_type = "Boxplot", x = "my group", y = jp)
  code <- paste(gs_code_plot(spec), collapse = "\n")
  expect_match(code, "x = `my group`", fixed = TRUE)
  expect_silent(parse(text = code))
})

test_that("labels are quoted safely", {
  spec <- iris_spec(title = 'He said "hi"', lab_x = "Species", lab_y = "Length")
  code <- paste(gs_code_plot(spec), collapse = "\n")
  expect_silent(parse(text = code))
  expect_match(code, "labs(", fixed = TRUE)
})

test_that("title_expr injects raw code rather than a string literal", {
  code <- paste(gs_code_plot(iris_spec(), title_expr = "nm"), collapse = "\n")
  expect_match(code, "labs(title = nm)", fixed = TRUE)
})

test_that("the generated code is one figure's worth, and nothing else", {
  spec <- gs_spec(plot_type = "Boxplot", x = "arm", y = "bmi",
                  strat_vars = c("sex", "treatment"), min_n = 10L,
                  data_name = "mydata", outdir = "figures")
  script <- gs_code_script(spec)

  expect_silent(parse(text = script))
  # Writing the files is the Export button's job; the code tab is the figure.
  for (unwanted in c("iwalk", "ggsave", "dir.create", "outdir", "split(")) {
    expect_false(grepl(unwanted, script, fixed = TRUE), info = unwanted)
  }
  expect_match(script, "library(ggplot2)", fixed = TRUE)
  expect_match(script, "p <- ggplot(", fixed = TRUE)
})

test_that("the code subsets the data down to the figure being shown", {
  spec <- gs_spec(plot_type = "Boxplot", y = "bmi", strat_vars = "sex",
                  show_n = TRUE)
  row <- data.table::data.table(var = "sex", level = "Male", label = "sex: Male",
                                file = "sex_Male", n = 42L, keep = TRUE)
  script <- gs_code_script(spec, row)

  expect_silent(parse(text = script))
  expect_match(script, 'd <- dt[as.character(sex) == "Male"]', fixed = TRUE)
  expect_match(script, "# Figure: sex: Male (N = 42)", fixed = TRUE)
  expect_match(script, 'labs(title = "sex_Male (N = 42)")', fixed = TRUE)

  # Nothing to subset: the figure is drawn from every row. With nothing to
  # derive or exclude either, it is drawn from the object as it stands, and
  # the code takes no data.table dependency to say so.
  plain <- gs_code_script(gs_spec(plot_type = "Scatter", x = "a", y = "b",
                                  data_name = "mydata"))
  expect_silent(parse(text = plain))
  expect_match(plain, "d <- mydata", fixed = TRUE)
  expect_false(grepl("library(data.table)", plain, fixed = TRUE))
  expect_false(grepl("Figure:", plain, fixed = TRUE))

  # A layer brings data.table back, because the exclusion is written in it.
  layered <- gs_code_script(gs_spec(plot_type = "Scatter", x = "a", y = "b",
                                    facet = "g"))
  expect_match(layered, "library(data.table)", fixed = TRUE)
  expect_match(layered, "d <- dt", fixed = TRUE)
})

test_that("a crossed stratum is taken apart into one test per variable", {
  spec <- gs_spec(plot_type = "Boxplot", y = "bmi",
                  strat_vars = c("sex", "site"), strat_mode = "crossed")
  row <- data.table::data.table(var = "sex x site", level = "Male | Site A",
                                label = "sex: Male | site: Site A",
                                file = "sex_Male__site_Site_A", n = 30L,
                                keep = TRUE)
  script <- gs_code_script(spec, row)

  expect_silent(parse(text = script))
  expect_match(script,
               paste0('d <- dt[as.character(sex) == "Male" & ',
                      'as.character(site) == "Site A"]'),
               fixed = TRUE)
})

test_that("a stratum that cannot be taken apart yields no subset at all", {
  # A level containing the separator itself would otherwise be split into the
  # wrong number of pieces, and a wrong subset is worse than none.
  spec <- gs_spec(plot_type = "Boxplot", y = "bmi",
                  strat_vars = c("sex", "site"), strat_mode = "crossed")
  row <- data.table::data.table(var = "sex x site", level = "Male",
                                label = "x", file = "x", n = 1L, keep = TRUE)
  expect_length(gs_stratum_levels(row, c("sex", "site"), "crossed"), 0L)
  expect_match(gs_code_script(spec, row), "d <- dt", fixed = TRUE)
})

test_that("the stratum size is written into the figure title", {
  row <- data.table::data.table(var = "sex", level = "M", label = "sex: M",
                                file = "sex_M", n = 42L, keep = TRUE)
  with_n <- gs_code_script(gs_spec(plot_type = "Boxplot", y = "bmi",
                                   strat_vars = "sex", show_n = TRUE), row)
  expect_match(with_n, 'labs(title = "sex_M (N = 42)")', fixed = TRUE)
  expect_silent(parse(text = with_n))

  without <- gs_code_script(gs_spec(plot_type = "Boxplot", y = "bmi",
                                    strat_vars = "sex", show_n = FALSE), row)
  expect_match(without, 'labs(title = "sex_M")', fixed = TRUE)
  expect_silent(parse(text = without))
})

test_that("a title the user typed replaces the stratum name", {
  # "blank = stratum name" is what the control promises, so a typed title has
  # to win.
  spec <- gs_spec(plot_type = "Boxplot", y = "bmi", strat_vars = "sex",
                  title = "Length of stay", show_n = TRUE)
  expect_equal(gs_title_literal(spec, "sex_M", 42L),
               "\"Length of stay (N = 42)\"")
  expect_equal(gs_title_literal(gs_spec(title = "", show_n = TRUE), "sex_M", 42L),
               "\"sex_M (N = 42)\"")
  expect_equal(gs_title_literal(gs_spec(title = "Fixed", show_n = FALSE),
                                "sex_M", 42L),
               "\"Fixed\"")
  expect_null(gs_title_literal(gs_spec(), "", NULL))
})

test_that("the generated code draws the figure it says it draws", {
  dir <- new_tempdir()
  spec <- gs_spec(plot_type = "Boxplot", y = "Sepal.Length", facet = "Species",
                  data_name = "iris")
  env <- new.env(parent = globalenv())
  p <- eval(parse(text = gs_code_script(spec)), envir = env)

  expect_s3_class(p, "ggplot")
  ggplot2::ggsave(file.path(dir, "p.png"), p, width = 5, height = 4,
                  device = ragg::agg_png)
  expect_gt(file.size(file.path(dir, "p.png")), 0)
})

test_that("each format is written through a device that can write it", {
  expect_identical(gs_device("png"), ragg::agg_png)
  # svglite when it is there, and the cairo device R ships with when it is not,
  # so that the option works without a further install either way.
  expect_true(is.function(gs_device("svg")))

  dir <- new_tempdir()
  p <- gs_eval_plot(gs_code_figure(gs_spec(plot_type = "Boxplot",
                                           y = "Sepal.Length"), "d"),
                    gs_prepare_data(iris))
  for (fmt in GS_FORMATS) {
    f <- file.path(dir, paste0("p.", fmt))
    ggplot2::ggsave(f, p, width = 5, height = 4, device = gs_device(fmt))
    expect_gt(file.size(f), 0)
  }
  # An SVG is text, and says so in its first line.
  expect_match(readLines(file.path(dir, "p.svg"), n = 2L, warn = FALSE)[1],
               "xml|svg")
})

test_that("an unknown format falls back to PNG rather than to nothing", {
  expect_equal(gs_spec(format = "tiff")$format, "png")
  expect_equal(gs_spec(format = "svg")$format, "svg")
})

test_that("gs_default_prefix produces file-safe stems", {
  expect_equal(gs_default_prefix("Boxplot"), "boxplot")
  expect_equal(gs_default_prefix("Dot + Error"), "dot_error")
})

test_that("gs_n formats numbers without scientific notation", {
  expect_equal(gs_n(300), "300")
  expect_equal(gs_n(0.6), "0.6")
  expect_equal(gs_n(1e6), "1000000")
  expect_equal(gs_n(NA), "NA")
})

km_spec <- function(...) {
  gs_spec(plot_type = GS_KM, time = "fu_days", event = "death", ...)
}

cohort <- function() gs_prepare_data(epi_cohort)

test_that("the Kaplan-Meier spec drops any lingering x and y selection", {
  spec <- km_spec(x = "sex", y = "bmi")
  expect_equal(spec$x, "")
  expect_equal(spec$y, "")
  # The variables it does read are the ones it is narrowed to.
  expect_setequal(gs_spec_cols(km_spec(group = "sex")),
                  c("fu_days", "death", "sex"))
})

test_that("a curve is drawable, with and without a grouping variable", {
  dt <- cohort()
  for (spec in list(km_spec(),
                    km_spec(group = "treatment"),
                    km_spec(group = "treatment", km_ci = TRUE, palette = "Set1"),
                    km_spec(km_censor = FALSE, km_ylim = FALSE))) {
    code <- gs_code_figure(spec, "d")
    expect_silent(parse(text = paste(code, collapse = "\n")))
    expect_no_error(ggplot2::ggplot_build(gs_eval_plot(code, dt)))
  }
})

test_that("the fit is computed within everything that splits the figure", {
  # A curve fitted across panels and then drawn in one of them would be the
  # wrong curve, so grouping and faceting both enter the `by`.
  expect_equal(gs_km_by(km_spec()), character())
  expect_equal(gs_km_by(km_spec(group = "sex")), "sex")
  # With the sizes on the strips, the fit is split by the labelled copy of
  # the facet column -- the same column the panels come from.
  expect_equal(gs_km_by(km_spec(group = "sex", facet = "site")),
               c("sex", ".facet_label"))
  expect_equal(gs_km_by(km_spec(group = "sex", facet = "site", show_n = FALSE)),
               c("sex", "site"))
  expect_equal(gs_km_by(km_spec(group = "sex"), facet_strata = TRUE),
               c("sex", ".strat_label"))

  prep <- gs_code_prep(km_spec(group = "sex"), "d")
  expect_match(prep[1], 'km_data(d, "fu_days", "death", by = c("sex"))',
               fixed = TRUE)
  expect_match(gs_code_prep(km_spec(), "d")[1],
               'km_data(d, "fu_days", "death")', fixed = TRUE)
})

test_that("the all-strata preview fits one curve per panel", {
  dt <- cohort()
  spec <- km_spec(group = "treatment", strat_vars = c("sex", "site"))
  long <- gs_long_strata(dt, spec$strat_vars, gs_spec_cols(spec), 10L)

  code <- gs_code_figure(spec, "d", facet_strata = TRUE)
  expect_match(paste(code, collapse = "\n"), '"treatment", ".strat_label"',
               fixed = TRUE)
  expect_no_error(ggplot2::ggplot_build(gs_eval_plot(code, long)))
})

test_that("the confidence band and the censoring marks are opt-in layers", {
  plain <- paste(gs_code_plot(km_spec()), collapse = "\n")
  expect_false(grepl("geom_ribbon", plain, fixed = TRUE))
  expect_match(plain, "geom_point(data = function(x) x[x$.ncens > 0, ]",
               fixed = TRUE)

  ci <- paste(gs_code_plot(km_spec(km_ci = TRUE, km_censor = FALSE)),
              collapse = "\n")
  expect_match(ci, "geom_ribbon(aes(ymin = .lower, ymax = .upper)", fixed = TRUE)
  expect_false(grepl("geom_point", ci, fixed = TRUE))
})

test_that("a grouped band colours both aesthetics from one palette", {
  code <- paste(gs_code_plot(km_spec(group = "sex", km_ci = TRUE,
                                     palette = "Set1", lab_legend = "Sex")),
                collapse = "\n")
  expect_match(code, "colour = sex, fill = sex", fixed = TRUE)
  expect_match(code, "scale_colour_brewer(palette = \"Set1\")", fixed = TRUE)
  expect_match(code, "scale_fill_brewer(palette = \"Set1\")", fixed = TRUE)
  # One renamed legend, not one renamed and one not.
  expect_match(code, 'colour = "Sex", fill = "Sex"', fixed = TRUE)
})

test_that("the axes are named, because .time and .surv are not names", {
  code <- paste(gs_code_plot(km_spec()), collapse = "\n")
  expect_match(code, 'labs(x = "Time", y = "Survival probability")', fixed = TRUE)

  # ...unless the user named them.
  mine <- paste(gs_code_plot(km_spec(lab_x = "Days", lab_y = "Alive")),
                collapse = "\n")
  expect_match(mine, 'labs(x = "Days", y = "Alive")', fixed = TRUE)
})

test_that("the y axis is clamped to 0-1 unless that is switched off", {
  expect_match(paste(gs_code_plot(km_spec()), collapse = "\n"),
               "coord_cartesian(ylim = c(0, 1))", fixed = TRUE)
  expect_false(grepl("coord_cartesian",
                     paste(gs_code_plot(km_spec(km_ylim = FALSE)), collapse = "\n"),
                     fixed = TRUE))

  # 0 to 1 is a default, and a range typed into the sidebar replaces it.
  typed <- paste(gs_code_plot(km_spec(ylim_min = 0.5)), collapse = "\n")
  expect_match(typed, "coord_cartesian(ylim = c(0.5, NA))", fixed = TRUE)
  expect_false(grepl("ylim = c(0, 1)", typed, fixed = TRUE))
})

test_that("the risk table is counted across the range the curve is shown over", {
  code <- paste(gs_code_figure(km_spec(km_risk = TRUE, xlim_max = 60), "d"),
                collapse = "\n")
  expect_match(code, "risk_max <- 60", fixed = TRUE)
  expect_match(code, "risk_min <- 0", fixed = TRUE)
  # The shared scale already carries the range, so the curve does not repeat
  # it as a coord_cartesian() the table would not have.
  expect_false(grepl("xlim = ", code, fixed = TRUE))

  auto <- paste(gs_code_figure(km_spec(km_risk = TRUE), "d"), collapse = "\n")
  expect_match(auto, "risk_max <- max(km$.time, na.rm = TRUE)", fixed = TRUE)
})

test_that("a curve is never drawn from a subsample", {
  # Sampling a distribution keeps its shape; sampling a survival curve makes
  # a different, wider curve.
  expect_equal(gs_sample_threshold(km_spec()), Inf)
  # An infinite cap is not reached by any table, so a small one shows it: the
  # threshold gs_downsample() takes by default is the spec's, and it is Inf.
  n <- 1000L
  d <- data.table::data.table(t = runif(n), e = rbinom(n, 1, 0.5))
  expect_identical(gs_downsample(d, km_spec(), TRUE), d)
  expect_lt(nrow(gs_downsample(d, km_spec(), TRUE, threshold = 100L)), n)
})

test_that("validation reports what a curve is missing", {
  info <- gs_classify_vars(cohort())
  expect_match(gs_validate_spec(gs_spec(plot_type = GS_KM)),
               "needs a time variable", all = FALSE)
  expect_match(gs_validate_spec(gs_spec(plot_type = GS_KM,
                                        time = "fu_days")),
               "needs an event variable")
  expect_length(gs_validate_spec(km_spec(), info), 0L)

  # Surv() takes 0/1, 1/2 or logical, and nothing else.
  bad <- gs_validate_spec(gs_spec(plot_type = GS_KM, time = "fu_days",
                                  event = "severity"), info)
  expect_match(bad, "must be 0/1, 1/2 or TRUE/FALSE")

  bad_time <- gs_validate_spec(gs_spec(plot_type = GS_KM, time = "sex",
                                       event = "death"), info)
  expect_match(bad_time, "time variable must be numeric")
})

test_that("gs_is_event_col accepts exactly what Surv() accepts", {
  expect_true(gs_is_event_col(c(0L, 1L, 1L)))
  expect_true(gs_is_event_col(c(1, 2, 2)))
  expect_true(gs_is_event_col(c(TRUE, FALSE, NA)))
  expect_true(gs_is_event_col(c(0, 0, 0)))
  expect_false(gs_is_event_col(c(0, 1, 2)))
  expect_false(gs_is_event_col(factor(c("yes", "no"))))
  expect_false(gs_is_event_col(c("0", "1")))
})

test_that("a column called time is not confused with the helper's argument", {
  dt <- cohort()
  data.table::setnames(dt, "fu_days", "time")
  code <- gs_code_figure(gs_spec(plot_type = GS_KM, time = "time",
                                 event = "death", group = "sex"), "d")
  expect_no_error(ggplot2::ggplot_build(gs_eval_plot(code, dt)))
})

test_that("the columns the fit produces cannot be shadowed by the data", {
  dt <- data.table::data.table(.time = 1:3, .surv = 1:3, x = 1:3)
  expect_equal(names(gs_clean_names(dt)), c(".time_", ".surv_", "x"))
})

test_that("a stratum in which nobody had the event still draws", {
  dt <- cohort()
  dt[sex == "Male", death := 0L]
  code <- gs_code_figure(km_spec(group = "sex", km_ci = TRUE), "d")
  expect_no_error(ggplot2::ggplot_build(gs_eval_plot(code, dt)))
})

test_that("the generated code carries the helper it needs and runs", {
  spec <- km_spec(group = "treatment", data_name = "epi_cohort")
  script <- gs_code_script(spec)

  expect_silent(parse(text = script))
  expect_equal(length(gregexpr("km_data <- function", script)[[1]]), 1L)
  expect_match(script, "survival::survfit", fixed = TRUE)
  # It stands on its own: it calls survival and data.table, never ggstratify.
  expect_false(grepl("ggstratify::", script, fixed = TRUE))

  env <- new.env(parent = globalenv())
  p <- eval(parse(text = script), envir = env)
  expect_s3_class(p, "ggplot")
})

# --- the number-at-risk table -------------------------------------------------

test_that("the risk table is opt-in, and only for a survival curve", {
  plain <- paste(gs_code_figure(km_spec(), "d"), collapse = "\n")
  expect_false(grepl("km_risk", plain, fixed = TRUE))

  # The control belongs to one plot type, so it cannot reach another.
  box <- gs_spec(plot_type = "Boxplot", y = "bmi", km_risk = TRUE)
  expect_false(box$km_risk)
  expect_false(grepl("km_risk", paste(gs_code_figure(box, "d"), collapse = "\n"),
                     fixed = TRUE))
})

test_that("the risk table is counted at the times the curve is labelled at", {
  code <- paste(gs_code_figure(km_spec(group = "treatment", km_risk = TRUE), "d"),
                collapse = "\n")

  expect_match(code, "km_risk <- function", fixed = TRUE)
  expect_match(code, "risk <- km_risk(d, \"fu_days\", \"death\", risk_times",
               fixed = TRUE)
  # One set of breaks, given to both panels: a count under the wrong point of
  # the curve is worse than no count at all.
  expect_equal(
    length(gregexpr(
      "scale_x_continuous(limits = c(risk_min, risk_max), breaks = risk_times)",
      code, fixed = TRUE)[[1]]), 2L)
  # The curve drops the axis the table underneath carries for both.
  expect_match(code, "axis.text.x = element_blank()", fixed = TRUE)
  expect_match(code, "patchwork::wrap_plots(p, tbl", fixed = TRUE)
})

test_that("the risk table draws, grouped, ungrouped and panelled", {
  dt <- cohort()
  for (spec in list(
    km_spec(km_risk = TRUE),
    km_spec(km_risk = TRUE, group = "treatment"),
    km_spec(km_risk = TRUE, group = "treatment", facet = "sex"),
    km_spec(km_risk = TRUE, km_ci = TRUE, group = "sex", palette = "Set1")
  )) {
    code <- gs_code_figure(spec, "d")
    expect_silent(parse(text = paste(code, collapse = "\n")))
    p <- gs_eval_plot(code, dt)
    expect_s3_class(p, "patchwork")
    # Building the assembly is what actually exercises both halves of it.
    expect_no_error(patchwork::patchworkGrob(p))
  }
})

test_that("the counts in the table are the numbers still being followed", {
  dt <- cohort()
  code <- gs_code_figure(km_spec(km_risk = TRUE), "d")
  env <- new.env(parent = asNamespace("ggstratify"))
  assign("d", dt, envir = env)
  eval(parse(text = paste(code, collapse = "\n")), envir = env)
  risk <- get("risk", envir = env)

  # Everyone is at risk at time zero, and nobody is added later.
  expect_equal(risk$.nrisk[risk$.time == 0], nrow(dt))
  expect_true(all(diff(risk$.nrisk) <= 0))
})

# Server logic. The expensive things -- reading the data and classifying its
# columns -- happen once per data set and are stored in reactiveVal()s. Every
# other reactive is cheap, and the preview render is cached on the spec so that
# flipping back to a previous setting is instant.

#' Build the ggstratify server function
#'
#' The data set is fixed for the life of the app: it is the object the launcher
#' was called with, and there is no way to load another one from inside. That
#' is deliberate -- see the *Before you start* section of [ggstratify()] -- and
#' it is what lets `raw` be an ordinary value rather than a reactive.
#'
#' @param dataset The data supplied to the launcher.
#' @param data_name What the generated code should call it.
#' @keywords internal
#' @noRd
gs_server <- function(dataset, data_name = "mydata") {
  # Coerced once, outside the session, so that every session of a hosted app
  # shares the work rather than repeating it.
  raw <- gs_prepare_data(dataset)

  function(input, output, session) {

    # Categorization rules, in the order the user added them.
    cuts <- shiny::reactiveVal(list())

    # The name the data was passed under, rather than the placeholder the
    # control starts with, so the code on the R-code tab runs as it is read.
    if (!identical(data_name, "mydata")) {
      shiny::updateTextInput(session, "data_name", value = data_name)
    }

    # Rules that can actually be applied to the data in hand.
    cuts_used <- shiny::reactive(gs_valid_cuts(cuts(), names(raw)))

    # What the whole app works on: the data plus the categorized columns. They
    # are built by evaluating the very lines the R-code tab shows, so the app
    # and the script cannot disagree about them.
    dat <- shiny::reactive(gs_apply_cuts(raw, cuts_used()))

    # Re-derived whenever a categorized column appears or disappears.
    info <- shiny::reactive(gs_classify_vars(dat()))

    # --- populate the selectors once per data set ----------------------------

    # Adding a categorized variable repopulates every selector, so a selector
    # keeps what the user picked whenever that choice still exists. Only a
    # selection that has become impossible falls back to "(none)", which is
    # also where every selector starts -- see gs_selector_choices().
    keep_or <- function(id, choices) {
      cur <- input[[id]]
      if (!is.null(cur) && length(cur) == 1L && cur %in% choices) cur else GS_NONE
    }

    shiny::observeEvent(info(), {
      nfo <- info()
      choices <- gs_selector_choices(nfo)
      for (id in names(choices)) {
        shiny::updateSelectInput(session, id, choices = choices[[id]],
                                 selected = keep_or(id, choices[[id]]))
      }
      shiny::updateCheckboxGroupInput(
        session, "strat_vars", choices = gs_stratify_choices(nfo),
        selected = intersect(input$strat_vars %||% character(),
                             gs_vars_of(nfo, "stratify")))
    })

    # --- categorizing --------------------------------------------------------

    # A name that does not collide with the data, refreshed as the source
    # variable changes but never overwriting something the user typed.
    shiny::observeEvent(list(input$cut_var, cuts_used()), {
      dt <- dat()
      shiny::req(input$cut_var)
      # Nothing chosen yet: there is no name to suggest.
      if (identical(input$cut_var, GS_NONE)) return()
      suggested <- gs_cut_name(input$cut_var, names(dt))
      current <- input$cut_name %||% ""
      if (!nzchar(current) || current %in% names(dt) ||
          grepl("_cat[0-9]*$", current)) {
        shiny::updateTextInput(session, "cut_name", value = suggested)
      }
    })

    shiny::observeEvent(input$add_cut, {
      dt <- dat()
      shiny::req(input$cut_var)
      if (identical(input$cut_var, GS_NONE)) {
        shiny::showNotification("Choose the continuous variable to categorize.",
                                type = "warning")
        return()
      }

      breaks <- gs_parse_breaks(input$cut_points)
      if (identical(input$cut_method, "breaks") && anyNA(breaks)) {
        shiny::showNotification(
          "Cut points must be numbers, separated by commas.",
          type = "error", duration = 8)
        return()
      }

      cut <- gs_cut(var = input$cut_var, new = trimws(input$cut_name %||% ""),
                    method = input$cut_method %||% "quantile",
                    n = input$cut_n, breaks = breaks)
      probs <- gs_check_cut(dt, cut)
      if (length(probs)) {
        shiny::showNotification(paste(probs, collapse = " "), type = "error",
                                duration = 10)
        return()
      }

      cuts(c(cuts(), list(cut)))
      shiny::showNotification(sprintf("Added %s", gs_cut_describe(cut)),
                              type = "message", duration = 5)
    })

    shiny::observeEvent(input$remove_cut, {
      cs <- cuts()
      idx <- suppressWarnings(as.integer(input$cut_remove))
      if (is.na(idx) || idx < 1L || idx > length(cs)) return()
      cuts(cs[-idx])
    })

    output$cuts_list <- shiny::renderUI({
      cs <- cuts_used()
      if (!length(cs)) {
        return(shiny::helpText("No categorized variables yet."))
      }
      labels <- vapply(cs, gs_cut_describe, character(1L))
      shiny::div(
        class = "mt-3",
        shiny::selectInput("cut_remove", "Categorized variables",
                           choices = stats::setNames(seq_along(cs), labels)),
        shiny::actionButton("remove_cut", "Remove", class = "btn-outline-danger btn-sm",
                            icon = shiny::icon("trash"))
      )
    })

    # --- spec ---------------------------------------------------------------

    spec_r <- shiny::debounce(
      shiny::reactive({
        spec <- gs_spec_from_input(input)
        # The categorization rules are held outside the inputs, because they
        # accumulate rather than being read off a control.
        spec$cuts <- cuts_used()
        # Which palette scale the group needs is a property of the data, not
        # of any control, so it is filled in here rather than read off one.
        nfo <- info()
        spec$group_continuous <- nzchar(spec$group) && !is.null(nfo) &&
          spec$group %in% gs_vars_of(nfo, "continuous")
        # Removing a categorization removes its column, and the browser takes
        # a round trip to notice. Until it does, the ticked variable would
        # otherwise reach the generated script, which would then not run.
        dt <- dat()
        spec$strat_vars <- intersect(spec$strat_vars, names(dt))
        if (!spec$facet %in% names(dt)) spec$facet <- ""
        gs_normalize_spec(spec)
      }),
      millis = 300
    )

    problems <- shiny::reactive(gs_validate_spec(spec_r(), info()))

    # --- the layers ----------------------------------------------------------

    # Every figure is drawn from rows that say which figure they belong to.
    # Excluding them here, once, is what keeps the n on the Strata tab, the n
    # in the title and the rows in the figure the same number.
    layer_data <- shiny::reactive({
      dt <- dat()
      dt[gs_complete_layers(dt, gs_layer_vars(spec_r()))]
    })

    missing_report <- shiny::reactive({
      gs_missing_report(dat(), gs_layer_vars(spec_r()))
    })

    output$layer_summary <- shiny::renderUI({
      spec <- spec_r()
      n <- gs_n_layers(spec)
      described <- if (identical(spec$plot_type, GS_KM)) {
        if (nzchar(spec$time)) spec$time else "the outcome"
      } else if (nzchar(spec$y)) spec$y else if (nzchar(spec$x)) spec$x else "nothing yet"
      parts <- sprintf("%d layer%s: describing %s", n, if (n == 1L) "" else "s",
                       described)
      if (nzchar(spec$facet)) {
        parts <- c(parts, sprintf("panelled by %s", spec$facet))
      }
      st <- if (length(spec$strat_vars)) strata() else NULL
      if (!is.null(st) && nrow(st)) {
        parts <- c(parts, sprintf("%d figure%s from %s", sum(st$keep),
                                  if (sum(st$keep) == 1L) "" else "s",
                                  paste(spec$strat_vars, collapse =
                                          if (identical(spec$strat_mode, "crossed"))
                                            " x " else ", ")))
      }
      shiny::div(class = "alert alert-light border py-2 px-3 small mt-2",
                 paste(parts, collapse = ", "))
    })

    # --- strata -------------------------------------------------------------

    strata <- shiny::reactive({
      spec <- spec_r()
      gs_strata_table(layer_data(), spec$strat_vars, spec$min_n, spec$strat_mode)
    })

    # updateSelectInput() moves the box back to its first choice, and strata()
    # is invalidated by every control in the sidebar -- an opacity, a theme --
    # not only by the ones that decide which figures there are. Pushing the
    # list unconditionally would therefore send the preview back to the first
    # figure whenever anything at all was adjusted. So it is pushed only when
    # it is a different list, and the figure on screen is followed to its new
    # position rather than dropped: it is the same figure, further down.
    shown_strata <- NULL
    shiny::observeEvent(strata(), {
      st <- strata()[keep == TRUE]
      choices <- if (nrow(st)) {
        stats::setNames(as.character(seq_len(nrow(st))), st$label)
      } else {
        character()
      }
      if (identical(choices, shown_strata)) return()
      shown <- names(shown_strata)[match(input$preview_stratum, shown_strata)]
      shown_strata <<- choices
      keep_showing <- length(shown) == 1L && !is.na(shown) &&
        shown %in% names(choices)
      shiny::updateSelectInput(
        session, "preview_stratum", choices = choices,
        selected = if (keep_showing) choices[[shown]] else NULL)
    })

    # The one figure the app is currently talking about: the one the preview
    # draws when the figures are shown one at a time, and the one the R-code
    # tab writes the code for. `NULL` when the layers make a single figure.
    current_stratum <- shiny::reactive({
      spec <- spec_r()
      if (!length(spec$strat_vars)) return(NULL)
      st <- strata()[keep == TRUE]
      if (!nrow(st)) return(NULL)
      idx <- suppressWarnings(as.integer(input$preview_stratum))
      if (is.na(idx) || idx < 1L || idx > nrow(st)) idx <- 1L
      st[idx]
    })

    # --- preview ------------------------------------------------------------

    # Showing every figure at once borrows the facet machinery, so both modes
    # are possible only when the figures are split and no facet variable has
    # claimed the panels already.
    preview_choice <- shiny::reactive({
      spec <- spec_r()
      length(spec$strat_vars) > 0 && !nzchar(spec$facet)
    })

    # The chosen variables decide what the preview can show, so they set it:
    # every figure at once as soon as that becomes possible, one at a time as
    # soon as it stops being. The switch is hidden in the second case, which
    # is why this needs no note on screen. A choice the user makes while both
    # modes are open is left alone, which is what the comparison is for:
    # observeEvent() fires whenever its expression is invalidated, and
    # preview_choice() reads the spec, so without it every change of opacity
    # would push the preview back to showing all the figures at once.
    shown_choice <- NULL
    shiny::observeEvent(preview_choice(), {
      if (identical(preview_choice(), shown_choice)) return()
      shown_choice <<- preview_choice()
      shiny::updateRadioButtons(
        session, "preview_mode",
        selected = if (preview_choice()) "facet" else "single")
    })

    facet_preview <- shiny::reactive({
      identical(input$preview_mode, "facet") && preview_choice()
    })

    output$preview_note <- shiny::renderUI({
      spec <- spec_r()
      thr <- gs_sample_threshold(spec)
      if (!isTRUE(input$sample_big) || nrow(dat()) <= thr) return(NULL)
      msg <- sprintf(
        paste("Preview drawn from a random sample of about %s rows%s.",
              "The exported figures and the generated code always use every",
              "row%s."),
        format(thr, big.mark = ","),
        # Trajectories are sampled whole: half a line is a different line.
        if (nzchar(spec$id)) sprintf(", taken one whole %s at a time", spec$id) else "",
        # The strips are counted from what is drawn, which here is the sample.
        # The Strata tab and the exported figures still count every row.
        if (gs_uses_facet_n(spec)) {
          ", including the N on each panel strip"
        } else {
          ""
        }
      )
      shiny::div(class = "alert alert-warning py-2 px-3 small", msg)
    })

    output$plot <- shiny::renderPlot({
      dt <- layer_data()
      probs <- problems()
      shiny::validate(shiny::need(!length(probs), paste(probs, collapse = "\n")))

      spec <- spec_r()
      cols <- gs_spec_cols(spec)

      if (facet_preview()) {
        long <- gs_long_strata(dt, spec$strat_vars, cols, spec$min_n,
                               spec$strat_mode, label_n = spec$show_n)
        shiny::validate(shiny::need(
          nrow(long) > 0,
          "No stratum reaches the minimum n. Lower it, or pick another variable."
        ))
        long <- gs_downsample(long, spec, isTRUE(input$sample_big),
                              by = ".strat_label")
        gs_eval_plot(gs_code_figure(spec, "d", facet_strata = TRUE), long)
      } else {
        row <- current_stratum()
        d <- dt
        title <- NULL
        if (!is.null(row)) {
          d <- gs_one_stratum(dt, spec, row)
          # n comes from the strata table, so it reports the true stratum size
          # even when the preview itself is drawn from a subsample.
          title <- gs_title_literal(spec, row$file[1L], row$n[1L])
        }
        d <- if (length(cols)) d[, cols, with = FALSE] else d
        d <- gs_downsample(d, spec, isTRUE(input$sample_big))
        gs_eval_plot(gs_code_figure(spec, "d", title_expr = title), d)
      }
    }, res = GS_PREVIEW_RES) |>
      # The data cannot change under the app, so the spec -- which carries the
      # categorization rules -- plus the preview controls is the whole key.
      shiny::bindCache(spec_r(), input$preview_mode, input$preview_stratum,
                       input$sample_big)

    # --- tables -------------------------------------------------------------

    output$data_table <- shiny::renderTable({
      utils::head(dat(), 200L)
    }, striped = TRUE, spacing = "xs", rownames = FALSE, digits = 4)

    # A one-line count of what will and will not be drawn, so the n = 0 case
    # and the excluded rows are visible without reading the whole table.
    output$strata_summary <- shiny::renderUI({
      st <- strata()
      miss <- missing_report()
      if (!nrow(st) && !nrow(miss)) return(NULL)

      parts <- character()
      if (nrow(st)) {
        n_empty <- sum(st$n == 0L)
        n_small <- sum(!st$keep & st$n > 0L)
        parts <- sprintf("%d strata, %d exported", nrow(st), sum(st$keep))
        if (n_small) parts <- c(parts, sprintf("%d below the minimum N", n_small))
        if (n_empty) parts <- c(parts, sprintf("%d with N = 0 (no figure)", n_empty))
        possible <- attr(st, "n_possible")
        if (!is.null(possible) && possible > nrow(st)) {
          parts <- c(parts, sprintf("%d of %d possible combinations occur",
                                    nrow(st), possible))
        }
      }
      if (nrow(miss)) {
        parts <- c(parts, sprintf(
          "excluded rows with a missing %s",
          paste(sprintf("%s (%d)", miss$var, miss$n_missing), collapse = ", ")))
      }
      shiny::div(
        class = if (nrow(miss) || !nrow(st) || any(!st$keep))
                  "alert alert-info py-2 px-3 small"
                else "alert alert-success py-2 px-3 small",
        paste(parts, collapse = " | ")
      )
    })

    # Whether a stratum is drawn is already said by its file name -- a stratum
    # with no file gets no figure -- and why it is not is said once, in the
    # summary line above. A status column would repeat both, row by row.
    output$strata_table <- shiny::renderTable({
      st <- strata()
      shiny::validate(shiny::need(
        nrow(st) > 0,
        "Tick at least one variable under 'Separate figures' in the sidebar."
      ))
      spec <- spec_r()
      # The same stem and the same extension the export will use, so that the
      # name listed here is the name that appears on disk.
      names <- paste0(gs_file_stem(gs_export_prefix(spec), st$file),
                      ".", spec$format)
      data.frame(
        Variable = st$var,
        Level = st$level,
        N = st$n,
        `File name` = ifelse(st$keep, names, "-"),
        check.names = FALSE,
        stringsAsFactors = FALSE
      )
    }, striped = TRUE, spacing = "xs", rownames = FALSE, digits = 0)

    # --- generated code -----------------------------------------------------

    output$code <- shiny::renderText({
      probs <- problems()
      if (length(probs)) {
        return(paste0("# Cannot generate code yet:\n# - ",
                      paste(probs, collapse = "\n# - ")))
      }
      # The code is for one figure, so it has to say which one when the Plot
      # tab is showing several at once.
      note <- if (facet_preview()) {
        paste("The Plot tab is showing every figure at once;",
              "this is the code for one of them.")
      } else {
        NULL
      }
      gs_code_script(spec_r(), current_stratum(), note)
    })

    # --- export -------------------------------------------------------------

    shiny::observeEvent(input$export, {
      dt <- layer_data()
      probs <- problems()
      if (length(probs)) {
        shiny::showNotification(paste(probs, collapse = " "), type = "error",
                                duration = 10)
        return()
      }

      spec <- spec_r()
      outdir <- spec$outdir
      dir.create(outdir, showWarnings = FALSE, recursive = TRUE)
      if (!dir.exists(outdir)) {
        shiny::showNotification(paste("Could not create folder:", outdir),
                                type = "error", duration = 10)
        return()
      }

      prefix <- gs_export_prefix(spec)
      dt_list <- gs_split_strata(dt, spec)
      st <- strata()
      n_skipped <- if (nrow(st)) sum(!st$keep) else 0L

      if (!length(dt_list)) {
        shiny::showNotification(
          "Every stratum is empty or below the minimum n; nothing was written.",
          type = "warning", duration = 8
        )
        return()
      }

      device <- gs_device(spec$format)
      ext <- paste0(".", spec$format)
      step <- 1 / length(dt_list)
      written <- 0L
      shiny::withProgress(message = "Writing figures", value = 0, {
        # Every figure is drawn by evaluating the code generator's output, the
        # same way the preview and the R-code tab are, so the file on disk and
        # the code on screen cannot drift apart.
        Map(function(d, nm) {
          stem <- gs_file_stem(prefix, nm)
          title <- if (nzchar(nm)) {
            gs_title_literal(spec, gs_safe_name(nm), nrow(d))
          } else {
            NULL
          }
          p <- gs_eval_plot(gs_code_figure(spec, "d", title_expr = title), d)
          ggplot2::ggsave(file.path(outdir, paste0(stem, ext)), plot = p,
                          width = spec$width, height = spec$height,
                          dpi = spec$dpi, device = device)
          written <<- written + 1L
          shiny::incProgress(step, detail = paste0(stem, ext))
          NULL
        }, dt_list, names(dt_list))
      })

      msg <- sprintf("Wrote %d %s file%s to %s", written,
                     toupper(spec$format),
                     if (written == 1L) "" else "s",
                     normalizePath(outdir, winslash = "/", mustWork = FALSE))
      if (n_skipped > 0L) {
        msg <- paste0(msg, sprintf("; %d stratum/strata skipped (see the Strata tab)",
                                   n_skipped))
      }
      shiny::showNotification(msg, type = "message", duration = 8)
    })
  }
}

#' The graphics device one of `GS_FORMATS` is written through
#'
#' PNG goes through \pkg{ragg}, which is both faster and better at finding a
#' font than the default device. SVG is a vector format and needs a vector
#' device: \pkg{svglite} when it is installed, and otherwise the cairo device
#' that ships with R, so that the option works without a further install.
#'
#' @param format `"png"` or `"svg"`. Anything else is an error.
#' @return A device function, ready to hand to [ggplot2::ggsave()].
#' @keywords internal
#' @noRd
gs_device <- function(format = "png") {
  if (identical(format, "png")) return(ragg::agg_png)
  # Anything but the two formats the UI offers is a bug on the way in, and is
  # worth saying so rather than quietly writing a PNG under an .svg name.
  if (!identical(format, "svg")) {
    stop("Unknown figure format: ", format, call. = FALSE)
  }
  if (requireNamespace("svglite", quietly = TRUE)) {
    return(svglite::svglite)
  }
  grDevices::svg
}

# Nominal resolution the preview is drawn at. shiny::renderPlot() defaults to
# 72, at which a 620-pixel-tall panel is treated as an 8.6-inch figure and
# ggplot2's 11-point text comes out looking tiny beside it. Raising the
# resolution shrinks the notional figure instead, so the same theme text fills
# the panel the way it fills the exported figure. A browser window wide enough
# to hold the sidebar and a 900-pixel plot puts the notional figure at about
# the 7 inches the export defaults to, which is what makes the two look alike.
# The plot code is untouched by this, so the preview still matches the R-code
# tab exactly.
GS_PREVIEW_RES <- 150

# Row counts above which a preview is drawn from a sample. Scatter and Line
# need a much lower cap: every row becomes its own mark or line segment, so
# they cost roughly 20x what a boxplot of the same data costs.
#
# Both caps are set high on purpose. Sampling is a compromise -- it is the one
# place where what is on screen is not what the data says -- so it should be
# reached by data that genuinely will not draw, not by data that is merely
# large. A boxplot, histogram or density summarises its rows before drawing
# anything, so five million of them is a few seconds of arithmetic; the caps
# below are where the wait stops being worth it, not where it starts.
GS_SAMPLE_SCATTER <- 1000000L
GS_SAMPLE_OTHER <- 10000000L

#' Row target for previewing a spec's figure
#'
#' A target rather than a hard cap: `gs_downsample()` will go over it to keep
#' a whole trajectory or a whole stratum, see there.
#'
#' A Kaplan-Meier curve is never sampled. The other plot types show the shape
#' of a distribution, which a large sample preserves; a survival curve *is*
#' the estimate, and a sampled one would be a different, wider curve shown
#' without saying so.
#' @keywords internal
#' @noRd
gs_sample_threshold <- function(spec) {
  type <- spec$plot_type
  if (identical(type, GS_KM)) return(Inf)
  if (type %in% c("Scatter", GS_LINE)) GS_SAMPLE_SCATTER else GS_SAMPLE_OTHER
}

#' Thin a data.table so the preview stays responsive
#'
#' Applies to previews only. The in-app export and the generated script always
#' use every row, so nothing written to disk is ever silently subsampled.
#'
#' `threshold` is a target, not a strict upper bound, and both of the grouped
#' branches can exceed it rather than drop a group entirely:
#'
#'   * A line plot with an ID variable is sampled one whole trajectory at a
#'     time -- dropping rows from within a line would draw a line nobody has,
#'     in the same way that a sampled survival curve is a different curve --
#'     so the rows returned are those of the sampled IDs, which is only
#'     `threshold` on average.
#'   * When `by` is given, sampling is done within each group so that small
#'     strata keep their rows instead of being washed out by large ones, and
#'     every group keeps at least one row. With more groups than `threshold`
#'     that alone puts the result over it -- which is the intended trade: a
#'     panel drawn from no rows at all is worse than a preview slightly larger
#'     than asked for. In practice the caps are in the millions and the panels
#'     in the tens, so the case is a corner one.
#'
#' The ungrouped branch is exact.
#'
#' @param d A `data.table`.
#' @param spec A spec list, which sets the sampling unit and, by default, the
#'   row target.
#' @param enabled Whether the user left subsampling switched on.
#' @param by Optional grouping column to sample within.
#' @param threshold Row target. Defaults to the one the spec's plot type asks
#'   for; passed explicitly by the tests, which would otherwise have to
#'   allocate a table of the real cap's size to reach this code at all.
#' @return `d`, or a sample of it.
#' @keywords internal
#' @noRd
gs_downsample <- function(d, spec, enabled = TRUE, by = NULL,
                          threshold = gs_sample_threshold(spec)) {
  if (!isTRUE(enabled) || !nrow(d)) return(d)
  if (nrow(d) <= threshold) return(d)

  # Blank for every plot type but Line, see gs_normalize_spec().
  id <- spec$id %||% ""
  if (nzchar(id) && id %in% names(d)) {
    ids <- unique(d[[id]])
    # The fraction first: two integer row counts multiplied together overflow.
    n_keep <- max(1L, as.integer(floor(length(ids) * (threshold / nrow(d)))))
    # Computed outside the `[` so that `id`, the local holding the column's
    # name, cannot be read as the column of that name instead.
    .gs_keep <- d[[id]] %in% sample(ids, n_keep)
    return(d[.gs_keep])
  }

  if (!is.null(by) && by %in% names(d)) {
    per_group <- max(1L, threshold %/% max(data.table::uniqueN(d[[by]]), 1L))
    return(d[, if (.N > per_group) .SD[sample.int(.N, per_group)] else .SD,
             by = c(by)])
  }
  d[sample.int(nrow(d), threshold)]
}

# The UI is built once and never regenerated; the server only pushes new
# *choices* into the selectors when a data set is loaded. Plot-type-dependent
# controls are hidden with conditionalPanel() so that switching plot type is a
# pure client-side operation with no server round-trip.

#' Build the ggstratify user interface
#'
#' There is no control for loading data. The data set is the argument the
#' launcher was called with, already read and already typed by the user; see
#' the *Before you start* section of [ggstratify()].
#' @keywords internal
#' @noRd
gs_ui <- function() {
  bslib::page_sidebar(
    title = "ggstratify",
    # font_scale multiplies Bootstrap's base font size, so every control,
    # label and table in the app grows together. The exported PNGs are drawn
    # by ggplot2 and are unaffected by it.
    theme = bslib::bs_theme(version = 5, preset = "flatly", font_scale = 1.35),
    sidebar = bslib::sidebar(
      width = 400,
      open = "desktop",

      bslib::accordion(
        multiple = TRUE,
        open = c("Describe", "Layers"),

        bslib::accordion_panel(
          "Describe",
          shiny::selectInput("plot_type", "Type of graph", GS_PLOT_TYPES,
                             selected = "Boxplot"),
          shiny::conditionalPanel(
            condition = gs_js_not_km(),
            shiny::selectInput("yvar", "Y-variable (continuous)", choices = NULL)
          ),
          shiny::conditionalPanel(
            condition = sprintf("!%s.includes(input.plot_type) && %s",
                                gs_js_array(GS_XONLY_TYPES), gs_js_not_km()),
            shiny::selectInput("xvar", "X-variable", choices = NULL)
          ),
          shiny::conditionalPanel(
            condition = gs_js_km(),
            shiny::selectInput("timevar", "Time variable", choices = NULL),
            shiny::selectInput("eventvar", "Event (1 = event, 0 = censored)",
                               choices = NULL)
          ),
          shiny::conditionalPanel(
            condition = gs_js_type(GS_LINE),
            shiny::selectInput("idvar", "One line per (e.g. patient ID)",
                               choices = NULL),
            shiny::helpText(
              "Leave this empty when each x value already has one value to",
              "draw -- a mean per visit, say. Choose the subject identifier",
              "to draw one line per subject instead."
            )
          ),
          shiny::selectInput("group", "Group (colour by this variable)",
                             choices = NULL)
        ),

        bslib::accordion_panel(
          "Categorize",
          shiny::helpText(
            "Turn a continuous variable into groups. The new variable can be",
            "used as a layer like any other categorical variable"
          ),
          shiny::selectInput("cut_var", "Continuous variable", choices = NULL),
          shiny::selectInput("cut_method", "How", choices = GS_CUT_METHODS),
          shiny::conditionalPanel(
            condition = "input.cut_method != 'breaks'",
            shiny::numericInput("cut_n", "Number of groups", value = 4,
                                min = 2, max = 20, step = 1)
          ),
          shiny::conditionalPanel(
            condition = "input.cut_method == 'breaks'",
            shiny::textInput("cut_points", "Cut points, comma separated",
                             placeholder = "50, 65, 80"),
            shiny::helpText(
              "Cut points are boundaries: 65 gives (-Inf, 65] and (65, Inf]."
            )
          ),
          shiny::textInput("cut_name", "New variable name"),
          shiny::actionButton("add_cut", "Add categorized variable",
                              class = "btn-secondary w-100",
                              icon = shiny::icon("plus")),
          shiny::uiOutput("cuts_list")
        ),

        bslib::accordion_panel(
          "Layers",
          shiny::helpText(
            "Choose how to split your figure using additional variables.",
            "Leave both boxes empty: create one figure with all data.",
            "Panels: show multiple small plots within one figure, one for",
            "each level of the variable.",
            "Separate figures: create a separate figure for each level of",
            "the variable.",
            "A variable can be used in either Panels or Separate figures,",
            "but not both."
          ),
          shiny::selectInput("facet", "Panels side by side in one figure",
                             choices = GS_NONE),
          shiny::checkboxGroupInput("strat_vars",
                                    "Separate figures (one file per level)",
                                    choices = character()),
          shiny::conditionalPanel(
            condition = "input.strat_vars && input.strat_vars.length > 1",
            shiny::radioButtons("strat_mode", NULL, choices = GS_STRAT_MODES,
                                selected = "independent")
          ),
          shiny::uiOutput("layer_summary"),
          shiny::numericInput("min_n", "Minimum rows per separate figure",
                              value = 10, min = 0, step = 1),
          shiny::helpText(
            "No separate figure is created for a group with fewer rows than",
            "this number. Set to 0 to include every non-empty group."
          ),
          shiny::checkboxInput("show_n",
                               "Show N in the title and on the panel strips",
                               TRUE),
          shiny::helpText(
            "Every figure says how many observations it was drawn from, and",
            "so does every panel inside it.",
            "A stratum with N = 0 (an unused factor level) is always listed",
            "on the Strata tab and never drawn.",
            "Rows missing a layer variable are excluded, and counted"
          )
        ),

        bslib::accordion_panel(
          "Plot options",
          shiny::conditionalPanel(
            condition = "['Boxplot','Violin','Dot + Error'].includes(input.plot_type)",
            shiny::checkboxInput("jitter", "Add jittered points", FALSE)
          ),
          shiny::conditionalPanel(
            condition = "['Histogram','Dotplot'].includes(input.plot_type)",
            shiny::numericInput("binwidth", "Binwidth (blank = automatic)",
                                value = NA, min = 0)
          ),
          shiny::conditionalPanel(
            condition = "input.plot_type == 'Density'",
            shiny::sliderInput("bw_adjust", "Bandwidth adjustment",
                               min = 0.1, max = 3, value = 1, step = 0.1)
          ),
          shiny::conditionalPanel(
            condition = gs_js_type(GS_LINE),
            shiny::checkboxInput("line_points", "Mark the observations", FALSE)
          ),
          shiny::conditionalPanel(
            condition = sprintf("%s.includes(input.plot_type)",
                                gs_js_array(GS_SMOOTH_TYPES)),
            shiny::checkboxInput("smooth", "Add a LOWESS smoother", FALSE),
            shiny::conditionalPanel(
              condition = "input.smooth == true",
              shiny::sliderInput("smooth_span", "Smoother span (larger = smoother)",
                                 min = 0.1, max = 1, value = 0.75, step = 0.05),
              shiny::checkboxInput("smooth_se", "Show the smoother's 95% band",
                                   FALSE),
              shiny::helpText(
                "Drawn with a grouping variable, the smoother is fitted",
                "separately within each group."
              )
            )
          ),
          shiny::conditionalPanel(
            condition = gs_js_km(),
            shiny::checkboxInput("km_ci", "Show 95% confidence band", FALSE),
            shiny::checkboxInput("km_censor", "Mark censored observations", TRUE),
            shiny::checkboxInput("km_ylim", "Y axis from 0 to 1", TRUE),
            shiny::checkboxInput("km_risk", "Add the number-at-risk table",
                                 FALSE),
            shiny::conditionalPanel(
              condition = "input.km_risk == true",
              shiny::helpText(
                "Drawn under the curve, one row per group, counted at the",
                "times the x axis is labelled at. The tail of a survival",
                "curve is read differently when 12 people are left than when",
                "300 are."
              )
            )
          ),
          shiny::sliderInput("alpha", "Alpha (opacity)",
                             min = 0.1, max = 1, value = 0.6, step = 0.05),
          shiny::checkboxInput(
            "sample_big",
            "Speed up previews of very large datasets",
            TRUE
          ),
          shiny::helpText(
            "For the on-screen preview only, randomly use a subset of rows",
            "when there are more than 1,000,000 rows for Scatter and Line",
            "plots, or more than 10,000,000 rows for other plot types.",
            "This does not remove groups or change which figures are made.",
            "Exports and generated code always use all rows. Kaplan-Meier",
            "curves are never sampled."
          )
        ),

        bslib::accordion_panel(
          "Appearance",
          shiny::selectInput("theme", "Theme", choices = GS_THEMES,
                             selected = "theme_bw()"),
          shiny::selectInput("palette", "Colour palette", choices = GS_PALETTES),
          shiny::helpText(
            "A palette colours the groups, so it only changes the figure once",
            "a Group (or colour) variable has been chosen under Describe."
          ),
          shiny::textInput("title", "Title (blank = stratum name)"),
          shiny::textInput("lab_x", "X-axis label"),
          shiny::textInput("lab_y", "Y-axis label"),
          shiny::textInput("lab_legend", "Legend label"),
          shiny::fluidRow(
            shiny::column(6, shiny::numericInput("xlim_min", "X axis from",
                                                 value = NA)),
            shiny::column(6, shiny::numericInput("xlim_max", "X axis to",
                                                 value = NA))
          ),
          shiny::fluidRow(
            shiny::column(6, shiny::numericInput("ylim_min", "Y axis from",
                                                 value = NA)),
            shiny::column(6, shiny::numericInput("ylim_max", "Y axis to",
                                                 value = NA))
          ),
          shiny::helpText(
            "Blank is automatic, and either end can be left blank on its own.",
            "The figure is zoomed into the range rather than filtered to it,",
            "so a boxplot's median and a density's shape stay the ones the",
            "whole data gives. A range typed here replaces a Kaplan-Meier",
            "curve's 'Y axis from 0 to 1'."
          )
        ),

        bslib::accordion_panel(
          "Export",
          shiny::textInput("data_name", "Data object name (used in the code)",
                           value = "mydata"),
          shiny::textInput("outdir", "Output folder", value = "figures"),
          shiny::helpText(
            "Relative to the working directory. The folder is created if it",
            "does not exist yet."
          ),
          shiny::textInput("prefix", "File-name prefix (blank = plot type)"),
          shiny::selectInput("format", "File format", GS_FORMATS,
                             selected = "png"),
          shiny::fluidRow(
            shiny::column(6, shiny::numericInput("width", "Width (in)",
                                                 value = 7, min = 1, step = 0.5)),
            shiny::column(6, shiny::numericInput("height", "Height (in)",
                                                 value = 5, min = 1, step = 0.5))
          ),
          shiny::conditionalPanel(
            condition = "input.format != 'svg'",
            shiny::numericInput("dpi", "Resolution (dpi)", value = 300,
                                min = 72, step = 10)
          ),
          shiny::actionButton("export", "Export all figures",
                              class = "btn-primary w-100",
                              icon = shiny::icon("download"))
        )
      )
    ),

    bslib::navset_card_tab(
      id = "main_tabs",

      bslib::nav_panel(
        "Plot",
        shiny::div(
          class = "d-flex flex-wrap align-items-end gap-4 mb-2",
          # The variables decide how many figures there are, so they decide
          # what the preview can show. The switch is offered only when both
          # answers are possible: with no separate figures there is only one
          # figure to show, and with the panels already spent on a facet
          # variable there is nowhere to put the others. The server keeps the
          # selection in step, so nothing here needs explaining on screen.
          shiny::conditionalPanel(
            condition = gs_js_has_strata(),
            shiny::radioButtons(
              "preview_mode", "Show",
              choices = c("All figures at once" = "facet",
                          "One figure at a time" = "single"),
              selected = "facet", inline = TRUE
            )
          ),
          # Which figure to show is a question only when there is more than
          # one, and only when they are shown one at a time -- which a facet
          # variable makes true whatever the switch above says.
          shiny::conditionalPanel(
            condition = sprintf("%s && (input.preview_mode == 'single' || %s)",
                                gs_js_strat_vars(), gs_js_has_facet()),
            shiny::selectInput("preview_stratum", "Which figure",
                               choices = NULL, width = "360px")
          )
        ),
        shiny::uiOutput("preview_note"),
        # Tall enough that the figure keeps roughly the 7:5 shape the export
        # defaults to once GS_PREVIEW_RES has been applied to it.
        shiny::plotOutput("plot", height = "660px")
      ),

      bslib::nav_panel(
        "Data",
        shiny::helpText("First 200 rows."),
        shiny::div(class = "gs-scroll", shiny::tableOutput("data_table"))
      ),

      bslib::nav_panel(
        "Strata",
        shiny::helpText(
          "Every figure the current layers produce, with the number of",
          "observations behind it. A stratum with no file name gets no",
          "figure: either it has N = 0, or it is below the minimum N. The",
          "line above says how many of each."
        ),
        shiny::uiOutput("strata_summary"),
        shiny::div(class = "gs-scroll", shiny::tableOutput("strata_table"))
      ),

      bslib::nav_panel(
        "R-code",
        shiny::helpText(
          "The ggplot2 code for the figure on the Plot tab, and nothing else.",
          "Writing the files is what the Export button is for; what is worth",
          "taking away from here is the figure itself, to paste into your own",
          "script and edit."
        ),
        shiny::div(
          class = "mb-2",
          shiny::actionButton("copy_code", "Copy to clipboard",
                              icon = shiny::icon("clipboard"))
        ),
        shiny::verbatimTextOutput("code")
      )
    ),

    shiny::tags$head(shiny::tags$style(shiny::HTML(GS_CSS))),
    shiny::tags$script(shiny::HTML(GS_COPY_JS))
  )
}

#' Render a string as a quoted JavaScript string literal
#'
#' Every conditionalPanel() condition below is built by pasting one of the
#' package's own constants into a fragment of JavaScript, so nothing hostile
#' can reach it. The escaping is here so that adding a plot type or a theme
#' whose name happens to contain an apostrophe -- `Tufte's boxplot` -- breaks
#' nothing more than the label it is written in.
#' @keywords internal
#' @noRd
gs_js_str <- function(x) {
  # fixed = TRUE on both: the strings being escaped are labels, not patterns.
  x <- gsub("\\", "\\\\", x, fixed = TRUE)
  x <- gsub("'", "\\'", x, fixed = TRUE)
  paste0("'", x, "'")
}

#' Render a character vector as a JavaScript array literal
#' @keywords internal
#' @noRd
gs_js_array <- function(x) {
  paste0("[", paste0(gs_js_str(x), collapse = ","), "]")
}

#' conditionalPanel() condition for a plot-type-only control
#' @keywords internal
#' @noRd
gs_js_type <- function(type) paste0("input.plot_type == ", gs_js_str(type))

#' conditionalPanel() conditions for the Kaplan-Meier-only controls
#' @keywords internal
#' @noRd
gs_js_km <- function() gs_js_type(GS_KM)

#' @rdname gs_js_km
#' @keywords internal
#' @noRd
gs_js_not_km <- function() paste0("input.plot_type != ", gs_js_str(GS_KM))

#' conditionalPanel() condition: at least one variable splits the figures
#' @keywords internal
#' @noRd
gs_js_strat_vars <- function() {
  "input.strat_vars && input.strat_vars.length > 0"
}

#' conditionalPanel() condition: a variable is panelling the figure
#' @keywords internal
#' @noRd
gs_js_has_facet <- function() paste0("input.facet != ", gs_js_str(GS_NONE))

#' conditionalPanel() condition: the preview could show either mode
#'
#' Both modes are possible only when the figures are split *and* the panels
#' are still free, because showing every figure at once borrows the same
#' `facet_wrap()` a facet variable would use.
#' @keywords internal
#' @noRd
gs_js_has_strata <- function() {
  sprintf("%s && !(%s)", gs_js_strat_vars(), gs_js_has_facet())
}

# bs_theme(font_scale=) raises Bootstrap's own base size, and everything
# Bootstrap sizes in `em` follows it. What is left here is the text that would
# not: rules written in px by Shiny or by the slider widget, and the handful
# Bootstrap deliberately shrinks. Sizes are given in `em` rather than `rem` on
# purpose -- `rem` is measured from the unscaled root font, so a rule in `rem`
# would pull this text back down instead of leaving it scaled.
GS_CSS <- "
#code {
  background: #f6f8fa;
  color: #24292f;
  border: 1px solid #d7dde3;
  border-radius: 6px;
  padding: 14px 16px;
  font-size: 0.95em;
  line-height: 1.55;
  max-height: 70vh;
  overflow: auto;
}
.accordion-button { font-weight: 600; }
.gs-problem { color: #b02a37; font-weight: 500; }
.gs-scroll { max-height: 65vh; overflow: auto; }
.gs-scroll table { font-size: 0.95em; }
.gs-skip { color: #b02a37; }

/* Notes, hints and the alerts the sidebar and the Strata tab use, all of
   which Bootstrap sets to 0.875em. */
.small, small, .form-text, .help-block, .alert { font-size: 0.95em; }

/* Labels, so a control is as easy to read as its value. */
.form-label, .control-label, .shiny-input-container > label {
  font-weight: 600;
}

/* ionRangeSlider ships its own px sizes, which no theme scale reaches. */
.irs-min, .irs-max, .irs-single, .irs-from, .irs-to { font-size: 0.8em; }
.irs-grid-text { font-size: 0.7em; }

/* The message shiny::validate() paints over the plot area. */
.shiny-output-error-validation { font-size: 1.05em; color: #b02a37; }
"

# Copying happens in the browser so that it also works over a remote session.
#
# The Clipboard API is only available in a secure context, which localhost is
# but a plain http:// address on another machine is not, so the button has to
# survive not having it: execCommand('copy') over a temporary textarea is the
# older path every browser still honours. Either way the button says what
# happened -- a copy button that silently does nothing leaves the user
# pasting whatever was on the clipboard before.
GS_COPY_JS <- "
$(document).on('click', '#copy_code', function() {
  var el = document.getElementById('code');
  if (!el) { return; }
  var btn = this;
  // Read once and remember: a second click while the button is still
  // saying 'Copied' would otherwise make that the label it restores.
  if (!btn.dataset.gsLabel) { btn.dataset.gsLabel = btn.innerHTML; }
  var label = btn.dataset.gsLabel;
  var say = function(msg) {
    btn.innerHTML = msg;
    setTimeout(function() { btn.innerHTML = label; }, 1500);
  };
  var fallback = function() {
    var ta = document.createElement('textarea');
    ta.value = el.innerText;
    ta.setAttribute('readonly', '');
    ta.style.position = 'fixed';
    ta.style.left = '-9999px';
    document.body.appendChild(ta);
    ta.select();
    var ok = false;
    try { ok = document.execCommand('copy'); } catch (e) { ok = false; }
    document.body.removeChild(ta);
    say(ok ? 'Copied' : 'Press Ctrl+C to copy');
  };
  if (navigator.clipboard && navigator.clipboard.writeText) {
    navigator.clipboard.writeText(el.innerText).then(
      function() { say('Copied'); }, fallback);
  } else {
    fallback();
  }
});
"

#' A synthetic clinical cohort
#'
#' Six hundred simulated patients, shaped like the descriptive tables that
#' motivate this package: a few continuous measurements crossed with several
#' categorical variables worth stratifying on.
#'
#' Two features are deliberate. `site` declares a fourth level, `"Site D"`,
#' that recruited nobody, so stratifying by `site` produces a stratum with
#' `n = 0`; the app lists it and draws no figure. `severity` has a small
#' `"Severe"` group -- 20 patients against 381 and 199 -- which is what the
#' minimum-n control is there to be tried on: it is drawn at the default
#' minimum of 10, and disappears from the figures, with a reason, as soon as
#' the minimum is raised past 20.
#'
#' The data are simulated. They describe no real patients and support no
#' clinical conclusion.
#'
#' `fu_days` and `death` make the data usable for a Kaplan-Meier curve, and
#' `age` for the categorization controls.
#'
#' @format A data frame with 600 rows and 11 columns:
#' \describe{
#'   \item{id}{Patient identifier, `"P0001"` to `"P0600"`.}
#'   \item{age}{Age in years.}
#'   \item{sex}{Factor: `"Male"`, `"Female"`.}
#'   \item{site}{Factor: `"Site A"`, `"Site B"`, `"Site C"`, `"Site D"`.
#'     `"Site D"` has no observations.}
#'   \item{treatment}{Factor: `"Control"`, `"Low dose"`, `"High dose"`.}
#'   \item{severity}{Factor: `"Mild"`, `"Moderate"`, `"Severe"`.}
#'   \item{bmi}{Body mass index, kg/m^2.}
#'   \item{crp}{C-reactive protein, mg/L.}
#'   \item{los_days}{Length of stay in days.}
#'   \item{fu_days}{Days of follow-up, to death or to censoring. Censoring is
#'     by dropout or by the end of the study at 365 days.}
#'   \item{death}{`1` if the patient died during follow-up, `0` if censored.}
#' }
#'
#' @source Simulated by `data-raw/epi_cohort.R`.
#'
#' @examples
#' str(epi_cohort)
#'
#' # The empty stratum that the app reports with n = 0.
#' table(epi_cohort$site)
"epi_cohort"

#' @keywords internal
#' @import data.table
#' @import ggplot2
#' @import shiny
#' @importFrom grDevices svg
# patchwork is called from generated code -- `patchwork::wrap_plots()` stacks
# the number-at-risk table under its curve -- which the package then evaluates.
# That is a real run-time dependency, but it lives in a string, where R CMD
# check cannot see it; naming one function here is what makes it visible.
#' @importFrom patchwork wrap_plots
#' @importFrom stats setNames
#' @importFrom survival Surv survfit
#' @importFrom utils head modifyList
"_PACKAGE"

# Columns referenced by data.table's non-standard evaluation.
utils::globalVariables(c(
  ".", ".N", ".SD", ".facet_label", ".gs_stratum", ".strat_label",
  "can_stratify", "file", "is_categorical",
  "is_continuous", "is_event", "is_numeric", "keep", "label", "level", "n",
  "n_levels", "n_missing", "status", "var"
))

# Turning a continuous variable into a categorical one. A "cut" is a small
# plain list describing one such rule; the spec carries a list of them.
#
# There is one implementation, not two: `gs_code_cuts()` writes the
# `data.table` lines, and `gs_apply_cuts()` runs those very lines against the
# data the app is holding. The derived column in the app is therefore always
# the column the generated script produces.

#' Build one categorization rule with defaults filled in
#'
#' @param var Name of the continuous column to categorize.
#' @param new Name of the derived column. Defaults to `<var>_cat`.
#' @param method One of `"quantile"` (equal-sized groups), `"equal"`
#'   (equal-width bins) or `"breaks"` (user-supplied cut points).
#' @param n Number of groups, for the first two methods.
#' @param breaks Internal cut points, for `method = "breaks"`. They are
#'   interpreted as boundaries between open-ended bins, so `65` gives
#'   `(-Inf, 65]` and `(65, Inf]`.
#' @return A list with elements `var`, `new`, `method`, `n`, `breaks`.
#' @keywords internal
#' @noRd
gs_cut <- function(var, new = NULL, method = "quantile", n = 4L,
                   breaks = numeric()) {
  method <- match.arg(method, unname(GS_CUT_METHODS))
  n <- as.integer(gs_num(n, 4))
  if (is.na(n)) n <- 4L
  breaks <- suppressWarnings(as.numeric(breaks))
  breaks <- sort(unique(breaks[is.finite(breaks)]))
  var <- as.character(var)
  if (is.null(new) || !nzchar(new)) new <- paste0(var, "_cat")
  # n is floored at 2: one group is not a categorization, and every caller
  # that builds a cut goes through here, so nothing downstream has to carry
  # the case. gs_check_cut() still holds the lower bound as well, for a cut
  # assembled as a plain list without this constructor.
  list(var = var, new = as.character(new), method = method,
       n = max(n, 2L), breaks = breaks)
}

#' Coerce whatever was handed to `gs_spec()` into a list of cuts
#' @keywords internal
#' @noRd
gs_as_cuts <- function(cuts) {
  if (is.null(cuts) || !length(cuts)) return(list())
  # A single cut, passed unwrapped.
  if (!is.null(cuts$var)) cuts <- list(cuts)
  lapply(cuts, function(cut) {
    do.call(gs_cut, cut[intersect(names(cut), c("var", "new", "method", "n",
                                                "breaks"))])
  })
}

#' Read cut points typed into a text box
#'
#' Commas, semicolons and whitespace all separate. `NA` in the result means
#' the user typed something that is not a number, which the caller reports.
#' @keywords internal
#' @noRd
gs_parse_breaks <- function(x) {
  parts <- unlist(strsplit(as.character(x %||% ""), "[,;[:space:]]+"))
  parts <- parts[nzchar(parts)]
  if (!length(parts)) return(numeric())
  suppressWarnings(as.numeric(parts))
}

#' A default name for a derived column that does not clash with the data
#' @keywords internal
#' @noRd
gs_cut_name <- function(var, taken = character()) {
  base <- paste0(var, "_cat")
  if (!base %in% taken) return(base)
  # Unbounded: a capped search that gives up and returns `base` would hand
  # back the very name it was asked to avoid. The loop ends on the first free
  # candidate, so it runs `length(taken)` times at the very worst.
  i <- 2L
  repeat {
    cand <- paste0(base, i)
    if (!cand %in% taken) return(cand)
    i <- i + 1L
  }
}

#' A one-line human description of a cut, for the sidebar list
#' @keywords internal
#' @noRd
gs_cut_describe <- function(cut) {
  how <- switch(
    cut$method,
    quantile = sprintf("%d quantile groups", cut$n),
    equal    = sprintf("%d equal-width bins", cut$n),
    breaks   = paste0("cut at ", paste(gs_n_vec(cut$breaks), collapse = ", "))
  )
  sprintf("%s <- %s (%s)", cut$new, cut$var, how)
}

#' Format a numeric vector for display
#' @keywords internal
#' @noRd
gs_n_vec <- function(x) vapply(x, gs_n, character(1L))

#' Drop cuts whose source column is missing, and cuts that collide
#'
#' Later cuts can legitimately read a column an earlier cut created, so the
#' set of known names grows as the list is walked.
#' @keywords internal
#' @noRd
gs_valid_cuts <- function(cuts, nms) {
  out <- list()
  for (cut in gs_as_cuts(cuts)) {
    if (!cut$var %in% nms) next
    if (cut$new %in% nms) next
    out[[length(out) + 1L]] <- cut
    nms <- c(nms, cut$new)
  }
  out
}

#' Add the derived columns to a copy of the data
#'
#' The copy is what keeps the caller's table untouched; the derived columns
#' are then written into it by reference.
#'
#' A rule that fails (non-unique quantile breaks, say) is skipped rather than
#' aborting the app, and its message is attached as the `gs_cut_error`
#' attribute so the caller can report it.
#'
#' @param dt A `data.table`.
#' @param cuts A list of cuts.
#' @return A `data.table` with one extra factor column per applied cut.
#' @keywords internal
#' @noRd
gs_apply_cuts <- function(dt, cuts) {
  cuts <- gs_valid_cuts(cuts, names(dt))
  if (!length(cuts)) return(dt)

  out <- data.table::copy(dt)
  # data.table's namespace, so that `:=` and friends resolve exactly as they
  # do in the generated script, which starts with library(data.table).
  env <- new.env(parent = asNamespace("data.table"))
  assign("dt", out, envir = env)

  errors <- character()
  for (cut in cuts) {
    line <- gs_code_cut_line(cut, "dt")
    # tryCatch rather than try(): the message comes from the condition the
    # handler is given, instead of from an attribute of the returned object.
    msg <- NULL
    tryCatch(eval(parse(text = line), envir = env),
             error = function(e) msg <<- conditionMessage(e))
    if (!is.null(msg)) {
      errors <- c(errors, sprintf("%s: %s", cut$new, msg))
      if (cut$new %in% names(out)) out[, (cut$new) := NULL]
    }
  }
  if (length(errors)) attr(out, "gs_cut_error") <- errors
  out[]
}

#' Check that one cut can be applied to the data
#'
#' Run before a cut is added, so that a mistake is reported as a notification
#' instead of appearing as a broken column.
#'
#' @param dt A `data.table`.
#' @param cut A cut, see `gs_cut()`.
#' @return A character vector of problems; empty when the cut is usable.
#' @keywords internal
#' @noRd
gs_check_cut <- function(dt, cut) {
  if (!cut$var %in% names(dt)) {
    return(sprintf("'%s' is not a column in the data.", cut$var))
  }
  problems <- character()
  if (!is.numeric(dt[[cut$var]])) {
    problems <- c(problems, sprintf(
      "'%s' is not numeric; only numeric variables can be categorized.",
      cut$var))
  }
  if (!nzchar(cut$new)) {
    problems <- c(problems, "Give the new variable a name.")
  } else if (cut$new %in% names(dt)) {
    problems <- c(problems, sprintf(
      "'%s' already exists in the data; choose another name.", cut$new))
  }
  if (cut$method %in% c("quantile", "equal") && (cut$n < 2L || cut$n > 20L)) {
    problems <- c(problems, "The number of groups must be between 2 and 20.")
  }
  if (cut$method == "breaks" && !length(cut$breaks)) {
    problems <- c(problems, "Give at least one cut point.")
  }
  if (length(problems)) return(problems)

  # Only the source column is copied, so the trial run stays cheap on wide
  # data.
  trial <- gs_apply_cuts(dt[, cut$var, with = FALSE], list(cut))
  err <- attr(trial, "gs_cut_error")
  if (length(err)) return(err)
  if (!cut$new %in% names(trial)) {
    return(sprintf("'%s' could not be categorized this way.", cut$var))
  }
  used <- sum(table(trial[[cut$new]]) > 0L)
  if (used < 2L) {
    return(sprintf(
      "This gives fewer than two non-empty groups of '%s'; try fewer groups or other cut points.",
      cut$var))
  }
  character()
}

cohort <- function() gs_prepare_data(epi_cohort)

test_that("gs_cut fills in a name and tidies the cut points", {
  cut <- gs_cut("age")
  expect_equal(cut$new, "age_cat")
  expect_equal(cut$method, "quantile")
  expect_equal(cut$n, 4L)

  # Unsorted, duplicated and non-finite cut points are all dealt with here so
  # that nothing downstream has to.
  expect_equal(gs_cut("age", method = "breaks",
                      breaks = c(80, 65, 65, NA, Inf))$breaks,
               c(65, 80))
  expect_equal(gs_cut("age", n = 1)$n, 2L)
})

test_that("cut points are read from what the user typed", {
  expect_equal(gs_parse_breaks("50, 65 ; 80"), c(50, 65, 80))
  expect_equal(gs_parse_breaks(" 65 "), 65)
  expect_equal(gs_parse_breaks(""), numeric())
  expect_true(anyNA(gs_parse_breaks("50, abc")))
})

test_that("each method produces the categorization it promises", {
  dt <- cohort()
  out <- gs_apply_cuts(dt, list(
    gs_cut("age", method = "quantile", n = 4),
    gs_cut("bmi", new = "bmi_bin", method = "equal", n = 3),
    gs_cut("age", new = "age65", method = "breaks", breaks = 65)
  ))

  expect_equal(nlevels(out$age_cat), 4L)
  # Quantile groups are equal-sized to within a tie.
  expect_lt(diff(range(table(out$age_cat))), nrow(dt) * 0.05)

  expect_equal(nlevels(out$bmi_bin), 3L)
  expect_equal(levels(out$age65), c("(-Inf,65]", "(65, Inf]"))
  expect_equal(sum(out$age65 == "(-Inf,65]"), sum(dt$age <= 65))
})

test_that("categorizing never touches the data it was given", {
  dt <- cohort()
  before <- names(dt)
  out <- gs_apply_cuts(dt, list(gs_cut("age")))
  expect_equal(names(dt), before)
  expect_true("age_cat" %in% names(out))
})

test_that("a derived variable is categorical and can be stratified on", {
  out <- gs_apply_cuts(cohort(), list(gs_cut("age", method = "breaks",
                                             breaks = 65)))
  info <- gs_classify_vars(out)
  expect_true(info[var == "age_cat", is_categorical])
  expect_true(info[var == "age_cat", can_stratify])
  expect_true("age_cat" %in% gs_vars_of(info, "stratify"))
})

test_that("later rules can read a column an earlier rule created", {
  # ...but a rule whose source is missing, or whose name is taken, is dropped.
  cuts <- list(gs_cut("age", new = "age_cat"),
               gs_cut("nope", new = "x"),
               gs_cut("bmi", new = "age"))
  kept <- gs_valid_cuts(cuts, names(cohort()))
  expect_length(kept, 1L)
  expect_equal(kept[[1]]$new, "age_cat")
})

test_that("gs_check_cut refuses what cannot work, before it is added", {
  dt <- cohort()
  expect_match(gs_check_cut(dt, gs_cut("sex")), "not numeric")
  expect_match(gs_check_cut(dt, gs_cut("nope")), "not a column")
  expect_match(gs_check_cut(dt, gs_cut("age", new = "sex")), "already exists")
  expect_match(gs_check_cut(dt, gs_cut("age", method = "breaks")),
               "at least one cut point")
  expect_match(gs_check_cut(dt, gs_cut("age", n = 99)), "between 2 and 20")
  # A binary column has no four quantile groups to give.
  expect_match(gs_check_cut(dt, gs_cut("death", n = 4)), "fewer than two")
  expect_length(gs_check_cut(dt, gs_cut("age", n = 3)), 0L)
})

test_that("a rule that fails is skipped rather than breaking the app", {
  # cut() refuses a character column. gs_check_cut() stops this reaching the
  # data in the app; if one ever does, the column is dropped and the message
  # kept rather than the whole table failing to build.
  dt <- data.table::data.table(x = letters[1:5])
  out <- gs_apply_cuts(dt, list(gs_cut("x", method = "breaks", breaks = 2)))
  expect_false("x_cat" %in% names(out))
  expect_match(attr(out, "gs_cut_error"), "x_cat")
})

test_that("the generated lines are the ones the app itself runs", {
  cuts <- list(gs_cut("age", method = "quantile", n = 3))
  lines <- gs_code_cuts(cuts, "dt")
  expect_match(lines[2], "dt[, age_cat := cut(age, breaks = unique(stats::quantile(",
               fixed = TRUE)
  expect_silent(parse(text = paste(lines, collapse = "\n")))

  # Running the emitted line by hand gives the app's column, value for value.
  dt <- cohort()
  by_app <- gs_apply_cuts(dt, cuts)$age_cat
  by_hand <- data.table::copy(dt)
  eval(parse(text = lines[2]), envir = list2env(list(dt = by_hand)))
  expect_equal(by_app, by_hand$age_cat)
})

test_that("awkward variable names survive the round trip", {
  dt <- data.table::data.table(`my age` = as.numeric(1:100))
  cut <- gs_cut("my age", new = "my group", method = "breaks", breaks = 50)
  expect_length(gs_check_cut(dt, cut), 0L)
  out <- gs_apply_cuts(dt, list(cut))
  expect_true("my group" %in% names(out))
  expect_equal(as.integer(table(out[["my group"]])), c(50L, 50L))
})

test_that("the code creates the derived columns before it uses them", {
  spec <- gs_spec(plot_type = "Boxplot", x = "sex", y = "los_days",
                  cuts = list(gs_cut("age", method = "breaks", breaks = 65)),
                  strat_vars = "age_cat", data_name = "epi_cohort")
  row <- data.table::data.table(var = "age_cat", level = "(65, Inf]",
                                label = "age_cat: (65, Inf]",
                                file = "age_cat_65_Inf", n = 200L, keep = TRUE)
  script <- gs_code_script(spec, row)

  expect_silent(parse(text = script))
  expect_match(script, "dt[, age_cat := cut(age, breaks = c(-Inf, 65, Inf))]",
               fixed = TRUE)
  # Created straight after the data is read, and before the subset reads it.
  expect_lt(regexpr("age_cat :=", script, fixed = TRUE),
            regexpr("d <- dt[", script, fixed = TRUE))
  # ...and the whole thing runs, which is the point of the ordering.
  env <- new.env(parent = globalenv())
  assign("epi_cohort", epi_cohort, envir = env)
  expect_s3_class(eval(parse(text = script), envir = env), "ggplot")
})

test_that("gs_spec keeps the rules it is handed", {
  # modifyList() would quietly drop an unnamed list of rules.
  spec <- gs_spec(cuts = list(gs_cut("age"), gs_cut("bmi")))
  expect_length(spec$cuts, 2L)
  expect_equal(spec$cuts[[2]]$new, "bmi_cat")

  # A bare rule, not wrapped in a list, is understood too.
  expect_length(gs_spec(cuts = gs_cut("age"))$cuts, 1L)
  expect_length(gs_spec()$cuts, 0L)
})

test_that("gs_cut_name avoids names the data already uses", {
  expect_equal(gs_cut_name("age", c("bmi")), "age_cat")
  expect_equal(gs_cut_name("age", c("age_cat")), "age_cat2")
  expect_equal(gs_cut_name("age", c("age_cat", "age_cat2")), "age_cat3")
})

test_that("a rule describes itself for the sidebar", {
  expect_equal(gs_cut_describe(gs_cut("age", n = 4)),
               "age_cat <- age (4 quantile groups)")
  expect_equal(gs_cut_describe(gs_cut("age", method = "equal", n = 3)),
               "age_cat <- age (3 equal-width bins)")
  expect_equal(gs_cut_describe(gs_cut("age", method = "breaks",
                                      breaks = c(50, 65))),
               "age_cat <- age (cut at 50, 65)")
})

test_that("a default cut name never returns a name already taken", {
  taken <- c("x_cat", paste0("x_cat", 2:150))
  nm <- gs_cut_name("x", taken)
  expect_false(nm %in% taken)
})

test_that("the group-count bounds hold for a cut built without gs_cut()", {
  # gs_cut() floors n at 2, so the lower bound cannot be reached through it --
  # see the constructor test above. A cut assembled as a plain list skips that
  # flooring, and gs_check_cut() is what still catches it.
  dt <- data.table::data.table(age = as.numeric(1:100))
  raw <- list(var = "age", new = "age_cat", method = "quantile", n = 1L,
              breaks = numeric())
  expect_match(gs_check_cut(dt, raw), "between 2 and 20")
  expect_match(gs_check_cut(dt, gs_cut("age", n = 25)), "between 2 and 20")
  expect_equal(gs_check_cut(dt, gs_cut("age", n = 4)), character())
})

test_that("a cut that cannot be applied is reported, not thrown", {
  dt <- data.table::data.table(x = rep(1, 50))
  out <- gs_apply_cuts(dt, list(gs_cut("x", new = "x_cat", n = 4)))
  expect_true(length(attr(out, "gs_cut_error")) > 0L)
  expect_false("x_cat" %in% names(out))
})

# A "spec" is the single plain-list description of what the user asked for.
# The preview, the in-app PNG export and the generated R code are all derived
# from it, so there is exactly one place where plot semantics live.

# The plot types offered by ggplotgui, kept identical on purpose, plus the
# Kaplan-Meier curve, which ggplotgui does not offer and which descriptive
# work on clinical cohorts needs constantly.
GS_PLOT_TYPES <- c("Boxplot", "Density", "Dot + Error", "Dotplot",
                   "Histogram", "Kaplan-Meier curve", "Line", "Scatter",
                   "Violin")

# Spelled once each, because they are tested against all over the package.
GS_KM <- "Kaplan-Meier curve"
GS_LINE <- "Line"

GS_THEMES <- c(
  "bw"        = "theme_bw()",
  "classic"   = "theme_classic()",
  "dark"      = "theme_dark()",
  "grey"      = "theme_grey()",
  "light"     = "theme_light()",
  "line_draw" = "theme_linedraw()",
  "minimal"   = "theme_minimal()"
)

# RColorBrewer palette names, grouped as ggplotgui groups them. Written out
# rather than queried from RColorBrewer so that it is not a dependency; the
# palettes themselves are applied by ggplot2's scale_*_brewer().
# The shape is deliberately not uniform: selectInput() reads a bare string as
# one option and a list as an <optgroup>, so wrapping the default in list()
# would turn it into a group of one nameless entry rather than the plain first
# choice it has to be.
GS_PALETTES <- list(
  "Default (ggplot2)" = "",
  Qualitative = as.list(c("Accent", "Dark2", "Paired", "Pastel1", "Pastel2",
                          "Set1", "Set2", "Set3")),
  Diverging = as.list(c("BrBG", "PiYG", "PRGn", "PuOr", "RdBu", "RdGy",
                        "RdYlBu", "RdYlGn", "Spectral")),
  Sequential = as.list(c("Blues", "BuGn", "BuPu", "GnBu", "Greens", "Greys",
                         "Oranges", "OrRd", "PuBu", "PuBuGn", "PuRd",
                         "Purples", "RdPu", "Reds", "YlGn", "YlGnBu",
                         "YlOrBr", "YlOrRd"))
)

# Plot types that describe the distribution of a single continuous variable
# on the x axis; they have no y variable.
GS_XONLY_TYPES <- c("Density", "Histogram")

# Plot types that put one variable on each axis and need both.
GS_XY_TYPES <- c("Line", "Scatter")

# Plot types a LOWESS smoother can be laid over. Both draw y against x with
# nothing summarised in between, which is what a smoother has to have.
GS_SMOOTH_TYPES <- c("Line", "Scatter")

# The ways a continuous variable can be turned into a categorical one.
GS_CUT_METHODS <- c(
  "Quantiles (equal-sized groups)" = "quantile",
  "Equal-width bins"               = "equal",
  "Custom cut points"              = "breaks"
)

# The file formats a figure can be written in. PNG goes through ragg; SVG is a
# vector format, so a figure stays sharp at any size and can still be edited in
# Illustrator or Inkscape after the fact -- which is what a journal usually
# asks for.
GS_FORMATS <- c("PNG (raster)" = "png", "SVG (vector)" = "svg")

#' Build a spec with defaults filled in
#' @keywords internal
#' @noRd
gs_spec <- function(...) {
  defaults <- list(
    plot_type   = "Boxplot",
    x           = "",
    y           = "",
    time        = "",             # Kaplan-Meier curve: follow-up time
    event       = "",             # Kaplan-Meier curve: 1 = event, 0 = censored
    id          = "",             # Line: one line per level of this variable
    group       = "",
    # Whether the grouping variable is continuous, which decides between the
    # discrete and the gradient form of the palette scale. Set by the server
    # from the classified data; FALSE is the safe default for a spec built by
    # hand, because a discrete scale is what a categorical group needs.
    group_continuous = FALSE,
    facet       = "",             # the facet_wrap layer: one figure, many panels
    strat_vars  = character(),    # the outer layers: one figure each
    strat_mode  = "independent",  # "independent" | "crossed"
    cuts        = list(),         # categorized continuous variables
    min_n       = 10L,
    show_n      = TRUE,
    jitter      = FALSE,
    line_points = FALSE,          # Line: draw the observations as well
    smooth      = FALSE,          # LOWESS smoother, see GS_SMOOTH_TYPES
    smooth_se   = FALSE,
    smooth_span = 0.75,
    binwidth    = NA_real_,
    alpha       = 0.6,
    bw_adjust   = 1,
    km_ci       = FALSE,
    km_censor   = TRUE,
    km_ylim     = TRUE,
    km_risk     = FALSE,          # the number-at-risk table under the curve
    # The axis ranges, as they were typed. NA at either end is "as far as the
    # data goes", which is what ggplot2 does when it is left to decide.
    xlim_min    = NA_real_,
    xlim_max    = NA_real_,
    ylim_min    = NA_real_,
    ylim_max    = NA_real_,
    theme       = "theme_bw()",
    palette     = "",
    title       = "",
    lab_x       = "",
    lab_y       = "",
    lab_legend  = "",
    data_name   = "mydata",
    outdir      = "figures",
    prefix      = "",
    format      = "png",          # see GS_FORMATS
    width       = 7,
    height      = 5,
    dpi         = 300
  )
  args <- list(...)
  # modifyList() merges two lists element by element, which is right for every
  # field except `cuts`, itself a list: it would try to merge the rules with
  # the default empty list and, finding no names to match, drop them.
  cuts <- if ("cuts" %in% names(args)) args$cuts else defaults$cuts
  args$cuts <- NULL

  spec <- utils::modifyList(defaults, args)
  spec$strat_vars <- as.character(spec$strat_vars)
  spec$cuts <- gs_as_cuts(cuts)
  gs_normalize_spec(spec)
}

# --- the layers --------------------------------------------------------------

# How the outer layers combine. Independent is the default because it is what
# descriptive work usually wants: "what does this look like within each level
# of each variable", not "of every combination".
GS_STRAT_MODES <- c(
  "Each variable separately" = "independent",
  "Every combination"        = "crossed"
)

#' Every variable that splits the data into panels or figures
#'
#' These are the "layers": the facet variable makes panels inside one figure,
#' the stratifying variables make separate figures. A row that is missing any
#' of them cannot be placed, which is why they are handled together.
#' @keywords internal
#' @noRd
gs_layer_vars <- function(spec) {
  v <- c(spec$strat_vars, spec$facet)
  unique(v[nzchar(v)])
}

#' How many layers a spec describes, the described variable included
#'
#' The described variable, plus one for each variable that panels or splits
#' it. What a given count looks like is the user's arrangement, not this
#' number: two layers is a `facet_wrap()` when the second variable is the
#' facet, and one figure per level when it is a stratifying variable instead.
#' @keywords internal
#' @noRd
gs_n_layers <- function(spec) 1L + length(gs_layer_vars(spec))

#' Fold plot-type quirks into the spec once, so nothing downstream repeats them
#'
#' As in ggplotgui, density and histogram take their continuous variable from
#' the Y-variable selector and draw it on the x axis; there is no y variable.
#' The Y-variable always wins here, because the UI hides the X selector for
#' these plot types and its last value would otherwise linger: typically a
#' categorical column that `stat_bin()` cannot use.
#'
#' A Kaplan-Meier curve takes its axes from the time and event variables
#' instead, so any lingering X/Y selection is dropped for the same reason.
#'
#' The options that belong to one plot type -- the line ID, the markers, the
#' smoother -- are switched off for every other type, so that a control the
#' user cannot currently see can never reach the generated code.
#' @keywords internal
#' @noRd
gs_normalize_spec <- function(spec) {
  if (spec$plot_type %in% GS_XONLY_TYPES) {
    if (nzchar(spec$y)) spec$x <- spec$y
    spec$y <- ""
  }
  if (identical(spec$plot_type, GS_KM)) {
    spec$x <- ""
    spec$y <- ""
  } else {
    spec$km_risk <- FALSE
  }
  if (!identical(spec$plot_type, GS_LINE)) {
    spec$id <- ""
    spec$line_points <- FALSE
  }
  if (!spec$plot_type %in% GS_SMOOTH_TYPES) spec$smooth <- FALSE
  if (!spec$strat_mode %in% GS_STRAT_MODES) spec$strat_mode <- "independent"
  if (!spec$format %in% GS_FORMATS) spec$format <- "png"
  # Nothing is grouped, so nothing is on a continuous scale either.
  if (!nzchar(spec$group)) spec$group_continuous <- FALSE
  # A variable can be one layer or another, never both: faceting by it and
  # then splitting the figures by it too would leave one panel per figure.
  if (nzchar(spec$facet)) {
    spec$strat_vars <- setdiff(spec$strat_vars, spec$facet)
  }
  spec
}

#' Turn the Shiny inputs into a spec
#' @keywords internal
#' @noRd
gs_spec_from_input <- function(input) {
  none <- function(x) if (is.null(x) || identical(x, "") || identical(x, GS_NONE)) "" else x
  gs_spec(
    plot_type  = input$plot_type %||% "Boxplot",
    x          = none(input$xvar),
    y          = none(input$yvar),
    time       = none(input$timevar),
    event      = none(input$eventvar),
    id         = none(input$idvar),
    group      = none(input$group),
    facet      = none(input$facet),
    strat_vars = input$strat_vars %||% character(),
    strat_mode = input$strat_mode %||% "independent",
    # Clearing the numeric box yields NA, which must not poison the size rules.
    min_n      = as.integer(gs_num(input$min_n, 10)),
    # NULL before the checkbox has registered; its UI default is TRUE.
    show_n     = !identical(input$show_n, FALSE),
    jitter     = isTRUE(input$jitter),
    line_points = isTRUE(input$line_points),
    smooth     = isTRUE(input$smooth),
    smooth_se  = isTRUE(input$smooth_se),
    smooth_span = gs_num(input$smooth_span, 0.75),
    binwidth   = gs_num(input$binwidth),
    alpha      = gs_num(input$alpha, 0.6),
    bw_adjust  = gs_num(input$bw_adjust, 1),
    km_ci      = isTRUE(input$km_ci),
    # NULL before the checkboxes have registered; both default to TRUE.
    km_censor  = !identical(input$km_censor, FALSE),
    km_ylim    = !identical(input$km_ylim, FALSE),
    km_risk    = isTRUE(input$km_risk),
    # An empty box is NA, which is the "let ggplot2 decide" the default is.
    xlim_min   = gs_num(input$xlim_min),
    xlim_max   = gs_num(input$xlim_max),
    ylim_min   = gs_num(input$ylim_min),
    ylim_max   = gs_num(input$ylim_max),
    theme      = input$theme %||% "theme_bw()",
    palette    = none(input$palette),
    title      = input$title %||% "",
    lab_x      = input$lab_x %||% "",
    lab_y      = input$lab_y %||% "",
    lab_legend = input$lab_legend %||% "",
    data_name  = input$data_name %||% "mydata",
    # A folder of nothing but spaces would pass nzchar() and then fail to be
    # created, so it is trimmed before it can reach dir.create().
    outdir     = if (nzchar(trimws(input$outdir %||% ""))) {
                   trimws(input$outdir)
                 } else {
                   "figures"
                 },
    prefix     = input$prefix %||% "",
    format     = input$format %||% "png",
    width      = gs_num(input$width, 7),
    height     = gs_num(input$height, 5),
    dpi        = gs_num(input$dpi, 300)
  )
}

#' Columns a spec actually reads from the data
#' @keywords internal
#' @noRd
gs_spec_cols <- function(spec) {
  # `id` is blank for every type but Line, so it needs no test of its own.
  cols <- c(spec$x, spec$y, spec$group, spec$facet, spec$id)
  if (identical(spec$plot_type, GS_KM)) cols <- c(spec$time, spec$event, cols)
  unique(cols[nzchar(cols)])
}

#' Check that a spec can be drawn; returns a character vector of problems
#' @keywords internal
#' @noRd
gs_validate_spec <- function(spec, info = NULL) {
  problems <- character()
  type <- spec$plot_type

  # A selected variable can vanish under the app's feet: removing a
  # categorization removes its column. Saying so beats letting the plot code
  # fail on a missing object, and nothing else is worth reporting until the
  # selection is valid again.
  if (!is.null(info) && nrow(info)) {
    gone <- setdiff(gs_spec_cols(spec), gs_vars_of(info, "all"))
    if (length(gone)) {
      return(sprintf("'%s' is no longer a column in the data.", gone))
    }
  }

  # Checked for every plot type, and before the Kaplan-Meier branch returns.
  # An inverted range is silently destructive rather than loud: coord_cartesian
  # draws an empty panel, and on a survival curve the number-at-risk times are
  # the ones inside the range, so a backwards range leaves the table with no
  # columns at all.
  problems <- c(problems, gs_validate_limits(spec))

  if (type == GS_KM) {
    if (!nzchar(spec$time)) {
      problems <- c(problems, sprintf("A %s needs a time variable.", GS_KM))
    }
    if (!nzchar(spec$event)) {
      problems <- c(problems, sprintf(
        "A %s needs an event variable (1 = event, 0 = censored).", GS_KM))
    }
    if (!is.null(info) && nrow(info)) {
      problems <- c(problems, gs_validate_km_vars(spec, info))
    }
    return(problems)
  }

  if (type %in% GS_XONLY_TYPES) {
    if (!nzchar(spec$x)) {
      problems <- c(problems, sprintf(
        "%s needs a continuous variable (choose one under Y-variable).", type))
    }
  } else if (type %in% GS_XY_TYPES) {
    if (!nzchar(spec$x) || !nzchar(spec$y)) {
      problems <- c(problems,
                    sprintf("%s needs both an X- and a Y-variable.", type))
    }
  } else {
    if (!nzchar(spec$y)) {
      problems <- c(problems, sprintf("%s needs a Y-variable (continuous).", type))
    }
  }

  if (!is.null(info) && nrow(info)) {
    continuous <- gs_vars_of(info, "continuous")
    needs_cont <- if (type %in% GS_XONLY_TYPES) spec$x else spec$y
    if (nzchar(needs_cont) && !needs_cont %in% continuous) {
      problems <- c(problems, sprintf(
        "'%s' does not look continuous; %s expects a continuous variable there.",
        needs_cont, type))
    }
    # Only Scatter is held to a continuous X. A line plot's X is time, which
    # is as often a date or a visit number -- neither of which is classified
    # as continuous -- as it is a measurement.
    if (type == "Scatter" && nzchar(spec$x) && !spec$x %in% continuous) {
      problems <- c(problems,
                    sprintf("Scatter expects a continuous X-variable; '%s' is not.", spec$x))
    }
  }

  problems
}

#' Axis ranges that a figure cannot be drawn in
#'
#' Both ends of a range are optional -- one alone means "from here" or "up to
#' here" -- so the check only applies when both are given. Equal ends are
#' rejected alongside reversed ones: a range of zero width is as undrawable as
#' a backwards one.
#' @keywords internal
#' @noRd
gs_validate_limits <- function(spec) {
  one <- function(lo, hi, axis) {
    lo <- gs_num(lo)
    hi <- gs_num(hi)
    if (is.na(lo) || is.na(hi) || lo < hi) return(character())
    sprintf(paste("The %s-axis range runs backwards: %s is not below %s.",
                  "Swap them, or clear one to leave that end free."),
            axis, gs_n(lo), gs_n(hi))
  }
  c(one(spec$xlim_min, spec$xlim_max, "X"),
    one(spec$ylim_min, spec$ylim_max, "Y"))
}

#' The type checks that only a Kaplan-Meier curve needs
#'
#' `survival::Surv()` refuses a non-numeric time and an event that is not
#' 0/1, 1/2 or logical, so both are checked here rather than left to surface
#' as an error from inside the fit.
#' @keywords internal
#' @noRd
gs_validate_km_vars <- function(spec, info) {
  problems <- character()
  numeric_vars <- gs_vars_of(info, "numeric")
  event_vars <- gs_vars_of(info, "event")

  if (nzchar(spec$time) && !spec$time %in% numeric_vars) {
    problems <- c(problems, sprintf(
      "The time variable must be numeric; '%s' is not.", spec$time))
  }
  if (nzchar(spec$event) && !spec$event %in% event_vars) {
    problems <- c(problems, sprintf(
      paste("The event variable must be 0/1, 1/2 or TRUE/FALSE; '%s' is not.",
            "Recode it before plotting."),
      spec$event))
  }
  problems
}

# --- small helpers -----------------------------------------------------------

`%||%` <- function(x, y) if (is.null(x)) y else x

GS_NONE <- "(none)"

#' Coerce an input to a finite number, falling back to a default
#' @keywords internal
#' @noRd
gs_num <- function(x, default = NA_real_) {
  if (is.null(x) || length(x) != 1L) return(default)
  x <- suppressWarnings(as.numeric(x))
  if (is.na(x) || !is.finite(x)) default else x
}

test_that("gs_spec fills defaults and keeps overrides", {
  spec <- gs_spec(plot_type = "Violin", y = "bmi")
  expect_equal(spec$plot_type, "Violin")
  expect_equal(spec$y, "bmi")
  expect_equal(spec$theme, "theme_bw()")
  expect_equal(spec$outdir, "figures")
  expect_equal(spec$strat_vars, character())
})

test_that("gs_spec_cols lists only the columns the plot reads", {
  spec <- gs_spec(plot_type = "Boxplot", x = "arm", y = "bmi",
                  group = "arm", facet = "site", strat_vars = "sex")
  # Deduplicated, and the stratifying variable is handled separately.
  expect_setequal(gs_spec_cols(spec), c("arm", "bmi", "site"))
})

test_that("validation catches missing variables per plot type", {
  expect_match(gs_validate_spec(gs_spec(plot_type = "Boxplot")),
               "needs a Y-variable")
  expect_match(gs_validate_spec(gs_spec(plot_type = "Density")),
               "needs a continuous variable")
  expect_match(gs_validate_spec(gs_spec(plot_type = "Scatter", x = "a")),
               "both an X- and a Y-variable")

  expect_length(gs_validate_spec(gs_spec(plot_type = "Boxplot", y = "bmi")), 0L)
})

test_that("validation warns when a categorical variable is used as continuous", {
  info <- gs_classify_vars(gs_prepare_data(iris))
  ok <- gs_validate_spec(gs_spec(plot_type = "Boxplot", x = "Species",
                                 y = "Sepal.Length"), info)
  expect_length(ok, 0L)

  bad <- gs_validate_spec(gs_spec(plot_type = "Boxplot", y = "Species"), info)
  expect_match(bad, "does not look continuous")
})

test_that("gs_num falls back on unusable input", {
  expect_equal(gs_num(3), 3)
  expect_equal(gs_num("2.5"), 2.5)
  expect_equal(gs_num(NULL, 7), 7)
  expect_equal(gs_num("abc", 7), 7)
  expect_equal(gs_num(NA, 7), 7)
})

test_that("an axis range that runs backwards is reported", {
  spec <- gs_spec(plot_type = "Boxplot", y = "a", xlim_min = 100,
                  xlim_max = 10)
  expect_match(gs_validate_spec(spec), "X-axis range runs backwards", all = FALSE)

  spec2 <- gs_spec(plot_type = "Boxplot", y = "a", ylim_min = 5, ylim_max = 5)
  expect_match(gs_validate_spec(spec2), "Y-axis range runs backwards", all = FALSE)

  # One end alone means "from here"/"up to here" and is not a problem.
  spec3 <- gs_spec(plot_type = "Boxplot", y = "a", xlim_min = 10)
  expect_false(any(grepl("runs backwards", gs_validate_spec(spec3))))

  spec4 <- gs_spec(plot_type = "Boxplot", y = "a", xlim_min = 10, xlim_max = 100)
  expect_false(any(grepl("runs backwards", gs_validate_spec(spec4))))
})

# The one exported entry point. Argument checking is done with checkmate so
# that a mistake is reported here, in one clear message, rather than surfacing
# as a broken reactive several layers into the Shiny app.

#' Describe a data set, one layer at a time
#'
#' Launches a point-and-click Shiny interface for descriptive analysis: pick
#' the variable you want to describe, then add the layers you want to see it
#' within. The figure, the number of observations behind it and the R code
#' that reproduces it all appear together.
#'
#' @section The layers:
#'
#' A description has a variable being described and, around it, the variables
#' you want to see it within. `ggstratify` calls those the layers, and asks
#' you to place each one:
#'
#' \describe{
#'   \item{One layer}{The variable on its own. A plain `ggplot2` figure.}
#'   \item{Two layers}{The second variable becomes the panels of a
#'     `facet_wrap()`, so the whole comparison is one figure.}
#'   \item{Three or more}{Two layers is what a single figure holds. Beyond
#'     that, you choose: one variable stays as the `facet_wrap()` panels, and
#'     the rest become separate figures -- one file each. With columns A, B, C
#'     and D you might describe D, panel it by C, and get one figure per
#'     level of A and of B.}
#' }
#'
#' The variables that make separate figures can be treated separately, which
#' is the default -- ticking `sex` and `treatment` gives the figures `sex_M`,
#' `sex_F`, `treatment_A`, ... -- or crossed, giving one figure per observed
#' combination. Separately is usually the right question for descriptive
#' work: crossing two five-level variables gives twenty-five mostly empty
#' figures.
#'
#' @section What it tells you:
#'
#' Every level is reported with the number of observations it contains, and
#' that number is written into the figure title and onto each panel strip
#' inside it: a strip reads `site: Site A (N = 303)` rather than `Site A`, so
#' that a level which does not explain itself -- the bare range a categorized
#' variable produces -- still says which variable it is a level of. A level
#' with no observations at all (an unused factor level) is listed with
#' `N = 0` and produces no figure. Rows with a missing value in any layer
#' variable are excluded -- a row that does not say which panel it belongs to
#' cannot be drawn in one -- and the number excluded is reported on the Strata
#' tab and by the generated code.
#'
#' @section Three things descriptive work keeps needing:
#'
#' Under **Categorize**, a continuous variable becomes a categorical one --
#' quantile groups, equal-width bins or your own cut points -- and can then be
#' used as a layer like any other categorical variable.
#'
#' Under **Type of graph**, *Kaplan-Meier curve* draws survival curves from a
#' time variable and an event indicator, fitted with [survival::survfit()]
#' within each group, panel and figure, optionally with a confidence band,
#' censoring marks and a number-at-risk table under the curve. *Line* draws
#' change over time: put time on the X axis and the measurement on the Y axis,
#' and name the subject identifier under **One line per** to get one line per
#' subject.
#'
#' A *Line* or *Scatter* figure can carry a LOWESS smoother -- `stats::loess()`
#' through [ggplot2::geom_smooth()] -- with its span under your control and,
#' when a grouping variable is set, one fit per group.
#'
#' @section Before you start:
#'
#' Convert each column to the type you mean it to have -- `numeric` for
#' measurements, `factor` for groups, with the levels in the order you want
#' them read -- before passing the data. `ggstratify` describes what it is
#' given; it does not guess what you meant. A grouping variable stored as
#' `1, 2, 3` will be described as a number.
#'
#' This is why the data must be an object you already have in your session.
#' There is no file to choose from inside the application, and no file path to
#' hand it: a data set that arrives by being read from disk arrives with types
#' that a reader guessed, and it is then described on those guessed types
#' without anyone having looked at them. Read the file yourself, run
#' [str()] or [summary()] over the result, fix the types that are wrong, and
#' pass that object:
#'
#' ```
#' cohort <- read.csv("cohort.csv")
#' cohort$treatment <- factor(cohort$treatment,
#'                            levels = c("Control", "Low dose", "High dose"))
#' str(cohort)
#' ggstratify(cohort)
#' ```
#'
#' The figure types, themes and palettes otherwise match those offered by the
#' \pkg{ggplotgui} package. All data handling uses \pkg{data.table}. Figures
#' are written as PNG (through \pkg{ragg}) or as SVG, and the "R-code" tab
#' shows the self-contained \pkg{ggplot2} code for the figure on screen --
#' just the figure, since writing the files is what the export button is for.
#' The preview, the export button and that code are all produced by the same
#' code generator, so the code you are shown is the code that made the
#' figure.
#'
#' @param dataset The data to describe, and the one thing the function needs:
#'   a data frame -- including a `tibble` or a `data.table` -- or a matrix,
#'   already read into your session and already of the types you mean. A file
#'   path is not accepted; see *Before you start*. Whatever it is given becomes
#'   a plain `data.table`, so a grouped `tibble` is described as its rows, not
#'   as its groups. The object is copied, never modified in place, and its name
#'   is what the generated code refers to it by -- so pass a named object,
#'   rather than an expression: `ggstratify(head(cohort))` is described
#'   perfectly well, but the code it prints says `d <- mydata`, because
#'   `head(cohort)` is not a name.
#' @param launch.browser Passed to [shiny::runApp()]. `TRUE` opens the system
#'   browser.
#' @param ... Further arguments passed to [shiny::runApp()].
#'
#' @return Invisibly `NULL`, after the app is closed. Called for its side
#'   effect of running a Shiny application.
#'
#' @examples
#' if (interactive()) {
#'   ggstratify(iris)
#'   ggstratify(epi_cohort)
#' }
#' @export
ggstratify <- function(dataset, launch.browser = TRUE, ...) {
  # Read before `dataset` is forced, so that the name in the generated code is
  # the expression the user wrote rather than the value it evaluated to.
  data_name <- gs_data_name(substitute(dataset))
  checkmate::assert_flag(launch.browser)
  checkmate::assert(
    checkmate::check_data_frame(dataset, min.cols = 1L),
    checkmate::check_matrix(dataset, min.cols = 1L),
    .var.name = "dataset",
    combine = "or"
  )
  # Fail here rather than inside the app, where the message is easy to miss.
  dataset <- gs_prepare_data(dataset)

  app <- shiny::shinyApp(
    ui = gs_ui(),
    server = gs_server(dataset, data_name),
    # The in-app preview only. renderPlot() draws through whatever device this
    # option names, and ragg is a markedly faster raster device than the
    # default on Windows. The exported files do not depend on it: gs_device()
    # hands ggsave() ragg::agg_png or an SVG device by name.
    onStart = function() {
      old <- options(shiny.useragg = TRUE)
      shiny::onStop(function() options(old))
    }
  )

  shiny::runApp(app, launch.browser = launch.browser, ...)
  invisible(NULL)
}

#' The name the generated code should call the data by
#'
#' `ggstratify(cohort)` should print code that starts `d <- cohort`, not
#' `d <- mydata`. The rule is deliberately narrow, and stated here because the
#' generated code is the visible half of it: the argument is used as written
#' only when it is a syntactic name, one `make.names()` leaves alone.
#' Everything else falls back to the placeholder the Export panel starts with,
#' which the user can overwrite. Three kinds of argument fall back:
#'
#'   * an expression -- `ggstratify(read.csv("x.csv"))`, a subset, a pipeline,
#'     a literal. These are accepted as data, but they are values rather than
#'     names, and a reader cannot paste one at the top of a script and get the
#'     same object back.
#'   * a quoted non-syntactic name -- ``ggstratify(`my data`)`` -- which would
#'     have to be written back with its backticks to parse.
#'   * a dot name -- `.`, `...`, `..1` -- which is a name only inside the call
#'     that supplied it, and means nothing in a script pasted elsewhere.
#'
#' @param expr The unevaluated `dataset` argument, from `substitute()`.
#' @return A single syntactic name.
#' @keywords internal
#' @noRd
gs_data_name <- function(expr) {
  if (!is.name(expr)) return("mydata")
  nm <- as.character(expr)
  # nzchar() guards the empty name, which is a name but never one that prints.
  if (!nzchar(nm) || !identical(make.names(nm), nm)) return("mydata")
  if (grepl("^([.]|[.]{3}|[.][.][0-9]+)$", nm)) return("mydata")
  nm
}

# The launcher must reject bad input before a Shiny app is ever started, so
# the user gets one clear message instead of a broken reactive.

test_that("ggstratify rejects input that is not tabular", {
  expect_error(ggstratify(1:10), "dataset")
  expect_error(ggstratify(data.frame()), "dataset")
  expect_error(ggstratify(NA), "dataset")
})

test_that("ggstratify takes an object, never a path to one", {
  # A file read here would be typed by a guess, and then described on that
  # guess. The user reads the file, checks the types, and passes the result.
  path <- tempfile(fileext = ".csv")
  on.exit(unlink(path), add = TRUE)
  utils::write.csv(iris, path, row.names = FALSE)
  expect_error(ggstratify(path), "dataset")

  # And the app has no way to load one either: nothing in the UI is a file
  # control, so there is no second route back to guessed types.
  ui_src <- paste(deparse(gs_ui), collapse = " ")
  expect_false(grepl("fileInput", ui_src, fixed = TRUE))
  expect_false(grepl("data_file", ui_src, fixed = TRUE))
})

test_that("the data set is required", {
  expect_error(ggstratify(), "dataset")
})

test_that("ggstratify rejects a non-flag launch.browser", {
  expect_error(ggstratify(iris, launch.browser = "yes"), "launch.browser")
})

test_that("the generated code calls the data by the name it was passed under", {
  # ggstratify(cohort) should print `d <- cohort`, not `d <- mydata`.
  cohort <- iris
  expect_equal(gs_data_name(quote(cohort)), "cohort")
  # Anything that is not a plain name is not something to paste into a script.
  expect_equal(gs_data_name(quote(subset(iris, Species == "setosa"))), "mydata")
  expect_equal(gs_data_name(quote(`odd name`)), "mydata")
  # Dot names are names, and syntactic ones, but they mean nothing outside the
  # call that supplied them.
  expect_equal(gs_data_name(quote(.)), "mydata")
  expect_equal(gs_data_name(quote(...)), "mydata")
  expect_equal(gs_data_name(quote(..1)), "mydata")
  # A leading dot is otherwise an ordinary name and stays one.
  expect_equal(gs_data_name(quote(.cohort)), ".cohort")
})

test_that("there is exactly one exported function to learn", {
  ns <- readLines(system.file("NAMESPACE", package = "ggstratify"))
  expect_equal(grep("^export\\(", ns, value = TRUE), "export(ggstratify)")
  # The old second entry point is gone, not merely unexported.
  expect_false(exists("ggstratify_classic", envir = asNamespace("ggstratify"),
                      inherits = FALSE))
})

test_that("the package does not depend on ggplotgui", {
  deps <- unlist(strsplit(
    paste(utils::packageDescription("ggstratify",
                                    fields = c("Depends", "Imports", "LinkingTo")),
          collapse = ", "),
    ",\\s*"
  ))
  expect_false(any(grepl("ggplotgui", deps)))
})

test_that("every namespace import is declared and every declared one is used", {
  ns <- readLines(system.file("NAMESPACE", package = "ggstratify"))
  imported <- unique(c(
    sub("^import[(]([^)]*)[)]$", "\\1", grep("^import[(]", ns, value = TRUE)),
    sub("^importFrom[(]([^,]*),.*$", "\\1", grep("^importFrom[(]", ns, value = TRUE))
  ))
  desc <- read.dcf(system.file("DESCRIPTION", package = "ggstratify"))
  declared <- trimws(gsub("[(][^)]*[)]", "",
                          strsplit(desc[1, "Imports"], ",")[[1]]))
  # A namespace import that DESCRIPTION does not list is an R CMD check
  # warning. `tools` was imported here for two functions nothing called.
  expect_equal(setdiff(imported, declared), character())
})

---
title: "Getting started with ggstratify"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting started with ggstratify}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8} 
---

```{r, include = FALSE}
knitr::opts_chunk$set(collapse = TRUE, comment = "#>")
```

```{r setup}
library(ggstratify)
```

## What is ggstratify?

Descriptive analysis often means making the same figure repeatedly for
different subgroups. `ggstratify` makes this easier: choose the variables
you want to describe and stratify by, and the app creates the figures for
you. It also gives you the R code used to create each figure.

Everything runs locally. Your data are not uploaded anywhere.

## Before you start

`ggstratify` uses your variables as they are typed. Make sure categorical
variables are factors and that their levels are in the order you want.

```{r, eval = FALSE}
dat <- transform(
  dat,
  sex = factor(sex, levels = c("Male", "Female")),
  severity = factor(severity, levels = c("Mild", "Moderate", "Severe")),
  age = as.numeric(age)
)
```

The app takes a data object, not a file. Read your data first, check the
variables, and then pass the object to `ggstratify()`:

```{r, eval = FALSE}
cohort <- read.csv("cohort.csv")
str(cohort)
ggstratify(cohort)
```

Your original data are never modified.

## Launch the app

```{r, eval = FALSE}
ggstratify(epi_cohort)
```

`epi_cohort` is a simulated dataset included with the package.

```{r}
str(epi_cohort)
```

The same function works with any data frame, tibble, data.table, or matrix.

## Stratifying your figures

The **Layers** panel controls how figures are split.

* **No layer:** one figure.
* **One panel variable:** several panels in one figure.
* **Additional stratification variables:** separate figures.

For example, you could describe `D`, show panels by `C`, and create separate
figures by `A` and `B`.

When several stratification variables are selected, you can either:

* **Separate:** make figures for each variable independently.
* **Cross:** make a figure for every combination of their levels.

The **Strata** tab shows the figures that will be produced and their sample
sizes before you export them.

## Missing values and small strata

Rows missing a value for a stratification variable are excluded because they
cannot be assigned to a figure or panel. The number of excluded rows is shown
in the app and in the generated code.

Strata with fewer observations than the selected **Minimum N per figure** are
listed but not plotted. Set the minimum to `0` if you want to include small
strata. Empty factor levels are never plotted.

## Categorizing a continuous variable

The **Categorize** panel lets you turn a continuous variable into groups
using:

* **Quantiles**
* **Equal-width bins**
* **Custom cut points**

For example, you can create an age group at 65 years:

```{r, eval = FALSE}
dt[, age_cat := cut(age, breaks = c(-Inf, 65, Inf))]
```

The new variable can then be used like any other categorical variable.

## Figure types

`ggstratify` supports several common descriptive plots, including:

* Boxplots
* Histograms
* Bar plots
* Scatter plots
* Line plots
* Kaplan-Meier curves

For line plots, choose a time variable and measurement, and optionally
specify an ID to draw one line per subject. A LOWESS smoother can also be
added.

For Kaplan-Meier curves, select the follow-up time and event variables.
Optional confidence intervals, censoring marks, and risk tables are available.

All plots are calculated within the selected panels and strata.

## Changing the axis range

**Appearance** lets you set the X- and Y-axis ranges.

The range only zooms the figure; it does not remove observations from the
analysis. This means that summary statistics such as boxplot medians remain
unchanged.

## Generated R code

The **R-code** tab shows the code used to create the figure currently shown.

The code includes the necessary data preparation, stratification, and
`ggplot2` commands, so you can copy it into your own analysis.

For example:

```{r, eval = FALSE}
library(data.table)
library(ggplot2)

dt <- as.data.table(epi_cohort)

d <- dt[sex == "Male"]

p <- ggplot(d, aes(x = treatment, y = los_days)) +
  geom_boxplot() +
  theme_bw()

p
```

The preview, exported figures, and R-code tab all use the same code generator,
so the generated code reproduces the figure shown in the app.

## Exporting figures

Click **Export all figures** to save the figures listed in the **Strata** tab.

You can export:

* **PNG** for raster images
* **SVG** for vector graphics

The exported figures use the full dataset, even when a large dataset is
sampled for the on-screen preview.

## In short

The basic workflow is:

1. Prepare and check your data.
2. Run `ggstratify(your_data)`.
3. Choose the variable to describe.
4. Add panels or stratification variables if needed.
5. Adjust the appearance.
6. Check the **Strata** tab.
7. Copy the R code or export the figures if needed.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\docType{package}
\name{ggstratify-package}
\alias{ggstratify-package}
\title{ggstratify: Fast Stratified Descriptive Figures with a Point-and-Click GUI}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A point-and-click 'shiny' interface for the descriptive analysis that comes before any model is chosen. Pass a data frame, pick the variable to describe, and add the layers you want to see it within: a second variable becomes the panels of a 'ggplot2' facet_wrap(), and further variables become separate figures, one file each, taken either one variable at a time or crossed. Every stratum is reported with the number of observations behind it, on the figure and on each of its panels; strata that contain none are listed rather than dropped, and rows with a missing value in a layer variable are excluded and counted. A continuous variable can be categorized into quantile groups, equal-width bins or user-supplied cut points and then used as a layer; the figure types follow those offered by the 'ggplotgui' package and add the line plot for change over time, an optional LOWESS smoother, and the Kaplan-Meier curve estimated by 'survival', with an optional number-at-risk table. Columns are described as they are typed, so convert each to the type you mean first. Figures are written as PNG or SVG, and the application prints the 'ggplot2' code behind the figure on screen, so that a description can be repeated, shared or accounted for later. Everything runs locally, with no network access and no AI involved.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/AkiShiroshita/ggstratify}
  \item \url{https://akishiroshita.github.io/ggstratify/}
  \item Report bugs at \url{https://github.com/AkiShiroshita/ggstratify/issues}
}

}
\author{
\strong{Maintainer}: Akihiro Shiroshita \email{akihirokun8@gmail.com} [copyright holder]

Authors:
\itemize{
  \item Akihiro Shiroshita \email{akihirokun8@gmail.com} [copyright holder]
  \item Yuki Kataoka \email{youkiti@gmail.com}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{epi_cohort}
\alias{epi_cohort}
\title{A synthetic clinical cohort}
\format{
A data frame with 600 rows and 11 columns:
\describe{
\item{id}{Patient identifier, \code{"P0001"} to \code{"P0600"}.}
\item{age}{Age in years.}
\item{sex}{Factor: \code{"Male"}, \code{"Female"}.}
\item{site}{Factor: \code{"Site A"}, \code{"Site B"}, \code{"Site C"}, \code{"Site D"}.
\code{"Site D"} has no observations.}
\item{treatment}{Factor: \code{"Control"}, \code{"Low dose"}, \code{"High dose"}.}
\item{severity}{Factor: \code{"Mild"}, \code{"Moderate"}, \code{"Severe"}.}
\item{bmi}{Body mass index, kg/m^2.}
\item{crp}{C-reactive protein, mg/L.}
\item{los_days}{Length of stay in days.}
\item{fu_days}{Days of follow-up, to death or to censoring. Censoring is
by dropout or by the end of the study at 365 days.}
\item{death}{\code{1} if the patient died during follow-up, \code{0} if censored.}
}
}
\source{
Simulated by \code{data-raw/epi_cohort.R}.
}
\usage{
epi_cohort
}
\description{
Six hundred simulated patients, shaped like the descriptive tables that
motivate this package: a few continuous measurements crossed with several
categorical variables worth stratifying on.
}
\details{
Two features are deliberate. \code{site} declares a fourth level, \code{"Site D"},
that recruited nobody, so stratifying by \code{site} produces a stratum with
\code{n = 0}; the app lists it and draws no figure. \code{severity} has a small
\code{"Severe"} group -- 20 patients against 381 and 199 -- which is what the
minimum-n control is there to be tried on: it is drawn at the default
minimum of 10, and disappears from the figures, with a reason, as soon as
the minimum is raised past 20.

The data are simulated. They describe no real patients and support no
clinical conclusion.

\code{fu_days} and \code{death} make the data usable for a Kaplan-Meier curve, and
\code{age} for the categorization controls.
}
\examples{
str(epi_cohort)

# The empty stratum that the app reports with n = 0.
table(epi_cohort$site)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app_launch.R
\name{ggstratify}
\alias{ggstratify}
\title{Describe a data set, one layer at a time}
\usage{
ggstratify(dataset, launch.browser = TRUE, ...)
}
\arguments{
\item{dataset}{The data to describe, and the one thing the function needs:
a data frame -- including a \code{tibble} or a \code{data.table} -- or a matrix,
already read into your session and already of the types you mean. A file
path is not accepted; see \emph{Before you start}. Whatever it is given becomes
a plain \code{data.table}, so a grouped \code{tibble} is described as its rows, not
as its groups. The object is copied, never modified in place, and its name
is what the generated code refers to it by -- so pass a named object,
rather than an expression: \code{ggstratify(head(cohort))} is described
perfectly well, but the code it prints says \code{d <- mydata}, because
\code{head(cohort)} is not a name.}

\item{launch.browser}{Passed to \code{\link[shiny:runApp]{shiny::runApp()}}. \code{TRUE} opens the system
browser.}

\item{...}{Further arguments passed to \code{\link[shiny:runApp]{shiny::runApp()}}.}
}
\value{
Invisibly \code{NULL}, after the app is closed. Called for its side
effect of running a Shiny application.
}
\description{
Launches a point-and-click Shiny interface for descriptive analysis: pick
the variable you want to describe, then add the layers you want to see it
within. The figure, the number of observations behind it and the R code
that reproduces it all appear together.
}
\section{The layers}{


A description has a variable being described and, around it, the variables
you want to see it within. \code{ggstratify} calls those the layers, and asks
you to place each one:

\describe{
\item{One layer}{The variable on its own. A plain \code{ggplot2} figure.}
\item{Two layers}{The second variable becomes the panels of a
\code{facet_wrap()}, so the whole comparison is one figure.}
\item{Three or more}{Two layers is what a single figure holds. Beyond
that, you choose: one variable stays as the \code{facet_wrap()} panels, and
the rest become separate figures -- one file each. With columns A, B, C
and D you might describe D, panel it by C, and get one figure per
level of A and of B.}
}

The variables that make separate figures can be treated separately, which
is the default -- ticking \code{sex} and \code{treatment} gives the figures \code{sex_M},
\code{sex_F}, \code{treatment_A}, ... -- or crossed, giving one figure per observed
combination. Separately is usually the right question for descriptive
work: crossing two five-level variables gives twenty-five mostly empty
figures.
}

\section{What it tells you}{


Every level is reported with the number of observations it contains, and
that number is written into the figure title and onto each panel strip
inside it: a strip reads \verb{site: Site A (N = 303)} rather than \verb{Site A}, so
that a level which does not explain itself -- the bare range a categorized
variable produces -- still says which variable it is a level of. A level
with no observations at all (an unused factor level) is listed with
\code{N = 0} and produces no figure. Rows with a missing value in any layer
variable are excluded -- a row that does not say which panel it belongs to
cannot be drawn in one -- and the number excluded is reported on the Strata
tab and by the generated code.
}

\section{Three things descriptive work keeps needing}{


Under \strong{Categorize}, a continuous variable becomes a categorical one --
quantile groups, equal-width bins or your own cut points -- and can then be
used as a layer like any other categorical variable.

Under \strong{Type of graph}, \emph{Kaplan-Meier curve} draws survival curves from a
time variable and an event indicator, fitted with \code{\link[survival:survfit]{survival::survfit()}}
within each group, panel and figure, optionally with a confidence band,
censoring marks and a number-at-risk table under the curve. \emph{Line} draws
change over time: put time on the X axis and the measurement on the Y axis,
and name the subject identifier under \strong{One line per} to get one line per
subject.

A \emph{Line} or \emph{Scatter} figure can carry a LOWESS smoother -- \code{stats::loess()}
through \code{\link[ggplot2:geom_smooth]{ggplot2::geom_smooth()}} -- with its span under your control and,
when a grouping variable is set, one fit per group.
}

\section{Before you start}{


Convert each column to the type you mean it to have -- \code{numeric} for
measurements, \code{factor} for groups, with the levels in the order you want
them read -- before passing the data. \code{ggstratify} describes what it is
given; it does not guess what you meant. A grouping variable stored as
\verb{1, 2, 3} will be described as a number.

This is why the data must be an object you already have in your session.
There is no file to choose from inside the application, and no file path to
hand it: a data set that arrives by being read from disk arrives with types
that a reader guessed, and it is then described on those guessed types
without anyone having looked at them. Read the file yourself, run
\code{\link[=str]{str()}} or \code{\link[=summary]{summary()}} over the result, fix the types that are wrong, and
pass that object:

\if{html}{\out{<div class="sourceCode">}}\preformatted{cohort <- read.csv("cohort.csv")
cohort$treatment <- factor(cohort$treatment,
                           levels = c("Control", "Low dose", "High dose"))
str(cohort)
ggstratify(cohort)
}\if{html}{\out{</div>}}

The figure types, themes and palettes otherwise match those offered by the
\pkg{ggplotgui} package. All data handling uses \pkg{data.table}. Figures
are written as PNG (through \pkg{ragg}) or as SVG, and the "R-code" tab
shows the self-contained \pkg{ggplot2} code for the figure on screen --
just the figure, since writing the files is what the export button is for.
The preview, the export button and that code are all produced by the same
code generator, so the code you are shown is the code that made the
figure.
}

\examples{
if (interactive()) {
  ggstratify(iris)
  ggstratify(epi_cohort)
}
}
